@@ -21,7 +21,7 @@ my %languages = (
     lua => {
         start  => qr/^\s*\--(?:\[\[\[+|-+)\s*$/,
         end    => qr/^\s*--(:?\]\]+|-+)\s*$/,
-        filter => qr/^(?:\s*--!?\s?)?(\s*\S?.+)\s*$/,
+        filter => qr/^(?:\s*--!?\s?)?(\s*\S.+)\s*$/,
     },
     sql => {
         start  => qr/^\s*\--(?:\[\[+|-+)\s*$/,
