@@ -18,7 +18,6 @@
 -- Definitions of header regexps
 
 local reconf = config['regexp']
-local rspamd_regexp = require "rspamd_regexp"
 
 -- Subject needs encoding
 -- Define encodings types
@@ -107,9 +106,17 @@ reconf['MISSING_TO'] = {
 }
 
 -- Detects undisclosed recipients
-local undisc_rcpt = 'To=/^<?undisclosed[- ]recipient/Hi'
 reconf['R_UNDISC_RCPT'] = {
-  re = string.format('(%s)', undisc_rcpt),
+  -- match:
+  -- To: undisclosed-recipients:;
+  -- To: Undisclosed recipients:;
+  -- To: undisclosed-recipients: ;
+  -- To: <Undisclosed-Recipient:;>
+  -- To: "undisclosed-recipients (utajeni adresati)": ;
+  -- To: Undisclosed recipients:
+  -- but do not match:
+  -- Undisclosed Recipient <user@example.org>
+  re = [[To=/^[<"]?undisclosed[- ]recipients?\b.*:/i{header}]],
   score = 3.0,
   description = 'Recipients are absent or undisclosed',
   group = 'headers',
@@ -139,22 +146,13 @@ reconf['R_RCVD_SPAMBOTS'] = {
 -- Charset is missing in message
 reconf['R_MISSING_CHARSET'] = {
   re = string.format('!is_empty_body() & content_type_is_type(text) & content_type_is_subtype(plain) & !content_type_has_param(charset) & !%s',
-    'compare_transfer_encoding(7bit)'),
-  score = 2.5,
+      'compare_transfer_encoding(7bit)'),
+  score = 0.5,
   description = 'Charset is missing in a message',
   group = 'headers',
   mime_only = true,
 }
 
--- Subject seems to be spam
-reconf['R_SAJDING'] = {
-  re = 'Subject=/\\bsajding(?:om|a)?\\b/iH',
-  score = 8.0,
-  description = 'Subject seems to be spam',
-  group = 'headers',
-  mime_only = true,
-}
-
 -- Find forged Outlook MUA
 -- Yahoo groups messages
 local yahoo_bulk = 'Received=/from \\[\\S+\\] by \\S+\\.(?:groups|scd|dcn)\\.yahoo\\.com with NNFMP/H'
@@ -316,10 +314,12 @@ local hotmail_baydav_msgid = 'Message-Id=/^<?BAY\\d+-DAV\\d+[A-Z0-9]{25}\\@phx\\
 -- Sympatico message id
 local sympatico_msgid = 'Message-Id=/^<?BAYC\\d+-PASMTP\\d+[A-Z0-9]{25}\\@CEZ\\.ICE>?$/H'
 -- Mailman message id
-local mailman_msgid = 'Message-ID=/^<mailman\\.\\d+\\.\\d+\\.\\d+\\..+\\@\\S+>$/H'
+-- https://bazaar.launchpad.net/~mailman-coders/mailman/2.1/view/head:/Mailman/Utils.py#L811
+local mailman_msgid = [[Message-ID=/^<mailman\.\d+\.\d+\.\d+\.[-+.:=\w]+@[-a-zA-Z\d.]+>$/H]]
 -- Message id seems to be forged
 local unusable_msgid = string.format('(%s | %s | %s | %s | %s | %s)',
-					lyris_ezml_remailer, wacky_sendmail_version, iplanet_messaging_server, hotmail_baydav_msgid, sympatico_msgid, mailman_msgid)
+    lyris_ezml_remailer, wacky_sendmail_version,
+    iplanet_messaging_server, hotmail_baydav_msgid, sympatico_msgid, mailman_msgid)
 -- Outlook express data seems to be forged
 local forged_oe = string.format('(%s & !%s & !%s & !%s)', oe_mua, oe_msgid_1, oe_msgid_2, unusable_msgid)
 -- Outlook specific headers
@@ -329,7 +329,7 @@ local vista_msgid = 'Message-Id=/^<?[A-F\\d]{32}\\@\\S+>?$/H'
 local ims_msgid = 'Message-Id=/^<?[A-F\\d]{36,40}\\@\\S+>?$/H'
 -- Forged outlook headers
 local forged_outlook_dollars = string.format('(%s & !%s & !%s & !%s & !%s & !%s)',
-					outlook_dollars_mua, oe_msgid_2, outlook_dollars_other, vista_msgid, ims_msgid, unusable_msgid)
+    outlook_dollars_mua, oe_msgid_2, outlook_dollars_other, vista_msgid, ims_msgid, unusable_msgid)
 -- Outlook versions that should be excluded from summary rule
 local fmo_excl_o3416 = 'X-Mailer=/^Microsoft Outlook, Build 10.0.3416$/H'
 local fmo_excl_oe3790 = 'X-Mailer=/^Microsoft Outlook Express 6.00.3790.3959$/H'
@@ -415,23 +415,6 @@ reconf['FORGED_MUA_THEBAT_MSGID_UNKNOWN'] = {
 local kmail_mua = 'User-Agent=/^\\s*KMail\\/1\\.\\d+\\.\\d+/H'
 -- KMail common Message-ID template
 local kmail_msgid_common = 'Message-Id=/^<?\\s*\\d+\\.\\d+\\.\\S+\\@\\S+>?$/mH'
-function kmail_msgid (task)
-	local regexp_text = '<(\\S+)>\\|(19[789]\\d|20\\d\\d)(0\\d|1[012])([012]\\d|3[01])([0-5]\\d)([0-5]\\d)\\.\\d+\\.\\1$'
-	local re = rspamd_regexp.create_cached(regexp_text)
-	local header_msgid = task:get_header('Message-Id')
-	if header_msgid then
-		local header_from = task:get_header('From')
-		if header_from and re:match(header_from.."|"..header_msgid) then return true end
-	end
-	return false
-end
--- Summary rule for forged KMail Message-ID header
-reconf['FORGED_MUA_KMAIL_MSGID'] = {
-  re = string.format('(%s) & (%s) & !(%s) & !(%s)', kmail_mua, kmail_msgid_common, 'kmail_msgid', unusable_msgid),
-  score = 3.0,
-  description = 'Message pretends to be send from KMail but has forged Message-ID',
-  group = 'mua'
-}
 -- Summary rule for forged KMail Message-ID header with unknown template
 reconf['FORGED_MUA_KMAIL_MSGID_UNKNOWN'] = {
   re = string.format('(%s) & !(%s) & !(%s)', kmail_mua, kmail_msgid_common, unusable_msgid),
@@ -445,20 +428,9 @@ reconf['FORGED_MUA_KMAIL_MSGID_UNKNOWN'] = {
 local opera1x_mua = 'User-Agent=/^\\s*Opera Mail\\/1[01]\\.\\d+ /H'
 -- Opera Mail Message-ID template
 local opera1x_msgid = 'Message-ID=/^<?op\\.[a-z\\d]{14}\\@\\S+>?$/H'
--- Suspicious Opera Mail User-Agent header
-local suspicious_opera10w_mua = 'User-Agent=/^\\s*Opera Mail\\/10\\.\\d+ \\(Windows\\)$/H'
--- Suspicious Opera Mail Message-ID, apparently from KMail
-local suspicious_opera10w_msgid = 'Message-Id=/^<?2009\\d{8}\\.\\d+\\.\\S+\\@\\S+?>$/H'
--- Summary rule for forged Opera Mail User-Agent header and Message-ID header from KMail
-reconf['SUSPICIOUS_OPERA_10W_MSGID'] = {
-  re = string.format('(%s) & (%s)', suspicious_opera10w_mua, suspicious_opera10w_msgid),
-  score = 4.0,
-  description = 'Message pretends to be send from suspicious Opera Mail/10.x (Windows) but has forged Message-ID, apparently from KMail',
-  group = 'mua'
-}
--- Summary rule for forged Opera Mail Message-ID header
+-- Rule for forged Opera Mail Message-ID header
 reconf['FORGED_MUA_OPERA_MSGID'] = {
-  re = string.format('(%s) & !(%s) & !(%s) & !(%s)', opera1x_mua, opera1x_msgid, reconf['SUSPICIOUS_OPERA_10W_MSGID']['re'], unusable_msgid),
+  re = string.format('(%s) & !(%s) & !(%s)', opera1x_mua, opera1x_msgid, unusable_msgid),
   score = 4.0,
   description = 'Message pretends to be send from Opera Mail but has forged Message-ID',
   group = 'mua'
@@ -474,7 +446,7 @@ local user_agent_mozilla	= string.format('(%s) & !(%s) & !(%s) & !(%s)', user_ag
 -- Mozilla based common Message-ID template
 local mozilla_msgid_common	= 'Message-ID=/^\\s*<[\\dA-F]{8}\\.\\d{1,7}\\@([^>\\.]+\\.)+[^>\\.]+>$/H'
 local mozilla_msgid_common_sec	= 'Message-ID=/^\\s*<[\\da-f]{8}-([\\da-f]{4}-){3}[\\da-f]{12}\\@([^>\\.]+\\.)+[^>\\.]+>$/H'
-local mozilla_msgid		= 'Message-ID=/^\\s*<(3[3-9A-F]|4[\\dA-F]|5[\\dA-F])[\\dA-F]{6}\\.(\\d0){1,4}\\d\\@([^>\\.]+\\.)+[^>\\.]+>$/H'
+local mozilla_msgid		= 'Message-ID=/^\\s*<(3[3-9A-F]|[4-9A-F][\\dA-F])[\\dA-F]{6}\\.(\\d0){1,4}\\d\\@([^>\\.]+\\.)+[^>\\.]+>$/H'
 -- Summary rule for forged Mozilla Mail Message-ID header
 reconf['FORGED_MUA_MOZILLA_MAIL_MSGID'] = {
   re = string.format('(%s) & (%s) & !(%s) & !(%s)', user_agent_mozilla, mozilla_msgid_common, mozilla_msgid, unusable_msgid),
@@ -574,16 +546,6 @@ reconf['STOX_REPLY_TYPE'] = {
   group = 'headers'
 }
 
--- Fake Verizon headers
-local fhelo_verizon = 'X-Spam-Relays-Untrusted=/^[^\\]]+ helo=[^ ]+verizon\\.net /iH'
-local fhost_verizon = 'X-Spam-Relays-Untrusted=/^[^\\]]+ rdns=[^ ]+verizon\\.net /iH'
-reconf['FM_FAKE_HELO_VERIZON'] = {
-  re = string.format('(%s) & !(%s)', fhelo_verizon, fhost_verizon),
-  score = 2.0,
-  description = 'Fake helo for verizon provider',
-  group = 'headers'
-}
-
 -- Forged yahoo msgid
 local at_yahoo_msgid = 'Message-Id=/\\@yahoo\\.com\\b/iH'
 local from_yahoo_com = 'From=/\\@yahoo\\.com\\b/iH'
@@ -628,18 +590,18 @@ reconf['YANDEX_RU_MAILER'] = {
 
 -- Detect 1C v8.2 and v8.3 mailers
 reconf['MAILER_1C_8'] = {
-    re = 'X-Mailer=/^1C:Enterprise 8\\.[23]$/H',
-    score = 0.0,
-    description = 'Sent with 1C:Enterprise 8',
-    group = 'headers'
+  re = 'X-Mailer=/^1C:Enterprise 8\\.[23]$/H',
+  score = 0.0,
+  description = 'Sent with 1C:Enterprise 8',
+  group = 'headers'
 }
 
 -- Detect rogue 'strongmail' MTA with IPv4 and '(-)' in Received line
 reconf['STRONGMAIL'] = {
-    re = [[Received=/^from\s+strongmail\s+\(\[\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\]\) by \S+ \(-\); /mH]],
-    score = 6.0,
-    description = 'Sent via rogue "strongmail" MTA',
-    group = 'headers'
+  re = [[Received=/^from\s+strongmail\s+\(\[\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\]\) by \S+ \(-\); /mH]],
+  score = 6.0,
+  description = 'Sent via rogue "strongmail" MTA',
+  group = 'headers'
 }
 
 -- Two received headers with ip addresses
@@ -661,31 +623,14 @@ reconf['REPTO_QUOTE_YAHOO'] = {
   group = 'headers'
 }
 
--- MUA definitions
-local xm_gnus = 'X-Mailer=/^Gnus v/H'
-local xm_msoe5 = 'X-Mailer=/^Microsoft Outlook Express 5/H'
-local xm_msoe6 = 'X-Mailer=/^Microsoft Outlook Express 6/H'
-local xm_moz4 = 'X-Mailer=/^Mozilla 4/H'
-local xm_skyri = 'X-Mailer=/^SKYRiXgreen/H'
-local xm_wwwmail = 'X-Mailer=/^WWW-Mail \\d/H'
-local ua_gnus = 'User-Agent=/^Gnus/H'
-local ua_knode = 'User-Agent=/^KNode/H'
-local ua_mutt = 'User-Agent=/^Mutt/H'
-local ua_pan = 'User-Agent=/^Pan/H'
-local ua_xnews = 'User-Agent=/^Xnews/H'
-local no_inr_yes_ref = string.format('(%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s)', xm_gnus, xm_msoe5, xm_msoe6, xm_moz4, xm_skyri, xm_wwwmail, ua_gnus, ua_knode, ua_mutt, ua_pan, ua_xnews)
-local subj_re = 'Subject=/^R[eE]:/H'
-local has_ref = '(header_exists(References) | header_exists(In-Reply-To))'
-local missing_ref = string.format('!(%s)', has_ref)
--- Fake reply (has RE in subject, but has not References header)
-reconf['FAKE_REPLY_C'] = {
-  re = string.format('(%s) & (%s) & (%s) & !(%s)', subj_re, missing_ref, no_inr_yes_ref, xm_msoe6),
-  score = 6.0,
-  description = 'Fake reply (has RE in subject, but has not References header)',
-  group = 'subject'
+reconf['FAKE_REPLY'] = {
+  re = [[Subject=/^re:/i{header} & !(header_exists(In-Reply-To) | header_exists(References))]],
+  description = 'Fake reply',
+  score = 1.0,
+  group = 'headers'
 }
 
--- Mime-OLE is needed but absent (e.g. fake Outlook or fake Ecxchange)
+-- Mime-OLE is needed but absent (e.g. fake Outlook or fake Exchange)
 local has_msmail_pri = 'header_exists(X-MSMail-Priority)'
 local has_mimeole = 'header_exists(X-MimeOLE)'
 local has_squirrelmail_in_mailer = 'X-Mailer=/SquirrelMail\\b/H'
@@ -701,85 +646,59 @@ reconf['MISSING_MIMEOLE'] = {
   group = 'headers'
 }
 
--- Header delimiters
-local yandex_from = 'From=/\\@(yandex\\.ru|yandex\\.net|ya\\.ru)/iX'
-local yandex_x_envelope_from = 'X-Envelope-From=/\\@(yandex\\.ru|yandex\\.net|ya\\.ru)/iX'
-local yandex_return_path = 'Return-Path=/\\@(yandex\\.ru|yandex\\.net|ya\\.ru)/iX'
-local yandex_received = 'Received=/^\\s*from \\S+\\.(yandex\\.ru|yandex\\.net)/mH'
-local yandex = string.format('(%s) & ((%s) | (%s) | (%s))', yandex_received, yandex_from, yandex_x_envelope_from, yandex_return_path)
--- Tabs as delimiters between header names and header values
-function check_header_delimiter_tab(task, header_name)
-	for _,rh in ipairs(task:get_header_full(header_name)) do
-		if rh['tab_separated'] then return true end
-	end
-	return false
-end
-reconf['HEADER_FROM_DELIMITER_TAB'] = {
-  re = string.format('(%s) & !(%s)', 'check_header_delimiter_tab(From)', yandex),
-  score = 1.0,
-  description = 'Header From begins with tab',
-  group = 'headers'
-}
-reconf['HEADER_TO_DELIMITER_TAB'] = {
-  re = string.format('(%s) & !(%s)', 'check_header_delimiter_tab(To)', yandex),
-  score = 1.0,
-  description = 'Header To begins with tab',
-  group = 'headers'
-}
-reconf['HEADER_CC_DELIMITER_TAB'] = {
-  re = string.format('(%s) & !(%s)', 'check_header_delimiter_tab(Cc)', yandex),
-  score = 1.0,
-  description = 'Header To begins with tab',
-  group = 'headers'
-}
-reconf['HEADER_REPLYTO_DELIMITER_TAB'] = {
-  re = string.format('(%s) & !(%s)', 'check_header_delimiter_tab(Reply-To)', yandex),
-  score = 1.0,
-  description = 'Header Reply-To begins with tab',
-  group = 'headers'
-}
-reconf['HEADER_DATE_DELIMITER_TAB'] = {
-  re = string.format('(%s) & !(%s)', 'check_header_delimiter_tab(Date)', yandex),
-  score = 1.0,
-  description = 'Header Date begins with tab',
-  group = 'headers'
-}
 -- Empty delimiters between header names and header values
-function check_header_delimiter_empty(task, header_name)
-	for _,rh in ipairs(task:get_header_full(header_name)) do
-		if rh['empty_separator'] then return true end
-	end
-	return false
+local function gen_check_header_delimiter_empty(header_name)
+  return function(task)
+    for _,rh in ipairs(task:get_header_full(header_name) or {}) do
+      if rh['empty_separator'] then return true end
+    end
+    return false
+  end
 end
 reconf['HEADER_FROM_EMPTY_DELIMITER'] = {
-  re = string.format('(%s)', 'check_header_delimiter_empty(From)'),
+  re = string.format('(%s)', 'lua:check_from_delim_empty'),
   score = 1.0,
   description = 'Header From has no delimiter between header name and header value',
-  group = 'headers'
+  group = 'headers',
+  functions = {
+    check_from_delim_empty = gen_check_header_delimiter_empty('From')
+  }
 }
 reconf['HEADER_TO_EMPTY_DELIMITER'] = {
-  re = string.format('(%s)', 'check_header_delimiter_empty(To)'),
+  re = string.format('(%s)', 'lua:check_to_delim_empty'),
   score = 1.0,
   description = 'Header To has no delimiter between header name and header value',
-  group = 'headers'
+  group = 'headers',
+  functions = {
+    check_to_delim_empty = gen_check_header_delimiter_empty('To')
+  }
 }
 reconf['HEADER_CC_EMPTY_DELIMITER'] = {
-  re = string.format('(%s)', 'check_header_delimiter_empty(Cc)'),
+  re = string.format('(%s)', 'lua:check_cc_delim_empty'),
   score = 1.0,
   description = 'Header Cc has no delimiter between header name and header value',
-  group = 'headers'
+  group = 'headers',
+  functions = {
+    check_cc_delim_empty = gen_check_header_delimiter_empty('Cc')
+  }
 }
 reconf['HEADER_REPLYTO_EMPTY_DELIMITER'] = {
-  re = string.format('(%s)', 'check_header_delimiter_empty(Reply-To)'),
+  re = string.format('(%s)', 'lua:check_repto_delim_empty'),
   score = 1.0,
   description = 'Header Reply-To has no delimiter between header name and header value',
-  group = 'headers'
+  group = 'headers',
+  functions = {
+    check_repto_delim_empty = gen_check_header_delimiter_empty('Reply-To')
+  }
 }
 reconf['HEADER_DATE_EMPTY_DELIMITER'] = {
-  re = string.format('(%s)', 'check_header_delimiter_empty(Date)'),
+  re = string.format('(%s)', 'lua:check_date_delim_empty'),
   score = 1.0,
   description = 'Header Date has no delimiter between header name and header value',
-  group = 'headers'
+  group = 'headers',
+  functions = {
+    check_date_delim_empty = gen_check_header_delimiter_empty('Date')
+  }
 }
 
 -- Definitions of received headers regexp
@@ -796,7 +715,8 @@ local MAIL_RU_From		= 'From=/\\@mail\\.ru>?$/iX'
 local MAIL_RU_Received		= 'Received=/from mail\\.ru \\(/mH'
 
 reconf['FAKE_RECEIVED_mail_ru'] = {
-  re = string.format('(%s) & !(((%s) | (%s)) & (%s))', MAIL_RU_Received, MAIL_RU_Return_Path, MAIL_RU_X_Envelope_From, MAIL_RU_From),
+  re = string.format('(%s) & !(((%s) | (%s)) & (%s))',
+      MAIL_RU_Received, MAIL_RU_Return_Path, MAIL_RU_X_Envelope_From, MAIL_RU_From),
   score = 4.0,
   description = 'Fake helo mail.ru in header Received from non mail.ru sender address',
   group = 'headers'
@@ -821,7 +741,14 @@ local RECEIVED_smtp_yandex_ru_8	= 'Received=/from \\S+ \\(HELO smtp\\.yandex\\.r
 local RECEIVED_smtp_yandex_ru_9	= 'Received=/from \\S+ \\(\\[\\d+\\.\\d+\\.\\d+\\.\\d+\\] helo=smtp\\.yandex\\.ru\\)/iX'
 
 reconf['FAKE_RECEIVED_smtp_yandex_ru'] = {
-  re = string.format('(((%s) & ((%s) | (%s))) | ((%s) & ((%s) | (%s))) | ((%s) & ((%s) | (%s)))) & (%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s)', MAIL_RU_From, MAIL_RU_Return_Path, MAIL_RU_X_Envelope_From, GMAIL_COM_From, GMAIL_COM_Return_Path, GMAIL_COM_X_Envelope_From, UKR_NET_From, UKR_NET_Return_Path, UKR_NET_X_Envelope_From, RECEIVED_smtp_yandex_ru_1, RECEIVED_smtp_yandex_ru_2, RECEIVED_smtp_yandex_ru_3, RECEIVED_smtp_yandex_ru_4, RECEIVED_smtp_yandex_ru_5, RECEIVED_smtp_yandex_ru_6, RECEIVED_smtp_yandex_ru_7, RECEIVED_smtp_yandex_ru_8, RECEIVED_smtp_yandex_ru_9),
+  re = string.format('(((%s) & ((%s) | (%s))) | ((%s) & ((%s) | (%s))) '..
+      ' | ((%s) & ((%s) | (%s)))) & (%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s) | (%s)',
+      MAIL_RU_From, MAIL_RU_Return_Path, MAIL_RU_X_Envelope_From, GMAIL_COM_From,
+      GMAIL_COM_Return_Path, GMAIL_COM_X_Envelope_From, UKR_NET_From, UKR_NET_Return_Path,
+      UKR_NET_X_Envelope_From, RECEIVED_smtp_yandex_ru_1, RECEIVED_smtp_yandex_ru_2,
+      RECEIVED_smtp_yandex_ru_3, RECEIVED_smtp_yandex_ru_4, RECEIVED_smtp_yandex_ru_5,
+      RECEIVED_smtp_yandex_ru_6, RECEIVED_smtp_yandex_ru_7, RECEIVED_smtp_yandex_ru_8,
+      RECEIVED_smtp_yandex_ru_9),
   score = 4.0,
   description = 'Fake smtp.yandex.ru Received',
   group = 'headers'
@@ -892,11 +819,11 @@ reconf['CTE_CASE'] = {
 
 reconf['HAS_INTERSPIRE_SIG'] = {
   re = string.format('((%s) & (%s) & (%s) & (%s)) | (%s)',
-                     'header_exists(X-Mailer-LID)',
-                     'header_exists(X-Mailer-RecptId)',
-                     'header_exists(X-Mailer-SID)',
-                     'header_exists(X-Mailer-Sent-By)',
-                     'List-Unsubscribe=/\\/unsubscribe\\.php\\?M=[^&]+&C=[^&]+&L=[^&]+&N=[^>]+>$/Xi'),
+      'header_exists(X-Mailer-LID)',
+      'header_exists(X-Mailer-RecptId)',
+      'header_exists(X-Mailer-SID)',
+      'header_exists(X-Mailer-Sent-By)',
+      'List-Unsubscribe=/\\/unsubscribe\\.php\\?M=[^&]+&C=[^&]+&L=[^&]+&N=[^>]+>$/Xi'),
   description = "Has Interspire fingerprint",
   score = 1.0,
   group = 'headers'
@@ -911,14 +838,14 @@ reconf['CT_EXTRA_SEMI'] = {
 
 reconf['SUBJECT_ENDS_EXCLAIM'] = {
   re = 'Subject=/!\\s*$/H',
-  description = 'Subject ends with an exclaimation',
+  description = 'Subject ends with an exclamation',
   score = 0.0,
   group = 'headers'
 }
 
 reconf['SUBJECT_HAS_EXCLAIM'] = {
   re = string.format('%s & !%s', 'Subject=/!/H', 'Subject=/!\\s*$/H'),
-  description = 'Subject contains an exclaimation',
+  description = 'Subject contains an exclamation',
   score = 0.0,
   group = 'headers'
 }
@@ -938,7 +865,7 @@ reconf['SUBJECT_HAS_QUESTION'] = {
 }
 
 reconf['SUBJECT_HAS_CURRENCY'] = {
-  re = 'Subject=/[$€$¢¥₽]/Hu',
+  re = 'Subject=/\\p{Sc}/Hu',
   description = 'Subject contains currency',
   score = 1.0,
   group = 'headers'
@@ -980,10 +907,10 @@ reconf['HAS_LIST_UNSUB'] = {
 }
 
 reconf['HAS_GUC_PROXY_URI'] = {
-   re = '/\\.googleusercontent\\.com\\/proxy/{url}i',
-   description = 'Has googleusercontent.com proxy URI',
-   score = 0.01,
-   group = 'experimental'
+  re = '/\\.googleusercontent\\.com\\/proxy/{url}i',
+  description = 'Has googleusercontent.com proxy URI',
+  score = 0.01,
+  group = 'experimental'
 }
 
 reconf['HAS_GOOGLE_REDIR'] = {
@@ -995,12 +922,80 @@ reconf['HAS_GOOGLE_REDIR'] = {
 
 reconf['XM_UA_NO_VERSION'] = {
   re = string.format('(!%s && !%s) && (%s || %s)',
-                     'X-Mailer=/https?:/H',
-                     'User-Agent=/https?:/H',
-                     'X-Mailer=/^[^0-9]+$/H',
-                     'User-Agent=/^[^0-9]+$/H'),
+      'X-Mailer=/https?:/H',
+      'User-Agent=/https?:/H',
+      'X-Mailer=/^[^0-9]+$/H',
+      'User-Agent=/^[^0-9]+$/H'),
   description = 'X-Mailer/User-Agent has no version',
   score = 0.01,
   group = 'experimental'
 }
 
+-- X-Mailer for old MUA versions which are forged by spammers
+local old_x_mailers = {
+  -- Outlook Express 6.0 was last included in Windows XP (EOL 2014).  Windows
+  -- XP is still used (in 2020) by relatively small number of internet users,
+  -- but this header is widely abused by spammers.
+  'Microsoft Outlook Express',
+  -- Qualcomm Eudora for Windows 7.1.0.9 was released in 2006
+  [[QUALCOMM Windows Eudora (Pro )?Version [1-6]\.]],
+  -- The Bat 3.0 was released in 2004
+  [[The Bat! \(v[12]\.]],
+  -- Can be found in public maillist archives, messages circa 2000
+  [[Microsoft Outlook IMO, Build 9\.0\.]],
+  -- Outlook 2002 (Office XP)
+  [[Microsoft Outlook, Build 10\.]],
+  -- Some old Apple iOS versions are used on old devices, match only very old
+  -- versions (iOS 4.3.5 buid 8L1 was supported until 2013) and less old
+  -- versions frequently seen in spam
+  [[i(Phone|Pad) Mail \((?:[1-8][A-L]|12H|13E)]],
+}
+
+reconf['OLD_X_MAILER'] = {
+  description = 'X-Mailer has a very old MUA version',
+  re = string.format('X-Mailer=/^(?:%s)/{header}', table.concat(old_x_mailers, '|')),
+  score = 2.0,
+  group = 'headers',
+}
+
+-- X-Mailer header values which should not occur (in the modern mail) at all
+local bad_x_mailers = {
+  -- header name repeated in the header value
+  [[X-Mailer: ]],
+  -- Mozilla Thunderbird uses User-Agent header, not X-Mailer
+  -- Early Thunderbird had U-A like:
+  -- Mozilla Thunderbird 1.0.2 (Windows/20050317)
+  -- Thunderbird 2.0.0.23 (X11/20090812)
+  [[(?:Mozilla )?Thunderbird \d]],
+  -- Was used by Yahoo Groups in 2000s, no one expected to use this in 2020s
+  [[eGroups Message Poster]],
+  -- Regexp for genuine iOS X-Mailer is below, anything which doesn't match it,
+  -- but starts with 'iPhone Mail' or 'iPad Mail' is likely fake
+  [[i(?:Phone|Pad) Mail]],
+}
+-- Apple iPhone/iPad Mail X-Mailer contains iOS build number, e. g. 9B206, 16H5, 18G5023c
+-- https://en.wikipedia.org/wiki/IOS_version_history
+local apple_ios_x_mailer = [[i(?:Phone|Pad) Mail \((?:1[AC]|[34][AB]|5[ABCFGH]|7[A-E]|8[ABCEFGHJKL]|9[AB]|\d{2}[A-Z])\d+[a-z]?\)]]
+
+reconf['FORGED_X_MAILER'] = {
+  description = 'Forged X-Mailer header',
+  re = string.format('X-Mailer=/^(?:%s)/{header} && !X-Mailer=/^%s/{header}',
+      table.concat(bad_x_mailers, '|'), apple_ios_x_mailer),
+  score = 4.5,
+  group = 'headers',
+}
+
+-- X-Mailer headers like: 'Internet Mail Service (5.5.2650.21)' are being
+-- forged by spammers, but MS Exchange 5.5 is still being used (in 2020) on
+-- some mail servers. Example of genuine headers (DC-EXMPL is a hostname which
+-- can be a FQDN):
+-- Received: by DC-EXMPL with Internet Mail Service (5.5.2656.59)
+-- 	id <HKH4BJQX>; Tue, 8 Dec 2020 07:10:54 -0600
+-- Message-ID: <E7209F9DB64FCC4BB1051420F0E955DD05C9D59F@DC-EXMPL>
+-- X-Mailer: Internet Mail Service (5.5.2656.59)
+reconf['FORGED_IMS'] = {
+  description = 'Forged X-Mailer: Internet Mail Service',
+  re = [[X-Mailer=/^Internet Mail Service \(5\./{header} & !Received=/^by \S+ with Internet Mail Service \(5\./{header}]],
+  score = 3.0,
+  group = 'headers',
+}
