@@ -54,27 +54,27 @@ static const gchar *local_db_name = "local";
 
 #define msg_err_fuzzy_update(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         session->name, session->uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_fuzzy_update(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         session->name, session->uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_fuzzy_update(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         session->name, session->uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_err_fuzzy_collection(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         "fuzzy_collection", session->uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_fuzzy_collection(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
        "fuzzy_collection", session->uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_fuzzy_collection(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
        "fuzzy_collection", session->uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 /* Init functions */
@@ -112,6 +112,7 @@ struct fuzzy_key_stat {
 	guint64 added;
 	guint64 deleted;
 	guint64 errors;
+	struct rspamd_cryptobox_keypair *keypair;
 	rspamd_lru_hash_t *last_ips;
 	ref_entry_t ref;
 };
@@ -144,11 +145,13 @@ struct rspamd_fuzzy_storage_ctx {
 	gdouble sync_timeout;
 	gdouble delay;
 	struct rspamd_radix_map_helper *update_ips;
+	struct rspamd_hash_map_helper *update_keys;
 	struct rspamd_radix_map_helper *blocked_ips;
 	struct rspamd_radix_map_helper *ratelimit_whitelist;
 	struct rspamd_radix_map_helper *delay_whitelist;
 
 	const ucl_object_t *update_map;
+	const ucl_object_t *update_keys_map;
 	const ucl_object_t *delay_whitelist_map;
 	const ucl_object_t *blocked_map;
 	const ucl_object_t *ratelimit_whitelist_map;
@@ -249,6 +252,10 @@ rspamd_fuzzy_check_ratelimit (struct fuzzy_session *session)
 	struct rspamd_leaky_bucket_elt *elt;
 	ev_tstamp now;
 
+	if (!session->addr) {
+		return TRUE;
+	}
+
 	if (session->ctx->ratelimit_whitelist != NULL) {
 		if (rspamd_match_radix_map_addr (session->ctx->ratelimit_whitelist,
 				session->addr) != NULL) {
@@ -355,7 +362,10 @@ rspamd_fuzzy_check_write (struct fuzzy_session *session)
 		return FALSE;
 	}
 
-	if (session->ctx->update_ips != NULL) {
+	if (session->ctx->update_ips != NULL && session->addr) {
+		if (rspamd_inet_address_get_af (session->addr) == AF_UNIX) {
+			return TRUE;
+		}
 		if (rspamd_match_radix_map_addr (session->ctx->update_ips,
 				session->addr) == NULL) {
 			return FALSE;
@@ -365,6 +375,20 @@ rspamd_fuzzy_check_write (struct fuzzy_session *session)
 		}
 	}
 
+	if (session->ctx->update_keys != NULL && session->key_stat && session->key_stat->keypair) {
+		static gchar base32_buf[rspamd_cryptobox_HASHBYTES * 2 + 1];
+		guint raw_len;
+		const guchar *pk_raw = rspamd_keypair_component(session->key_stat->keypair,
+				RSPAMD_KEYPAIR_COMPONENT_ID, &raw_len);
+		gint encoded_len = rspamd_encode_base32_buf(pk_raw, raw_len,
+				base32_buf,sizeof(base32_buf),
+				RSPAMD_BASE32_DEFAULT);
+
+		if (rspamd_match_hash_map (session->ctx->update_keys, base32_buf, encoded_len)) {
+			return TRUE;
+		}
+	}
+
 	return FALSE;
 }
 
@@ -377,6 +401,10 @@ fuzzy_key_stat_dtor (gpointer p)
 		rspamd_lru_hash_destroy (st->last_ips);
 	}
 
+	if (st->keypair) {
+		rspamd_keypair_unref(st->keypair);
+	}
+
 	g_free (st);
 }
 
@@ -849,7 +877,12 @@ rspamd_fuzzy_check_callback (struct rspamd_fuzzy_reply *result, void *ud)
 		/* function */
 		lua_rawgeti (L, LUA_REGISTRYINDEX, session->ctx->lua_post_handler_cbref);
 		/* client IP */
-		rspamd_lua_ip_push (L, session->addr);
+		if (session->addr) {
+			rspamd_lua_ip_push(L, session->addr);
+		}
+		else {
+			lua_pushnil (L);
+		}
 		/* client command */
 		lua_pushinteger (L, cmd->cmd);
 		/* command value (push as rspamd_text) */
@@ -1093,7 +1126,7 @@ rspamd_fuzzy_process_command (struct fuzzy_session *session)
 		return;
 	}
 
-	if (session->key_stat) {
+	if (session->key_stat && session->addr) {
 		ip_stat = rspamd_lru_hash_lookup (session->key_stat->last_ips,
 				session->addr, -1);
 
@@ -1554,6 +1587,13 @@ fuzzy_session_destroy (gpointer d)
 #define MSGVEC_LEN 1
 #endif
 
+union sa_union {
+	struct sockaddr sa;
+	struct sockaddr_in s4;
+	struct sockaddr_in6 s6;
+	struct sockaddr_un su;
+	struct sockaddr_storage ss;
+};
 /*
  * Accept new connection and construct task
  */
@@ -1566,7 +1606,7 @@ accept_fuzzy_socket (EV_P_ ev_io *w, int revents)
 	guint64 *nerrors;
 	struct iovec iovs[MSGVEC_LEN];
 	guint8 bufs[MSGVEC_LEN][FUZZY_INPUT_BUFLEN];
-	struct sockaddr_storage peer_sa[MSGVEC_LEN];
+	union sa_union peer_sa[MSGVEC_LEN];
 	socklen_t salen = sizeof (peer_sa[0]);
 #ifdef HAVE_RECVMMSG
 #define MSG_FIELD(msg, field) msg.msg_hdr.field
@@ -1622,13 +1662,17 @@ accept_fuzzy_socket (EV_P_ ev_io *w, int revents)
 			for (int i = 0; i < r; i ++) {
 				rspamd_inet_addr_t *client_addr;
 
-				client_addr = rspamd_inet_address_from_sa (MSG_FIELD(msg[i], msg_name),
-						MSG_FIELD(msg[i], msg_namelen));
-
-				if (!rspamd_fuzzy_check_client (worker->ctx, client_addr)) {
-					/* Disallow forbidden clients silently */
-					rspamd_inet_address_free (client_addr);
-					continue;
+				if (MSG_FIELD(msg[i], msg_namelen) >= sizeof(struct sockaddr)) {
+					client_addr = rspamd_inet_address_from_sa(MSG_FIELD(msg[i], msg_name),
+							MSG_FIELD(msg[i], msg_namelen));
+					if (!rspamd_fuzzy_check_client (worker->ctx, client_addr)) {
+						/* Disallow forbidden clients silently */
+						rspamd_inet_address_free (client_addr);
+						continue;
+					}
+				}
+				else {
+					client_addr = NULL;
 				}
 
 				session = g_malloc0 (sizeof (*session));
@@ -1655,18 +1699,20 @@ accept_fuzzy_socket (EV_P_ ev_io *w, int revents)
 					session->ctx->stat.invalid_requests ++;
 					msg_debug ("invalid fuzzy command of size %z received", r);
 
-					nerrors = rspamd_lru_hash_lookup (session->ctx->errors_ips,
-							session->addr, -1);
+					if (session->addr) {
+						nerrors = rspamd_lru_hash_lookup(session->ctx->errors_ips,
+								session->addr, -1);
 
-					if (nerrors == NULL) {
-						nerrors = g_malloc (sizeof (*nerrors));
-						*nerrors = 1;
-						rspamd_lru_hash_insert (session->ctx->errors_ips,
-								rspamd_inet_address_copy(session->addr, NULL),
-								nerrors, -1, -1);
-					}
-					else {
-						*nerrors = *nerrors + 1;
+						if (nerrors == NULL) {
+							nerrors = g_malloc(sizeof(*nerrors));
+							*nerrors = 1;
+							rspamd_lru_hash_insert(session->ctx->errors_ips,
+									rspamd_inet_address_copy(session->addr, NULL),
+									nerrors, -1, -1);
+						}
+						else {
+							*nerrors = *nerrors + 1;
+						}
 					}
 				}
 
@@ -2106,9 +2152,11 @@ fuzzy_parse_keypair (rspamd_mempool_t *pool,
 		key->stat = keystat;
 		pk = rspamd_keypair_component (kp, RSPAMD_KEYPAIR_COMPONENT_PK,
 				NULL);
+		keystat->keypair = rspamd_keypair_ref(kp);
+		/* We map entries by pubkey in binary form for speed lookup */
 		g_hash_table_insert (ctx->keys, (gpointer)pk, key);
 		ctx->default_key = key;
-		msg_info_pool ("loaded keypair %*xs", 8, pk);
+		msg_debug_pool_check("loaded keypair %*xs", 8, pk);
 	}
 	else if (ucl_object_type (obj) == UCL_ARRAY) {
 		while ((cur = ucl_object_iterate (obj, &it, true)) != NULL) {
@@ -2210,6 +2258,15 @@ init_fuzzy (struct rspamd_config *cfg)
 			0,
 			"Allow modifications from the following IP addresses");
 
+	rspamd_rcl_register_worker_option (cfg,
+			type,
+			"allow_update_keys",
+			rspamd_rcl_parse_struct_ucl,
+			ctx,
+			G_STRUCT_OFFSET (struct rspamd_fuzzy_storage_ctx, update_keys_map),
+			0,
+			"Allow modifications for those using specific public keys");
+
 	rspamd_rcl_register_worker_option (cfg,
 			type,
 			"delay_whitelist",
@@ -2528,13 +2585,30 @@ start_fuzzy (struct rspamd_worker *worker)
 	rspamd_control_worker_add_cmd_handler (worker, RSPAMD_CONTROL_FUZZY_SYNC,
 			rspamd_fuzzy_storage_sync, ctx);
 
-	/* Create radix trees */
+
 	if (ctx->update_map != NULL) {
 		rspamd_config_radix_from_ucl (worker->srv->cfg, ctx->update_map,
 				"Allow fuzzy updates from specified addresses",
 				&ctx->update_ips, NULL, worker, "fuzzy update");
 	}
 
+	if (ctx->update_keys_map != NULL) {
+		struct rspamd_map *m;
+
+		if ((m = rspamd_map_add_from_ucl (worker->srv->cfg, ctx->update_keys_map,
+				"Allow fuzzy updates from specified public keys",
+				rspamd_kv_list_read,
+				rspamd_kv_list_fin,
+				rspamd_kv_list_dtor,
+				(void **)&ctx->update_keys, worker, RSPAMD_MAP_DEFAULT)) == NULL) {
+			msg_warn_config ("cannot load allow keys map from %s",
+					ucl_object_tostring (ctx->update_keys_map));
+		}
+		else {
+			m->active_http = TRUE;
+		}
+	}
+
 	if (ctx->skip_map != NULL) {
 		struct rspamd_map *m;
 
@@ -2678,6 +2752,7 @@ start_fuzzy (struct rspamd_worker *worker)
 
 	REF_RELEASE (ctx->cfg);
 	rspamd_log_close (worker->srv->logger);
+	rspamd_unset_crash_handler (worker->srv);
 
 	exit (EXIT_SUCCESS);
 }
