@@ -27,6 +27,8 @@
 #include "base64/base64.h"
 #include "ottery.h"
 #include "printf.h"
+#define XXH_INLINE_ALL
+#define XXH_PRIVATE_API
 #include "xxhash.h"
 #define MUM_TARGET_INDEPENDENT_HASH 1 /* For 32/64 bit equal hashes */
 #include "../../contrib/mumhash/mum.h"
@@ -118,7 +120,7 @@ rspamd_cryptobox_test_instr (gint instr)
 	}
 
 	switch (instr) {
-#ifdef HAVE_SSE2
+#if defined HAVE_SSE2 && defined (__x86_64__)
 	case CPUID_SSE2:
 		__asm__ volatile ("psubb %xmm0, %xmm0");
 		break;
@@ -146,7 +148,7 @@ rspamd_cryptobox_test_instr (gint instr)
 		__asm__ volatile ("pcmpeqq %xmm0, %xmm0");
 		break;
 #endif
-#ifdef HAVE_SSE42
+#if defined HAVE_SSE42 && defined(__x86_64__)
 	case CPUID_SSE42:
 		__asm__ volatile ("pushq %rax\n"
 				"xorq %rax, %rax\n"
@@ -586,7 +588,7 @@ static gsize
 rspamd_cryptobox_auth_ctx_len (enum rspamd_cryptobox_mode mode)
 {
 	if (G_LIKELY (mode == RSPAMD_CRYPTOBOX_MODE_25519)) {
-		return sizeof (crypto_onetimeauth_state) + _Alignof (crypto_onetimeauth_state);
+		return sizeof (crypto_onetimeauth_state) + RSPAMD_ALIGNOF(crypto_onetimeauth_state);
 	}
 	else {
 #ifndef HAVE_USABLE_OPENSSL
@@ -1438,7 +1440,7 @@ void
 rspamd_cryptobox_hash_init (rspamd_cryptobox_hash_state_t *p, const guchar *key, gsize keylen)
 {
 	crypto_generichash_blake2b_state *st = cryptobox_align_ptr (p,
-			_Alignof(crypto_generichash_blake2b_state));
+			RSPAMD_ALIGNOF(crypto_generichash_blake2b_state));
 	crypto_generichash_blake2b_init (st, key, keylen,
 			crypto_generichash_blake2b_BYTES_MAX);
 }
@@ -1450,7 +1452,7 @@ void
 rspamd_cryptobox_hash_update (rspamd_cryptobox_hash_state_t *p, const guchar *data, gsize len)
 {
 	crypto_generichash_blake2b_state *st = cryptobox_align_ptr (p,
-			_Alignof(crypto_generichash_blake2b_state));
+			RSPAMD_ALIGNOF(crypto_generichash_blake2b_state));
 	crypto_generichash_blake2b_update (st, data, len);
 }
 
@@ -1461,7 +1463,7 @@ void
 rspamd_cryptobox_hash_final (rspamd_cryptobox_hash_state_t *p, guchar *out)
 {
 	crypto_generichash_blake2b_state *st = cryptobox_align_ptr (p,
-			_Alignof(crypto_generichash_blake2b_state));
+			RSPAMD_ALIGNOF(crypto_generichash_blake2b_state));
 	crypto_generichash_blake2b_final (st, out, crypto_generichash_blake2b_BYTES_MAX);
 }
 
@@ -1480,7 +1482,7 @@ void rspamd_cryptobox_hash (guchar *out,
 
 G_STATIC_ASSERT (sizeof (t1ha_context_t) <=
 		sizeof (((rspamd_cryptobox_fast_hash_state_t *)NULL)->opaque));
-G_STATIC_ASSERT (sizeof (XXH64_state_t) <=
+G_STATIC_ASSERT (sizeof (struct XXH3_state_s) <=
 				 sizeof (((rspamd_cryptobox_fast_hash_state_t *)NULL)->opaque));
 
 
@@ -1493,13 +1495,34 @@ struct RSPAMD_ALIGNED(16) _mum_iuf {
 	unsigned rem;
 };
 
+rspamd_cryptobox_fast_hash_state_t*
+rspamd_cryptobox_fast_hash_new(void)
+{
+	rspamd_cryptobox_fast_hash_state_t *nst;
+	int ret = posix_memalign ((void **)&nst, RSPAMD_ALIGNOF(rspamd_cryptobox_fast_hash_state_t),
+			sizeof(rspamd_cryptobox_fast_hash_state_t));
+
+	if (ret != 0) {
+		abort();
+	}
+
+	return nst;
+}
+
+void
+rspamd_cryptobox_fast_hash_free(rspamd_cryptobox_fast_hash_state_t *st)
+{
+	free(st);
+}
+
 void
 rspamd_cryptobox_fast_hash_init (rspamd_cryptobox_fast_hash_state_t *st,
 		guint64 seed)
 {
-	t1ha_context_t *rst = (t1ha_context_t *)st->opaque;
-	st->type = RSPAMD_CRYPTOBOX_T1HA;
-	t1ha2_init (rst, seed, 0);
+	XXH3_state_t *xst = (XXH3_state_t *)st->opaque;
+	st->type = RSPAMD_CRYPTOBOX_XXHASH3;
+	XXH3_INITSTATE (xst);
+	XXH3_64bits_reset_withSeed (xst, seed);
 }
 
 void
@@ -1518,6 +1541,7 @@ rspamd_cryptobox_fast_hash_init_specific (rspamd_cryptobox_fast_hash_state_t *st
 	}
 	case RSPAMD_CRYPTOBOX_XXHASH64: {
 		XXH64_state_t *xst = (XXH64_state_t *)  st->opaque;
+		memset(xst, 0, sizeof(*xst));
 		st->type = RSPAMD_CRYPTOBOX_XXHASH64;
 		XXH64_reset (xst, seed);
 		break;
@@ -1525,10 +1549,19 @@ rspamd_cryptobox_fast_hash_init_specific (rspamd_cryptobox_fast_hash_state_t *st
 	case RSPAMD_CRYPTOBOX_XXHASH32:
 	{
 		XXH32_state_t *xst = (XXH32_state_t *)  st->opaque;
+		memset(xst, 0, sizeof(*xst));
 		st->type = RSPAMD_CRYPTOBOX_XXHASH32;
 		XXH32_reset (xst, seed);
 		break;
 	}
+	case RSPAMD_CRYPTOBOX_XXHASH3:
+	{
+		XXH3_state_t *xst = (XXH3_state_t *)  st->opaque;
+		XXH3_INITSTATE(xst);
+		st->type = RSPAMD_CRYPTOBOX_XXHASH3;
+		XXH3_64bits_reset_withSeed (xst, seed);
+		break;
+	}
 	case RSPAMD_CRYPTOBOX_MUMHASH: {
 		struct _mum_iuf *iuf = (struct _mum_iuf *)  st->opaque;
 		st->type = RSPAMD_CRYPTOBOX_MUMHASH;
@@ -1544,7 +1577,7 @@ void
 rspamd_cryptobox_fast_hash_update (rspamd_cryptobox_fast_hash_state_t *st,
 		const void *data, gsize len)
 {
-	if (G_LIKELY (st->type) == RSPAMD_CRYPTOBOX_T1HA) {
+	if (st->type == RSPAMD_CRYPTOBOX_T1HA) {
 		t1ha_context_t *rst = (t1ha_context_t *) st->opaque;
 		t1ha2_update (rst, data, len);
 	}
@@ -1561,6 +1594,12 @@ rspamd_cryptobox_fast_hash_update (rspamd_cryptobox_fast_hash_state_t *st,
 			XXH32_update (xst, data, len);
 			break;
 		}
+		case RSPAMD_CRYPTOBOX_XXHASH3:
+		{
+			XXH3_state_t *xst = (XXH3_state_t *)  st->opaque;
+			XXH3_64bits_update (xst, data, len);
+			break;
+		}
 		case RSPAMD_CRYPTOBOX_MUMHASH: {
 			struct _mum_iuf *iuf = (struct _mum_iuf *)  st->opaque;
 			gsize drem = len;
@@ -1614,7 +1653,7 @@ rspamd_cryptobox_fast_hash_final (rspamd_cryptobox_fast_hash_state_t *st)
 {
 	guint64 ret;
 
-	if (G_LIKELY (st->type) == RSPAMD_CRYPTOBOX_T1HA) {
+	if (st->type == RSPAMD_CRYPTOBOX_T1HA) {
 		t1ha_context_t *rst = (t1ha_context_t *) st->opaque;
 
 		return t1ha2_final (rst, NULL);
@@ -1631,6 +1670,11 @@ rspamd_cryptobox_fast_hash_final (rspamd_cryptobox_fast_hash_state_t *st)
 			ret = XXH32_digest (xst);
 			break;
 		}
+		case RSPAMD_CRYPTOBOX_XXHASH3: {
+			XXH3_state_t *xst = (XXH3_state_t *)  st->opaque;
+			ret = XXH3_64bits_digest (xst);
+			break;
+		}
 		case RSPAMD_CRYPTOBOX_MUMHASH: {
 			struct _mum_iuf *iuf = (struct _mum_iuf *)  st->opaque;
 			iuf->h = mum_hash_step (iuf->h, iuf->buf.ll);
@@ -1658,14 +1702,14 @@ static inline guint64
 rspamd_cryptobox_fast_hash_machdep (const void *data,
 		gsize len, guint64 seed)
 {
-	return t1ha2_atonce (data, len, seed);
+	return XXH3_64bits_withSeed(data, len, seed);
 }
 
 static inline guint64
 rspamd_cryptobox_fast_hash_indep (const void *data,
 		gsize len, guint64 seed)
 {
-	return t1ha2_atonce (data, len, seed);
+	return XXH3_64bits_withSeed(data, len, seed);
 }
 
 guint64
@@ -1684,6 +1728,8 @@ rspamd_cryptobox_fast_hash_specific (
 	switch (type) {
 	case RSPAMD_CRYPTOBOX_XXHASH32:
 		return XXH32 (data, len, seed);
+	case RSPAMD_CRYPTOBOX_XXHASH3:
+		return XXH3_64bits_withSeed (data, len, seed);
 	case RSPAMD_CRYPTOBOX_XXHASH64:
 		return XXH64 (data, len, seed);
 	case RSPAMD_CRYPTOBOX_MUMHASH:
