@@ -4,6 +4,10 @@
 #include "config.h"
 #include "fstring.h"
 
+#ifdef  __cplusplus
+extern "C" {
+#endif
+
 struct html_image;
 struct rspamd_task;
 struct rspamd_mime_part;
@@ -35,20 +39,38 @@ struct rspamd_image {
  */
 void rspamd_images_process (struct rspamd_task *task);
 
+/**
+ * Process image if possible in a single mime part
+ * @param task
+ * @param part
+ * @return
+ */
+bool rspamd_images_process_mime_part_maybe (struct rspamd_task *task,
+		struct rspamd_mime_part *part);
+
+/*
+ * Link embedded images to the HTML parts
+ */
+void rspamd_images_link (struct rspamd_task *task);
+
 /**
  * Processes image in raw data
  * @param task
  * @param data
  * @return
  */
-struct rspamd_image* rspamd_maybe_process_image (rspamd_mempool_t *pool,
-		rspamd_ftok_t *data);
+struct rspamd_image *rspamd_maybe_process_image (rspamd_mempool_t *pool,
+												 rspamd_ftok_t *data);
 
 /*
  * Get textual representation of an image's type
  */
-const gchar * rspamd_image_type_str (enum rspamd_image_type type);
+const gchar *rspamd_image_type_str (enum rspamd_image_type type);
 
 void rspamd_image_normalize (struct rspamd_task *task, struct rspamd_image *img);
 
+#ifdef  __cplusplus
+}
+#endif
+
 #endif /* IMAGES_H_ */
