@@ -20,7 +20,7 @@
 #include "rspamd.h"
 #include "message.h"
 #include "mime_expressions.h"
-#include "html.h"
+#include "libserver/html/html.h"
 #include "lua/lua_common.h"
 #include "utlist.h"
 
@@ -784,7 +784,7 @@ rspamd_mime_expr_parse (const gchar *line, gsize len,
 		case bad_atom:
 			g_set_error (err, rspamd_mime_expr_quark(), 100, "cannot parse"
 					" mime atom '%s' when reading symbol '%c' at offset %d, "
-					"near %*.s", line, t, (gint)(p - line),
+					"near %.*s", line, t, (gint)(p - line),
 					(gint)MIN (end - p, 10), p);
 			return NULL;
 		case end_atom:
@@ -827,6 +827,7 @@ set:
 						g_set_error (err, rspamd_mime_expr_quark (), 320,
 								"re_conditions is not a table for '%s'",
 								mime_atom->str);
+						rspamd_regexp_unref (mime_atom->d.re->regexp);
 						goto err;
 					}
 
@@ -838,6 +839,7 @@ set:
 							g_set_error (err, rspamd_mime_expr_quark (), 320,
 									"condition for '%s' is invalid, must be function",
 									mime_atom->str);
+							rspamd_regexp_unref (mime_atom->d.re->regexp);
 							goto err;
 						}
 
@@ -966,14 +968,14 @@ set:
 
 		if (function_obj == NULL) {
 			g_set_error (err, rspamd_mime_expr_quark(), 320,
-					"function %*.s is not found for '%s'",
+					"function %.*s is not found for '%s'",
 					(int)(p - c), c, mime_atom->str);
 			goto err;
 		}
 
 		if (ucl_object_type (function_obj) != UCL_USERDATA) {
 			g_set_error (err, rspamd_mime_expr_quark(), 320,
-					"function %*.s has invalid type for '%s'",
+					"function %.*s has invalid type for '%s'",
 					(int)(p - c), c, mime_atom->str);
 			goto err;
 		}
@@ -1060,28 +1062,34 @@ rspamd_mime_expr_priority (rspamd_expression_atom_t *atom)
 	switch (mime_atom->type) {
 	case MIME_ATOM_INTERNAL_FUNCTION:
 		/* Prioritize internal functions slightly */
-		ret = 50;
+		ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 8;
 		break;
 	case MIME_ATOM_LUA_FUNCTION:
 	case MIME_ATOM_LOCAL_LUA_FUNCTION:
-		ret = 50;
+		ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 4;
 		break;
 	case MIME_ATOM_REGEXP:
 		switch (mime_atom->d.re->type) {
 		case RSPAMD_RE_HEADER:
 		case RSPAMD_RE_RAWHEADER:
-			ret = 100;
+			ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 16;
 			break;
 		case RSPAMD_RE_URL:
 		case RSPAMD_RE_EMAIL:
-			ret = 90;
+			ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 8;
+			break;
+		case RSPAMD_RE_SELECTOR:
+			ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 8;
 			break;
 		case RSPAMD_RE_MIME:
 		case RSPAMD_RE_RAWMIME:
-			ret = 10;
+			ret = RSPAMD_EXPRESSION_MAX_PRIORITY - RSPAMD_EXPRESSION_MAX_PRIORITY / 2;
 			break;
+		case RSPAMD_RE_WORDS:
+		case RSPAMD_RE_RAWWORDS:
+		case RSPAMD_RE_STEMWORDS:
 		default:
-			/* For message regexp */
+			/* For expensive regexps */
 			ret = 0;
 			break;
 		}
@@ -1264,8 +1272,8 @@ rspamd_header_exists (struct rspamd_task * task, GArray * args, void *unused)
 		return FALSE;
 	}
 
-	rh = rspamd_message_get_header_array (task,
-			(gchar *)arg->data);
+	rh = rspamd_message_get_header_array(task,
+			(gchar *) arg->data, FALSE);
 
 	debug_task ("try to get header %s: %d", (gchar *)arg->data,
 			(rh != NULL));
@@ -1564,24 +1572,8 @@ rspamd_compare_transfer_encoding (struct rspamd_task * task,
 gboolean
 rspamd_is_html_balanced (struct rspamd_task * task, GArray * args, void *unused)
 {
-	struct rspamd_mime_text_part *p;
-	guint i;
-	gboolean res = TRUE;
-
-	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, text_parts), i, p) {
-		if (IS_TEXT_PART_HTML (p)) {
-			if (p->flags & RSPAMD_MIME_TEXT_PART_FLAG_BALANCED) {
-				res = TRUE;
-			}
-			else {
-				res = FALSE;
-				break;
-			}
-		}
-	}
-
-	return res;
-
+	/* Totally broken but seems to be never used */
+	return TRUE;
 }
 
 gboolean
@@ -1625,7 +1617,7 @@ rspamd_has_fake_html (struct rspamd_task * task, GArray * args, void *unused)
 	gboolean res = FALSE;
 
 	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, text_parts), i, p) {
-		if (IS_TEXT_PART_HTML (p) && (p->html == NULL || p->html->html_tags == NULL)) {
+		if (IS_TEXT_PART_HTML (p) && (rspamd_html_get_tags_count(p->html) < 2)) {
 			res = TRUE;
 		}
 
@@ -1653,7 +1645,7 @@ rspamd_raw_header_exists (struct rspamd_task *task, GArray * args, void *unused)
 		return FALSE;
 	}
 
-	return rspamd_message_get_header_array (task, arg->data) != NULL;
+	return rspamd_message_get_header_array(task, arg->data, FALSE) != NULL;
 }
 
 static gboolean
@@ -1733,7 +1725,7 @@ rspamd_check_smtp_data (struct rspamd_task *task, GArray * args, void *unused)
 		case 'u':
 		case 'U':
 			if (g_ascii_strcasecmp (type, "user") == 0) {
-				str = task->user;
+				str = task->auth_user;
 			}
 			else {
 				msg_warn_task ("bad argument to function: %s", type);
