@@ -17,9 +17,9 @@
 #include "mime_headers.h"
 #include "smtp_parsers.h"
 #include "mime_encoding.h"
+#include "received.h"
 #include "contrib/uthash/utlist.h"
 #include "libserver/mempool_vars_internal.h"
-#include "libserver/url.h"
 #include "libserver/cfg_file.h"
 #include "libutil/util.h"
 #include <unicode/utf8.h>
@@ -33,18 +33,14 @@ struct rspamd_mime_headers_table {
 	ref_entry_t ref;
 };
 
-#define RSPAMD_INET_ADDRESS_PARSE_RECEIVED \
-	(RSPAMD_INET_ADDRESS_PARSE_REMOTE|RSPAMD_INET_ADDRESS_PARSE_NO_UNIX)
-
 static void
 rspamd_mime_header_check_special (struct rspamd_task *task,
 		struct rspamd_mime_header *rh)
 {
 	guint64 h;
-	struct rspamd_received_header *recv;
 	const gchar *p, *end;
 	gchar *id;
-	gint max_recipients = -1;
+	gint max_recipients = -1, len;
 
 	if (task->cfg) {
 		max_recipients = task->cfg->max_recipients;
@@ -54,16 +50,9 @@ rspamd_mime_header_check_special (struct rspamd_task *task,
 
 	switch (h) {
 	case 0x88705DC4D9D61ABULL:	/* received */
-		recv = rspamd_mempool_alloc0 (task->task_pool,
-				sizeof (struct rspamd_received_header));
-		recv->hdr = rh;
-
-		if (rspamd_smtp_received_parse (task, rh->decoded,
-				strlen (rh->decoded), recv) != -1) {
-			DL_APPEND (MESSAGE_FIELD (task, received), recv);
+		if (rspamd_received_header_parse(task, rh->decoded, strlen (rh->decoded), rh)) {
+			rh->flags |= RSPAMD_HEADER_RECEIVED;
 		}
-
-		rh->flags |= RSPAMD_HEADER_RECEIVED;
 		break;
 	case 0x76F31A09F4352521ULL:	/* to */
 		MESSAGE_FIELD (task, rcpt_mime) = rspamd_email_address_from_mime (task->task_pool,
@@ -93,7 +82,11 @@ rspamd_mime_header_check_special (struct rspamd_task *task,
 
 		rh->flags = RSPAMD_HEADER_MESSAGE_ID|RSPAMD_HEADER_UNIQUE;
 		p = rh->decoded;
-		end = p + strlen (p);
+		len = rspamd_strip_smtp_comments_inplace(rh->decoded, strlen(p));
+		rh->decoded[len] = '\0'; /* Zero terminate after stripping */
+		/* Strip surrounding spaces */
+		rh->decoded = g_strstrip (rh->decoded);
+		end = p + len;
 
 		if (*p == '<') {
 			p++;
@@ -401,7 +394,8 @@ rspamd_mime_headers_process (struct rspamd_task *task,
 			}
 
 			if (nh->decoded == NULL) {
-				nh->decoded = "";
+				/* As we strip comments in place... */
+				nh->decoded = rspamd_mempool_strdup (task->task_pool, "");
 			}
 
 			/* We also validate utf8 and replace all non-valid utf8 chars */
@@ -413,8 +407,8 @@ rspamd_mime_headers_process (struct rspamd_task *task,
 			break;
 		case 5:
 			/* Header has only name, no value */
-			nh->value = "";
-			nh->decoded = "";
+			nh->value = rspamd_mempool_strdup (task->task_pool, "");;
+			nh->decoded = rspamd_mempool_strdup (task->task_pool, "");;
 			nh->raw_len = p - nh->raw_value;
 			nh->order = norder ++;
 			rspamd_mime_header_add (task, &target->htb, order_ptr, nh, check_newlines);
@@ -670,6 +664,14 @@ rspamd_mime_header_decode (rspamd_mempool_t *pool, const gchar *in,
 			if (*p == '?') {
 				state = got_more_qmark;
 				qmarks ++;
+
+				/* Skip multiple ? signs */
+				p ++;
+				while (p < end && *p == '?') {
+					p ++;
+				}
+
+				continue;
 			}
 			p ++;
 			break;
@@ -902,771 +904,18 @@ rspamd_mime_message_id_generate (const gchar *fqdn)
 	return g_string_free (out, FALSE);
 }
 
-enum rspamd_received_part_type {
-	RSPAMD_RECEIVED_PART_FROM,
-	RSPAMD_RECEIVED_PART_BY,
-	RSPAMD_RECEIVED_PART_FOR,
-	RSPAMD_RECEIVED_PART_WITH,
-	RSPAMD_RECEIVED_PART_ID,
-	RSPAMD_RECEIVED_PART_UNKNOWN,
-};
-
-struct rspamd_received_comment {
-	gchar *data;
-	gsize dlen;
-	struct rspamd_received_comment *prev;
-};
-
-struct rspamd_received_part {
-	enum rspamd_received_part_type type;
-	gchar *data;
-	gsize dlen;
-	struct rspamd_received_comment *tail_comment;
-	struct rspamd_received_comment *head_comment;
-	struct rspamd_received_part *prev, *next;
-};
-
-static void
-rspamd_smtp_received_part_set_or_append (struct rspamd_task *task,
-										 const gchar *begin,
-										 gsize len,
-										 gchar **dest,
-										 gsize *destlen)
-{
-	if (len == 0) {
-		return;
-	}
-
-	if (*dest) {
-		/* Append */
-		gsize total_len = *destlen + len;
-		gchar *new_dest;
-
-		new_dest = rspamd_mempool_alloc (task->task_pool, total_len);
-		memcpy (new_dest, *dest, *destlen);
-		memcpy (new_dest + *destlen, begin, len);
-		rspamd_str_lc (new_dest + *destlen, len);
-		*dest = new_dest;
-		*destlen = total_len;
-	}
-	else {
-		/* Set */
-		*dest = rspamd_mempool_alloc (task->task_pool, len);
-		memcpy (*dest, begin, len);
-		rspamd_str_lc (*dest, len);
-		*dest = (gchar *)rspamd_string_len_strip (*dest, &len, " \t");
-		*destlen = len;
-	}
-}
-
-static struct rspamd_received_part *
-rspamd_smtp_received_process_part (struct rspamd_task *task,
-								   const char *data,
-								   size_t len,
-								   enum rspamd_received_part_type type,
-								   goffset *last)
-{
-	struct rspamd_received_part *npart;
-	const guchar *p, *c, *end;
-	guint obraces = 0, ebraces = 0;
-	gboolean seen_tcpinfo = FALSE;
-	enum _parse_state {
-		skip_spaces,
-		in_comment,
-		read_data,
-		read_tcpinfo,
-		all_done
-	} state, next_state;
-
-	npart = rspamd_mempool_alloc0 (task->task_pool, sizeof (*npart));
-	npart->type = type;
-
-	/* In this function, we just process comments and data separately */
-	p = data;
-	end = data + len;
-	c = data;
-	state = skip_spaces;
-	next_state = read_data;
-
-	while (p < end) {
-		switch (state) {
-		case skip_spaces:
-			if (!g_ascii_isspace (*p)) {
-				c = p;
-				state = next_state;
-			}
-			else {
-				p ++;
-			}
-			break;
-		case in_comment:
-			if (*p == '(') {
-				obraces ++;
-			}
-			else if (*p == ')') {
-				ebraces ++;
-
-				if (ebraces >= obraces) {
-					if (type != RSPAMD_RECEIVED_PART_UNKNOWN) {
-						if (p > c) {
-							struct rspamd_received_comment *comment;
-
-
-							comment = rspamd_mempool_alloc0 (task->task_pool,
-									sizeof (*comment));
-							rspamd_smtp_received_part_set_or_append (task,
-									c, p - c,
-									&comment->data, &comment->dlen);
-
-							if (!npart->head_comment) {
-								comment->prev = NULL;
-								npart->head_comment = comment;
-								npart->tail_comment = comment;
-							}
-							else {
-								comment->prev = npart->tail_comment;
-								npart->tail_comment = comment;
-							}
-						}
-					}
-
-					p ++;
-					c = p;
-					state = skip_spaces;
-					next_state = read_data;
-
-					continue;
-				}
-			}
-
-			p ++;
-			break;
-		case read_data:
-			if (*p == '(') {
-				if (p > c) {
-					if (type != RSPAMD_RECEIVED_PART_UNKNOWN) {
-						rspamd_smtp_received_part_set_or_append (task,
-								c, p - c,
-								&npart->data, &npart->dlen);
-					}
-				}
-
-				state = in_comment;
-				obraces = 1;
-				ebraces = 0;
-				p ++;
-				c = p;
-			}
-			else if (g_ascii_isspace (*p)) {
-				if (p > c) {
-					if (type != RSPAMD_RECEIVED_PART_UNKNOWN) {
-						rspamd_smtp_received_part_set_or_append (task,
-								c, p - c,
-								&npart->data, &npart->dlen);
-					}
-				}
-
-				state = skip_spaces;
-				next_state = read_data;
-				c = p;
-			}
-			else if (*p == ';') {
-				/* It is actually delimiter of date part if not in the comments */
-				if (p > c) {
-					if (type != RSPAMD_RECEIVED_PART_UNKNOWN) {
-						rspamd_smtp_received_part_set_or_append (task,
-								c, p - c,
-								&npart->data, &npart->dlen);
-					}
-				}
-
-				state = all_done;
-				continue;
-			}
-			else if (npart->dlen > 0) {
-				/* We have already received data and find something with no ( */
-				if (!seen_tcpinfo && type == RSPAMD_RECEIVED_PART_FROM) {
-					/* Check if we have something special here, such as TCPinfo */
-					if (*c == '[') {
-						state = read_tcpinfo;
-						p ++;
-					}
-					else {
-						state = all_done;
-						continue;
-					}
-				}
-				else {
-					state = all_done;
-					continue;
-				}
-			}
-			else {
-				p ++;
-			}
-			break;
-		case read_tcpinfo:
-			if (*p == ']') {
-				rspamd_smtp_received_part_set_or_append (task,
-						c, p - c + 1,
-						&npart->data, &npart->dlen);
-				seen_tcpinfo = TRUE;
-				state = skip_spaces;
-				next_state = read_data;
-				c = p;
-			}
-			p ++;
-			break;
-		case all_done:
-			if (p > (const guchar *)data) {
-				*last = p - (const guchar *) data;
-				return npart;
-			}
-			else {
-				/* Empty element */
-				return NULL;
-			}
-			break;
-		}
-	}
-
-	/* Leftover */
-	switch (state) {
-	case read_data:
-		if (p > c) {
-			if (type != RSPAMD_RECEIVED_PART_UNKNOWN) {
-				rspamd_smtp_received_part_set_or_append (task,
-						c, p - c,
-						&npart->data, &npart->dlen);
-			}
-
-			*last = p - (const guchar *)data;
-
-			return npart;
-		}
-		break;
-	case skip_spaces:
-		if (p > (const guchar *)data) {
-			*last = p - (const guchar *) data;
-
-			return npart;
-		}
-	default:
-		break;
-	}
-
-	return NULL;
-}
-
-static struct rspamd_received_part *
-rspamd_smtp_received_spill (struct rspamd_task *task,
-							const char *data,
-							size_t len,
-							goffset *date_pos)
-{
-	const guchar *p, *end;
-	struct rspamd_received_part *cur_part, *head = NULL;
-	goffset pos = 0;
-
-	p = data;
-	end = data + len;
-
-	while (p < end && g_ascii_isspace (*p)) {
-		p ++;
-	}
-
-	len = end - p;
-
-	/* Ignore all received but those started from from part */
-	if (len <= 4 || (lc_map[p[0]] != 'f' &&
-					 lc_map[p[1]] != 'r' &&
-					 lc_map[p[2]] != 'o' &&
-					 lc_map[p[3]] != 'm')) {
-		return NULL;
-	}
-
-	p += sizeof ("from") - 1;
-
-	/* We can now store from part */
-	cur_part = rspamd_smtp_received_process_part (task, p, end - p,
-			RSPAMD_RECEIVED_PART_FROM, &pos);
-
-	if (!cur_part) {
-		return NULL;
-	}
-
-	g_assert (pos != 0);
-	p += pos;
-	len = end > p ? end - p : 0;
-	DL_APPEND (head, cur_part);
-
-	if (len > 2 && (lc_map[p[0]] == 'b' &&
-					lc_map[p[1]] == 'y')) {
-		p += sizeof ("by") - 1;
-
-		cur_part = rspamd_smtp_received_process_part (task, p, end - p,
-				RSPAMD_RECEIVED_PART_BY, &pos);
-
-		if (!cur_part) {
-			return NULL;
-		}
-
-		g_assert (pos != 0);
-		p += pos;
-		len = end > p ? end - p : 0;
-		DL_APPEND (head, cur_part);
-	}
-
-	while (p < end) {
-		if (*p == ';') {
-			/* We are at the date separator, stop here */
-			*date_pos = p - (const guchar *)data + 1;
-			break;
-		}
-		else {
-			if (len > sizeof ("with") && (lc_map[p[0]] == 'w' &&
-										  lc_map[p[1]] == 'i' &&
-										  lc_map[p[2]] == 't' &&
-										  lc_map[p[3]] == 'h')) {
-				p += sizeof ("with") - 1;
-
-				cur_part = rspamd_smtp_received_process_part (task, p, end - p,
-						RSPAMD_RECEIVED_PART_WITH, &pos);
-			}
-			else if (len > sizeof ("for") && (lc_map[p[0]] == 'f' &&
-											  lc_map[p[1]] == 'o' &&
-											  lc_map[p[2]] == 'r')) {
-				p += sizeof ("for") - 1;
-				cur_part = rspamd_smtp_received_process_part (task, p, end - p,
-						RSPAMD_RECEIVED_PART_FOR, &pos);
-			}
-			else if (len > sizeof ("id") && (lc_map[p[0]] == 'i' &&
-											  lc_map[p[1]] == 'd')) {
-				p += sizeof ("id") - 1;
-				cur_part = rspamd_smtp_received_process_part (task, p, end - p,
-						RSPAMD_RECEIVED_PART_ID, &pos);
-			}
-			else {
-				while (p < end) {
-					if (!(g_ascii_isspace (*p) || *p == '(' || *p == ';')) {
-						p ++;
-					}
-					else {
-						break;
-					}
-				}
-
-				if (p == end) {
-					return NULL;
-				}
-				else if (*p == ';') {
-					*date_pos = p - (const guchar *)data + 1;
-					break;
-				}
-				else {
-					cur_part = rspamd_smtp_received_process_part (task, p, end - p,
-							RSPAMD_RECEIVED_PART_UNKNOWN, &pos);
-				}
-			}
-
-			if (!cur_part) {
-				p ++;
-				len = end > p ? end - p : 0;
-			}
-			else {
-				g_assert (pos != 0);
-				p += pos;
-				len = end > p ? end - p : 0;
-				DL_APPEND (head, cur_part);
-			}
-		}
-	}
-
-	return head;
-}
-
-static gboolean
-rspamd_smtp_received_process_rdns (struct rspamd_task *task,
-								   const gchar *begin,
-								   gsize len,
-								   const gchar **pdest)
-{
-	const gchar *p, *end;
-	gsize hlen = 0;
-	gboolean seen_dot = FALSE;
-
-	p = begin;
-	end = begin + len;
-
-	while (p < end) {
-		if (!g_ascii_isspace (*p) && rspamd_url_is_domain (*p)) {
-			if (*p == '.') {
-				seen_dot = TRUE;
-			}
-
-			hlen ++;
-		}
-		else {
-			break;
-		}
-
-		p ++;
-	}
-
-	if (hlen > 0) {
-		if (p == end) {
-			/* All data looks like a hostname */
-			gchar *dest;
-
-			dest = rspamd_mempool_alloc (task->task_pool,
-					hlen + 1);
-			rspamd_strlcpy (dest, begin, hlen + 1);
-			*pdest = dest;
-
-			return TRUE;
-		}
-		else if (seen_dot && (g_ascii_isspace (*p) || *p == '[' || *p == '(')) {
-			gchar *dest;
-
-			dest = rspamd_mempool_alloc (task->task_pool,
-					hlen + 1);
-			rspamd_strlcpy (dest, begin, hlen + 1);
-			*pdest = dest;
-
-			return TRUE;
-		}
-	}
-
-	return FALSE;
-}
-
-static gboolean
-rspamd_smtp_received_process_host_tcpinfo (struct rspamd_task *task,
-										   struct rspamd_received_header *rh,
-										   const gchar *data,
-										   gsize len)
-{
-	rspamd_inet_addr_t *addr = NULL;
-	gboolean ret = FALSE;
-
-	if (data[0] == '[') {
-		/* Likely Exim version */
-
-		const gchar *brace_pos = memchr (data, ']', len);
-
-		if (brace_pos) {
-			addr = rspamd_parse_inet_address_pool (data + 1,
-					brace_pos - data - 1,
-					task->task_pool,
-					RSPAMD_INET_ADDRESS_PARSE_RECEIVED);
-
-			if (addr) {
-				rh->addr = addr;
-				rh->real_ip = rspamd_mempool_strdup (task->task_pool,
-						rspamd_inet_address_to_string (addr));
-				rh->from_ip = rh->real_ip;
-			}
-		}
-	}
-	else {
-		if (g_ascii_isxdigit (data[0])) {
-			/* Try to parse IP address */
-			addr = rspamd_parse_inet_address_pool (data,
-					len, task->task_pool, RSPAMD_INET_ADDRESS_PARSE_RECEIVED);
-			if (addr) {
-				rh->addr = addr;
-				rh->real_ip = rspamd_mempool_strdup (task->task_pool,
-						rspamd_inet_address_to_string (addr));
-				rh->from_ip = rh->real_ip;
-			}
-		}
-
-		if (!addr) {
-			/* Try canonical Postfix version: rdns [ip] */
-			const gchar *obrace_pos = memchr (data, '[', len),
-					*ebrace_pos, *dend;
-
-			if (obrace_pos) {
-				dend = data + len;
-				ebrace_pos = memchr (obrace_pos, ']', dend - obrace_pos);
-
-				if (ebrace_pos) {
-					addr = rspamd_parse_inet_address_pool (obrace_pos + 1,
-							ebrace_pos - obrace_pos - 1,
-							task->task_pool,
-							RSPAMD_INET_ADDRESS_PARSE_RECEIVED);
-
-					if (addr) {
-						rh->addr = addr;
-						rh->real_ip = rspamd_mempool_strdup (task->task_pool,
-								rspamd_inet_address_to_string (addr));
-						rh->from_ip = rh->real_ip;
-
-						/* Process with rDNS */
-						if (rspamd_smtp_received_process_rdns (task,
-								data,
-								obrace_pos - data,
-								&rh->real_hostname)) {
-							ret = TRUE;
-						}
-					}
-				}
-			}
-			else {
-				/* Hostname or some crap, sigh... */
-				if (rspamd_smtp_received_process_rdns (task,
-						data,
-						len,
-						&rh->real_hostname)) {
-					ret = TRUE;
-				}
-			}
-		}
-	}
-
-	return ret;
-}
-
-static void
-rspamd_smtp_received_process_from (struct rspamd_task *task,
-								   struct rspamd_received_part *rpart,
-								   struct rspamd_received_header *rh)
-{
-	if (rpart->dlen > 0) {
-		/* We have seen multiple cases:
-		 * - [ip] (hostname/unknown [real_ip])
-		 * - helo (hostname/unknown [real_ip])
-		 * - [ip]
-		 * - hostname
-		 * - hostname ([ip]:port helo=xxx)
-		 * Maybe more...
-		 */
-		gboolean seen_ip_in_data = FALSE;
-
-		if (rpart->head_comment && rpart->head_comment->dlen > 0) {
-			/* We can have info within comment as part of RFC */
-			rspamd_smtp_received_process_host_tcpinfo (
-					task, rh,
-					rpart->head_comment->data, rpart->head_comment->dlen);
-		}
-
-		if (!rh->real_ip) {
-			if (rpart->data[0] == '[') {
-				/* No comment, just something that looks like SMTP IP */
-				const gchar *brace_pos = memchr (rpart->data, ']', rpart->dlen);
-				rspamd_inet_addr_t *addr;
-
-				if (brace_pos) {
-					addr = rspamd_parse_inet_address_pool (rpart->data + 1,
-							brace_pos - rpart->data - 1,
-							task->task_pool,
-							RSPAMD_INET_ADDRESS_PARSE_RECEIVED);
-
-					if (addr) {
-						seen_ip_in_data = TRUE;
-						rh->addr = addr;
-						rh->real_ip = rspamd_mempool_strdup (task->task_pool,
-								rspamd_inet_address_to_string (addr));
-						rh->from_ip = rh->real_ip;
-					}
-				}
-			} else if (g_ascii_isxdigit (rpart->data[0])) {
-				/* Try to parse IP address */
-				rspamd_inet_addr_t *addr;
-				addr = rspamd_parse_inet_address_pool (rpart->data,
-						rpart->dlen, task->task_pool,
-						RSPAMD_INET_ADDRESS_PARSE_RECEIVED);
-				if (addr) {
-					seen_ip_in_data = TRUE;
-					rh->addr = addr;
-					rh->real_ip = rspamd_mempool_strdup (task->task_pool,
-							rspamd_inet_address_to_string (addr));
-					rh->from_ip = rh->real_ip;
-				}
-			}
-		}
-
-		if (!seen_ip_in_data) {
-			if (rh->real_ip) {
-				/* Get anounced hostname (usually helo) */
-				rspamd_smtp_received_process_rdns (task,
-						rpart->data,
-						rpart->dlen,
-						&rh->from_hostname);
-			}
-			else {
-				rspamd_smtp_received_process_host_tcpinfo (task,
-						rh, rpart->data, rpart->dlen);
-			}
-		}
-	}
-	else {
-		/* rpart->dlen = 0 */
-
-		if (rpart->head_comment && rpart->head_comment->dlen > 0) {
-			rspamd_smtp_received_process_host_tcpinfo (task,
-					rh,
-					rpart->head_comment->data,
-					rpart->head_comment->dlen);
-		}
-	}
-}
-
-int
-rspamd_smtp_received_parse (struct rspamd_task *task,
-							const char *data,
-							size_t len,
-							struct rspamd_received_header *rh)
-{
-	goffset date_pos = -1;
-	struct rspamd_received_part *head, *cur;
-	rspamd_ftok_t t1, t2;
-
-	head = rspamd_smtp_received_spill (task, data, len, &date_pos);
-
-	if (head == NULL) {
-		return -1;
-	}
-
-	rh->flags = RSPAMD_RECEIVED_UNKNOWN;
-
-	DL_FOREACH (head, cur) {
-		switch (cur->type) {
-		case RSPAMD_RECEIVED_PART_FROM:
-			rspamd_smtp_received_process_from (task, cur, rh);
-			break;
-		case RSPAMD_RECEIVED_PART_BY:
-			rspamd_smtp_received_process_rdns (task,
-					cur->data,
-					cur->dlen,
-					&rh->by_hostname);
-			break;
-		case RSPAMD_RECEIVED_PART_WITH:
-			t1.begin = cur->data;
-			t1.len = cur->dlen;
-
-			if (t1.len > 0) {
-				RSPAMD_FTOK_ASSIGN (&t2, "smtp");
-
-				if (rspamd_ftok_cmp (&t1, &t2) == 0) {
-					rh->flags = RSPAMD_RECEIVED_SMTP;
-				}
-
-				RSPAMD_FTOK_ASSIGN (&t2, "esmtp");
-
-				if (rspamd_ftok_starts_with (&t1, &t2)) {
-					/*
-					 * esmtp, esmtps, esmtpsa
-					 */
-					if (t1.len == t2.len + 1) {
-						if (t1.begin[t2.len] == 'a') {
-							rh->flags = RSPAMD_RECEIVED_ESMTPA;
-							rh->flags |= RSPAMD_RECEIVED_FLAG_AUTHENTICATED;
-						}
-						else if (t1.begin[t2.len] == 's') {
-							rh->flags = RSPAMD_RECEIVED_ESMTPS;
-							rh->flags |= RSPAMD_RECEIVED_FLAG_SSL;
-						}
-						continue;
-					}
-					else if (t1.len == t2.len + 2) {
-						if (t1.begin[t2.len] == 's' &&
-								t1.begin[t2.len + 1] == 'a') {
-							rh->flags = RSPAMD_RECEIVED_ESMTPSA;
-							rh->flags |= RSPAMD_RECEIVED_FLAG_AUTHENTICATED;
-							rh->flags |= RSPAMD_RECEIVED_FLAG_SSL;
-						}
-						continue;
-					}
-					else if (t1.len == t2.len) {
-						rh->flags = RSPAMD_RECEIVED_ESMTP;
-						continue;
-					}
-				}
-
-				RSPAMD_FTOK_ASSIGN (&t2, "lmtp");
-
-				if (rspamd_ftok_cmp (&t1, &t2) == 0) {
-					rh->flags = RSPAMD_RECEIVED_LMTP;
-					continue;
-				}
-
-				RSPAMD_FTOK_ASSIGN (&t2, "imap");
-
-				if (rspamd_ftok_cmp (&t1, &t2) == 0) {
-					rh->flags = RSPAMD_RECEIVED_IMAP;
-					continue;
-				}
-
-				RSPAMD_FTOK_ASSIGN (&t2, "local");
-
-				if (rspamd_ftok_cmp (&t1, &t2) == 0) {
-					rh->flags = RSPAMD_RECEIVED_LOCAL;
-					continue;
-				}
-
-				RSPAMD_FTOK_ASSIGN (&t2, "http");
-
-				if (rspamd_ftok_starts_with (&t1, &t2)) {
-					if (t1.len == t2.len + 1) {
-						if (t1.begin[t2.len] == 's') {
-							rh->flags = RSPAMD_RECEIVED_HTTP;
-							rh->flags |= RSPAMD_RECEIVED_FLAG_SSL;
-						}
-					}
-					else if (t1.len == t2.len) {
-						rh->flags = RSPAMD_RECEIVED_HTTP;
-					}
-
-					continue;
-				}
-			}
-
-			break;
-		case RSPAMD_RECEIVED_PART_FOR:
-			rh->for_addr = rspamd_email_address_from_smtp (cur->data, cur->dlen);
-
-			if (rh->for_addr) {
-				if (rh->for_addr->addr_len > 0) {
-					t1.begin = rh->for_addr->addr;
-					t1.len = rh->for_addr->addr_len;
-					rh->for_mbox = rspamd_mempool_ftokdup (task->task_pool,
-							&t1);
-				}
-
-				rspamd_mempool_add_destructor (task->task_pool,
-						(rspamd_mempool_destruct_t)rspamd_email_address_free,
-						rh->for_addr);
-			}
-			break;
-		default:
-			/* Do nothing */
-			break;
-		}
-	}
-
-	if (rh->real_ip && !rh->from_ip) {
-		rh->from_ip = rh->real_ip;
-	}
-
-	if (rh->real_hostname && !rh->from_hostname) {
-		rh->from_hostname = rh->real_hostname;
-	}
-
-	if (date_pos > 0 && date_pos < len) {
-		rh->timestamp = rspamd_parse_smtp_date (data + date_pos,
-				len - date_pos, NULL);
-	}
-
-	return 0;
-}
-
 struct rspamd_mime_header *
-rspamd_message_get_header_from_hash (struct rspamd_mime_headers_table *headers,
-									 const gchar *field)
+rspamd_message_get_header_from_hash (struct rspamd_mime_headers_table *hdrs,
+									 const gchar *field,
+									 gboolean need_modified)
 {
+	if (hdrs == NULL) {
+		return NULL;
+	}
+
 	khiter_t k;
-	khash_t(rspamd_mime_headers_htb) *htb = &headers->htb;
+	khash_t(rspamd_mime_headers_htb) *htb = &hdrs->htb;
+	struct rspamd_mime_header *hdr;
 
 	if (htb) {
 		k = kh_get (rspamd_mime_headers_htb, htb, (gchar *) field);
@@ -1675,19 +924,60 @@ rspamd_message_get_header_from_hash (struct rspamd_mime_headers_table *headers,
 			return NULL;
 		}
 
-		return kh_value (htb, k);
+		hdr = kh_value (htb, k);
+
+		if (!need_modified) {
+			if (hdr->flags & RSPAMD_HEADER_NON_EXISTING) {
+				return NULL;
+			}
+
+			return hdr;
+		}
+		else {
+			if (hdr->flags & RSPAMD_HEADER_MODIFIED) {
+				return hdr->modified_chain;
+			}
+
+			return hdr;
+		}
 	}
 
 	return NULL;
 }
 
 struct rspamd_mime_header *
-rspamd_message_get_header_array (struct rspamd_task *task,
-								 const gchar *field)
+rspamd_message_get_header_array (struct rspamd_task *task, const gchar *field,
+		gboolean need_modified)
 {
-	return rspamd_message_get_header_from_hash (
+	return rspamd_message_get_header_from_hash(
 			MESSAGE_FIELD_CHECK (task, raw_headers),
-			field);
+			field, need_modified);
+}
+
+gsize
+rspamd_mime_headers_count (struct rspamd_mime_headers_table *hdrs)
+{
+	if (hdrs) {
+		return kh_size (&hdrs->htb);
+	}
+
+	return 0;
+}
+
+bool
+rspamd_mime_headers_foreach(const struct rspamd_mime_headers_table *hdrs,
+								 rspamd_hdr_traverse_func_t func, void *ud)
+{
+	const gchar *name;
+	struct rspamd_mime_header *hdr;
+
+	kh_foreach(&hdrs->htb, name, hdr, {
+		if (!func(name, hdr, ud)) {
+			return false;
+		}
+	});
+
+	return true;
 }
 
 static void
@@ -1724,4 +1014,309 @@ rspamd_message_headers_new (void)
 	REF_INIT_RETAIN (nhdrs, rspamd_message_headers_dtor);
 
 	return nhdrs;
+}
+
+void
+rspamd_message_set_modified_header (struct rspamd_task *task,
+									struct rspamd_mime_headers_table *hdrs,
+									const gchar *hdr_name,
+									const ucl_object_t *obj)
+{
+	khiter_t k;
+	khash_t(rspamd_mime_headers_htb) *htb = &hdrs->htb;
+	struct rspamd_mime_header *hdr_elt, *existing_chain;
+	int i;
+
+	if (htb) {
+		k = kh_get (rspamd_mime_headers_htb, htb, (gchar *)hdr_name);
+
+		if (k == kh_end (htb)) {
+			hdr_elt = rspamd_mempool_alloc0 (task->task_pool, sizeof (*hdr_elt));
+
+			hdr_elt->flags |= RSPAMD_HEADER_MODIFIED|RSPAMD_HEADER_NON_EXISTING;
+			hdr_elt->name = rspamd_mempool_strdup (task->task_pool, hdr_name);
+
+			int r;
+			k = kh_put (rspamd_mime_headers_htb, htb, hdr_elt->name, &r);
+
+			kh_value (htb, k) = hdr_elt;
+		}
+		else {
+			hdr_elt = kh_value (htb, k);
+		}
+	}
+	else {
+		/* No hash, no modification */
+		msg_err_task ("internal error: calling for set_modified_header for no headers");
+		return;
+	}
+
+	if (hdr_elt->flags & RSPAMD_HEADER_MODIFIED) {
+		existing_chain = hdr_elt->modified_chain;
+	}
+	else {
+		existing_chain = hdr_elt;
+	}
+
+	const ucl_object_t *elt, *cur;
+	ucl_object_iter_t it;
+
+	/* First, deal with removed headers, copying the relevant headers with remove flag */
+	elt = ucl_object_lookup (obj, "remove");
+
+	/*
+	 * remove:  {1, 2 ...}
+	 * where number is the header's position starting from '1'
+	 */
+	if (elt && ucl_object_type (elt) == UCL_ARRAY) {
+		/* First, use a temporary array to keep all headers */
+		GPtrArray *existing_ar = g_ptr_array_new ();
+		struct rspamd_mime_header *cur_hdr;
+
+		/* Exclude removed headers */
+		LL_FOREACH (existing_chain, cur_hdr) {
+			if (!(cur_hdr->flags & RSPAMD_HEADER_REMOVED)) {
+				g_ptr_array_add (existing_ar, cur_hdr);
+			}
+		}
+
+		it = NULL;
+
+		while ((cur = ucl_object_iterate (elt, &it, true)) != NULL) {
+			if (ucl_object_type (cur) == UCL_INT) {
+				int ord = ucl_object_toint (cur);
+
+				if (ord == 0) {
+					/* Remove all headers in the existing chain */
+					PTR_ARRAY_FOREACH (existing_ar, i, cur_hdr) {
+						cur_hdr->flags |= RSPAMD_HEADER_MODIFIED|RSPAMD_HEADER_REMOVED;
+					}
+				}
+				else if (ord > 0) {
+					/* Start from the top */
+
+					if (ord <= existing_ar->len) {
+						cur_hdr = g_ptr_array_index (existing_ar, ord - 1);
+						cur_hdr->flags |= RSPAMD_HEADER_MODIFIED|RSPAMD_HEADER_REMOVED;
+					}
+				}
+				else {
+					/* Start from the bottom; ord < 0 */
+					if ((-ord) <= existing_ar->len) {
+						cur_hdr = g_ptr_array_index (existing_ar, existing_ar->len + ord);
+						cur_hdr->flags |= RSPAMD_HEADER_MODIFIED|RSPAMD_HEADER_REMOVED;
+					}
+				}
+			}
+		}
+
+		/*
+		 * Next, we return all headers modified to the existing chain
+		 * This implies an additional copy of all structures but is safe enough to
+		 * deal with it
+		 */
+		hdr_elt->flags |= RSPAMD_HEADER_MODIFIED;
+		hdr_elt->modified_chain = NULL;
+		gint new_chain_length = 0;
+
+		PTR_ARRAY_FOREACH (existing_ar, i, cur_hdr) {
+			if (!(cur_hdr->flags & RSPAMD_HEADER_REMOVED)) {
+				struct rspamd_mime_header *nhdr = rspamd_mempool_alloc (
+						task->task_pool, sizeof (*nhdr));
+				memcpy (nhdr, cur_hdr, sizeof (*nhdr));
+				nhdr->modified_chain = NULL;
+				nhdr->prev = NULL;
+				nhdr->next = NULL;
+				nhdr->ord_next = NULL;
+
+				DL_APPEND (hdr_elt->modified_chain, nhdr);
+				new_chain_length ++;
+			}
+		}
+
+		g_ptr_array_free (existing_ar, TRUE);
+
+		/* End of headers removal logic */
+	}
+
+	/* We can now deal with headers additions */
+	elt = ucl_object_lookup (obj, "add");
+	if (elt && ucl_object_type (elt) == UCL_ARRAY) {
+		if (!(hdr_elt->flags & RSPAMD_HEADER_MODIFIED)) {
+			/* Copy the header itself to the modified chain */
+			struct rspamd_mime_header *nhdr;
+			hdr_elt->flags |= RSPAMD_HEADER_MODIFIED;
+			nhdr = rspamd_mempool_alloc (
+					task->task_pool, sizeof (*nhdr));
+			memcpy (nhdr, hdr_elt, sizeof (*hdr_elt));
+			nhdr->modified_chain = NULL;
+			nhdr->next = NULL;
+			nhdr->ord_next = NULL;
+			nhdr->prev = nhdr;
+			hdr_elt->modified_chain = nhdr;
+		}
+
+		/*
+		 * add:  {{1, "foo"}, {-1, "bar"} ...}
+		 * where number is the header's position starting from '1'
+		 */
+		it = NULL;
+
+		while ((cur = ucl_object_iterate (elt, &it, true)) != NULL) {
+			if (ucl_object_type (cur) == UCL_ARRAY) {
+				const ucl_object_t *order = ucl_array_find_index (cur, 0),
+					*value = ucl_array_find_index (cur, 1);
+
+				if (order && value &&
+					(ucl_object_type (order) == UCL_INT &&
+					 ucl_object_type (value) == UCL_STRING)) {
+					int ord = ucl_object_toint (order);
+					const char *raw_value;
+					gsize raw_len;
+
+					raw_value = ucl_object_tolstring (value, &raw_len);
+
+					if (raw_len == 0) {
+						continue;
+					}
+
+					struct rspamd_mime_header *nhdr = rspamd_mempool_alloc0 (
+							task->task_pool, sizeof (*nhdr));
+
+					nhdr->flags |= RSPAMD_HEADER_ADDED;
+					nhdr->name = hdr_elt->name;
+					nhdr->value = rspamd_mempool_alloc (task->task_pool,
+							raw_len + 1);
+					nhdr->raw_len = rspamd_strlcpy (nhdr->value, raw_value,
+							raw_len + 1);
+					nhdr->raw_value = nhdr->value;
+					nhdr->decoded = rspamd_mime_header_decode (task->task_pool,
+							raw_value, raw_len, NULL);
+
+					/* Now find a position to insert a value */
+					struct rspamd_mime_header **pos = &hdr_elt->modified_chain;
+
+					if (ord == 0) {
+						DL_PREPEND (hdr_elt->modified_chain, nhdr);
+					}
+					else if (ord == -1) {
+						DL_APPEND (hdr_elt->modified_chain, nhdr);
+					}
+					else if (ord > 0) {
+						while (ord > 0 && (*pos)) {
+							ord --;
+							pos = &((*pos)->next);
+						}
+						if (*pos) {
+							/* pos is &(elt)->next */
+							nhdr->next = (*pos);
+							nhdr->prev = (*pos)->prev;
+							(*pos)->prev = nhdr;
+							*pos = nhdr;
+						}
+						else {
+							/* Last element */
+							DL_APPEND (*pos, nhdr);
+						}
+					}
+					else {
+						/* NYI: negative order is not defined */
+						msg_err_task ("internal error: calling for set_modified_header "
+									  "with negative add order header");
+					}
+				}
+				else {
+					msg_err_task ("internal error: calling for set_modified_header "
+								  "with invalid header");
+				}
+			}
+		}
+	}
+}
+
+gsize
+rspamd_strip_smtp_comments_inplace (gchar *input, gsize len)
+{
+	enum parser_state {
+		parse_normal,
+		parse_obrace,
+		parse_comment,
+		parse_quoted_copy,
+		parse_quoted_ignore,
+	} state = parse_normal, next_state = parse_normal;
+	gchar *d = input, *end = input + len, *start = input;
+	gchar t;
+	int obraces = 0, ebraces = 0;
+
+	while (input < end) {
+		t = *input;
+		switch (state) {
+		case parse_normal:
+			if (t == '(') {
+				state = parse_obrace;
+			}
+			else if (t == '\\') {
+				state = parse_quoted_copy;
+				next_state = parse_normal;
+			}
+			else {
+				*d++ = t;
+			}
+			input ++;
+			break;
+		case parse_obrace:
+			obraces ++;
+			if (t == '(') {
+				obraces ++;
+			}
+			else if (t == ')') {
+				ebraces ++;
+
+				if (obraces == ebraces) {
+					obraces = 0;
+					ebraces = 0;
+					state = parse_normal;
+				}
+			}
+			else if (t == '\\') {
+				state = parse_quoted_ignore;
+				next_state = parse_comment;
+			}
+			else {
+				state = parse_comment;
+			}
+			input ++;
+			break;
+		case parse_comment:
+			if (t == '(') {
+				state = parse_obrace;
+			}
+			else if (t == ')') {
+				ebraces ++;
+
+				if (obraces == ebraces) {
+					obraces = 0;
+					ebraces = 0;
+					state = parse_normal;
+				}
+			}
+			else if (t == '\\') {
+				state = parse_quoted_ignore;
+				next_state = parse_comment;
+			}
+			input ++;
+			break;
+		case parse_quoted_copy:
+			*d++ = t;
+			state = next_state;
+			input ++;
+			break;
+		case parse_quoted_ignore:
+			state = next_state;
+			input ++;
+			break;
+		}
+	}
+
+	return (d - start);
 }
\ No newline at end of file
