@@ -25,6 +25,7 @@
 #include "contrib/uthash/utlist.h"
 #include <openssl/cms.h>
 #include <openssl/pkcs7.h>
+#include "contrib/fastutf8/fastutf8.h"
 
 struct rspamd_mime_parser_lib_ctx {
 	struct rspamd_multipattern *mp_boundary;
@@ -39,7 +40,7 @@ static const guint max_key_usages = 10000;
 
 #define msg_debug_mime(...)  rspamd_conditional_debug_fast (NULL, task->from_addr, \
         rspamd_mime_log_id, "mime", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(mime)
@@ -392,7 +393,7 @@ rspamd_mime_part_get_cte (struct rspamd_task *task,
 	enum rspamd_cte cte = RSPAMD_CTE_UNKNOWN;
 	gboolean parent_propagated = FALSE;
 
-	hdr = rspamd_message_get_header_from_hash (hdrs, "Content-Transfer-Encoding");
+	hdr = rspamd_message_get_header_from_hash(hdrs, "Content-Transfer-Encoding", FALSE);
 
 	if (hdr == NULL) {
 		if (part->parent_part && part->parent_part->cte != RSPAMD_CTE_UNKNOWN &&
@@ -471,15 +472,15 @@ rspamd_mime_part_get_cd (struct rspamd_task *task, struct rspamd_mime_part *part
 	rspamd_ftok_t srch;
 	struct rspamd_content_type_param *found;
 
-	hdr = rspamd_message_get_header_from_hash (part->raw_headers,
-			"Content-Disposition");
+	hdr = rspamd_message_get_header_from_hash(part->raw_headers,
+			"Content-Disposition", FALSE);
 
 
 	if (hdr == NULL) {
 		cd = rspamd_mempool_alloc0 (task->task_pool, sizeof (*cd));
 		cd->type = RSPAMD_CT_INLINE;
 
-		/* We can also have content dispositon definitions in Content-Type */
+		/* We can also have content disposition definitions in Content-Type */
 		if (part->ct && part->ct->attrs) {
 			RSPAMD_FTOK_ASSIGN (&srch, "name");
 			found = g_hash_table_lookup (part->ct->attrs, &srch);
@@ -500,9 +501,9 @@ rspamd_mime_part_get_cd (struct rspamd_task *task, struct rspamd_mime_part *part
 			gsize hlen;
 			cd = NULL;
 
-			if (cur->decoded) {
-				hlen = strlen (cur->decoded);
-				cd = rspamd_content_disposition_parse (cur->decoded, hlen,
+			if (cur->value) {
+				hlen = strlen (cur->value);
+				cd = rspamd_content_disposition_parse (cur->value, hlen,
 						task->task_pool);
 			}
 
@@ -538,7 +539,7 @@ rspamd_mime_part_get_cd (struct rspamd_task *task, struct rspamd_mime_part *part
 				cd = rspamd_mempool_alloc0 (task->task_pool, sizeof (*cd));
 				cd->type = RSPAMD_CT_INLINE;
 
-				/* We can also have content dispositon definitions in Content-Type */
+				/* We can also have content disposition definitions in Content-Type */
 				if (part->ct->attrs) {
 					RSPAMD_FTOK_ASSIGN (&srch, "name");
 					found = g_hash_table_lookup (part->ct->attrs, &srch);
@@ -595,14 +596,14 @@ rspamd_mime_parse_normal_part (struct rspamd_task *task,
 	g_assert (part != NULL);
 
 	rspamd_mime_part_get_cte (task, part->raw_headers, part,
-			!(part->ct->flags & RSPAMD_CONTENT_TYPE_MESSAGE));
+			part->ct && !(part->ct->flags & RSPAMD_CONTENT_TYPE_MESSAGE));
 	rspamd_mime_part_get_cd (task, part);
 
 	switch (part->cte) {
 	case RSPAMD_CTE_7BIT:
 	case RSPAMD_CTE_8BIT:
 	case RSPAMD_CTE_UNKNOWN:
-		if (part->ct->flags & RSPAMD_CONTENT_TYPE_MISSING) {
+		if (part->ct && (part->ct->flags & RSPAMD_CONTENT_TYPE_MISSING)) {
 			if (part->cte != RSPAMD_CTE_7BIT) {
 				/* We have something that has a missing content-type,
 				 * but it has non-7bit characters.
@@ -610,8 +611,20 @@ rspamd_mime_parse_normal_part (struct rspamd_task *task,
 				 * In theory, it is very unsafe to process it as a text part
 				 * as we unlikely get some sane result
 				 */
-				part->ct->flags &= ~RSPAMD_CONTENT_TYPE_TEXT;
-				part->ct->flags |= RSPAMD_CONTENT_TYPE_BROKEN;
+
+				/*
+				 * On the other hand, there is an evidence that some
+				 * emails actually rely on that.
+				 * So we apply an expensive hack here:
+				 * if there are no 8bit characters -OR- the content is valid
+				 * UTF8, we can still imply Content-Type == text/plain
+				 */
+
+				if (rspamd_str_has_8bit (part->raw_data.begin, part->raw_data.len) &&
+					!rspamd_fast_utf8_validate (part->raw_data.begin, part->raw_data.len)) {
+					part->ct->flags &= ~RSPAMD_CONTENT_TYPE_TEXT;
+					part->ct->flags |= RSPAMD_CONTENT_TYPE_BROKEN;
+				}
 			}
 		}
 
@@ -645,7 +658,9 @@ rspamd_mime_parse_normal_part (struct rspamd_task *task,
 		}
 		else {
 			msg_err_task ("invalid quoted-printable encoded part, assume 8bit");
-			part->ct->flags |= RSPAMD_CONTENT_TYPE_BROKEN;
+			if (part->ct) {
+				part->ct->flags |= RSPAMD_CONTENT_TYPE_BROKEN;
+			}
 			part->cte = RSPAMD_CTE_8BIT;
 			memcpy (parsed->str, part->raw_data.begin, part->raw_data.len);
 			parsed->len = part->raw_data.len;
@@ -681,7 +696,9 @@ rspamd_mime_parse_normal_part (struct rspamd_task *task,
 		}
 		else {
 			msg_err_task ("invalid uuencoding in encoded part, assume 8bit");
-			part->ct->flags |= RSPAMD_CONTENT_TYPE_BROKEN;
+			if (part->ct) {
+				part->ct->flags |= RSPAMD_CONTENT_TYPE_BROKEN;
+			}
 			part->cte = RSPAMD_CTE_8BIT;
 			parsed->len = MIN (part->raw_data.len, parsed->allocated);
 			memcpy (parsed->str, part->raw_data.begin, parsed->len);
@@ -726,7 +743,7 @@ rspamd_mime_parse_normal_part (struct rspamd_task *task,
 
 						ct_nid =  OBJ_obj2nid (p7_signed_content->type);
 
-						if (ct_nid == NID_pkcs7_data) {
+						if (ct_nid == NID_pkcs7_data && p7_signed_content->d.data) {
 							int ret;
 
 							msg_debug_mime ("found an additional part inside of "
@@ -859,8 +876,8 @@ rspamd_mime_process_multipart_node (struct rspamd_task *task,
 			}
 		}
 
-		hdr = rspamd_message_get_header_from_hash (npart->raw_headers,
-				"Content-Type");
+		hdr = rspamd_message_get_header_from_hash(npart->raw_headers,
+				"Content-Type", FALSE);
 
 	}
 	else {
@@ -874,7 +891,7 @@ rspamd_mime_process_multipart_node (struct rspamd_task *task,
 	if (hdr != NULL) {
 
 		DL_FOREACH (hdr, cur) {
-			ct = rspamd_content_type_parse (cur->decoded, strlen (cur->decoded),
+			ct = rspamd_content_type_parse (cur->value, strlen (cur->value),
 					task->task_pool);
 
 			/* Here we prefer multipart content-type or any content-type */
@@ -1136,7 +1153,6 @@ rspamd_mime_preprocess_cb (struct rspamd_multipattern *mp,
 		void *context)
 {
 	const gchar *end = text + len, *p = text + match_pos, *bend;
-	gchar *lc_copy;
 	gsize blen;
 	gboolean closing = FALSE;
 	struct rspamd_mime_boundary b;
@@ -1208,13 +1224,21 @@ rspamd_mime_preprocess_cb (struct rspamd_multipattern *mp,
 			b.boundary = p - st->start - 2;
 			b.start = bend - st->start;
 
-			if (closing) {
+			/* Small optimisation as boundaries are usually short strings */
+			gchar *lc_copy, lc_copy_buf[128];
+
+			if (blen + 2 < sizeof(lc_copy_buf)) {
+				lc_copy = lc_copy_buf;
+			}
+			else {
 				lc_copy = g_malloc (blen + 2);
+			}
+
+			if (closing) {
 				memcpy (lc_copy, p, blen + 2);
 				rspamd_str_lc (lc_copy, blen + 2);
 			}
 			else {
-				lc_copy = g_malloc (blen);
 				memcpy (lc_copy, p, blen);
 				rspamd_str_lc (lc_copy, blen);
 			}
@@ -1239,7 +1263,10 @@ rspamd_mime_preprocess_cb (struct rspamd_multipattern *mp,
 				b.closed_hash = 0;
 			}
 
-			g_free (lc_copy);
+			/* Check if a string has been allocated on the heap */
+			if (blen + 2 >= sizeof(lc_copy_buf)) {
+				g_free(lc_copy);
+			}
 			g_array_append_val (st->boundaries, b);
 		}
 	}
@@ -1402,9 +1429,9 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 				}
 			}
 
-			hdr = rspamd_message_get_header_from_hash (
+			hdr = rspamd_message_get_header_from_hash(
 					MESSAGE_FIELD (task, raw_headers),
-					"Content-Type");
+					"Content-Type", FALSE);
 		}
 		else {
 			/* First apply heuristic, maybe we have just headers */
@@ -1432,9 +1459,9 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 					}
 				}
 
-				hdr = rspamd_message_get_header_from_hash (
+				hdr = rspamd_message_get_header_from_hash(
 						MESSAGE_FIELD (task, raw_headers),
-						"Content-Type");
+						"Content-Type", FALSE);
 				task->flags |= RSPAMD_TASK_FLAG_BROKEN_HEADERS;
 			}
 			else {
@@ -1488,8 +1515,8 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 				}
 			}
 
-			hdr = rspamd_message_get_header_from_hash (npart->raw_headers,
-					"Content-Type");
+			hdr = rspamd_message_get_header_from_hash(npart->raw_headers,
+					"Content-Type", FALSE);
 		}
 		else {
 			body_pos = 0;
@@ -1508,7 +1535,7 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 	}
 	else {
 		DL_FOREACH (hdr, cur) {
-			ct = rspamd_content_type_parse (cur->decoded, strlen (cur->decoded),
+			ct = rspamd_content_type_parse (cur->value, strlen (cur->value),
 					task->task_pool);
 
 			/* Here we prefer multipart content-type or any content-type */
