@@ -27,7 +27,7 @@
 
 #define msg_debug_metric(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_metric_log_id, "metric", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(metric)
@@ -39,7 +39,7 @@ static void
 rspamd_scan_result_dtor (gpointer d)
 {
 	struct rspamd_scan_result *r = (struct rspamd_scan_result *)d;
-	struct rspamd_symbol_result sres;
+	struct rspamd_symbol_result *sres;
 
 	rspamd_set_counter_ema (&symbols_count, kh_size (r->symbols), 0.5);
 
@@ -48,8 +48,8 @@ rspamd_scan_result_dtor (gpointer d)
 	}
 
 	kh_foreach_value (r->symbols, sres, {
-		if (sres.options) {
-			kh_destroy (rspamd_options_hash, sres.options);
+		if (sres->options) {
+			kh_destroy (rspamd_options_hash, sres->options);
 		}
 	});
 	kh_destroy (rspamd_symbols_hash, r->symbols);
@@ -91,15 +91,19 @@ rspamd_create_metric_result (struct rspamd_task *task,
 	if (task->cfg) {
 		struct rspamd_action *act, *tmp;
 
-		metric_res->actions_limits = rspamd_mempool_alloc0 (task->task_pool,
-			sizeof (struct rspamd_action_result) * HASH_COUNT (task->cfg->actions));
+		metric_res->actions_config = rspamd_mempool_alloc0 (task->task_pool,
+				sizeof (struct rspamd_action_config) * HASH_COUNT (task->cfg->actions));
 		i = 0;
 
 		HASH_ITER (hh, task->cfg->actions, act, tmp) {
+			metric_res->actions_config[i].flags = RSPAMD_ACTION_RESULT_DEFAULT;
 			if (!(act->flags & RSPAMD_ACTION_NO_THRESHOLD)) {
-				metric_res->actions_limits[i].cur_limit = act->threshold;
+				metric_res->actions_config[i].cur_limit = act->threshold;
 			}
-			metric_res->actions_limits[i].action = act;
+			else {
+				metric_res->actions_config[i].flags |= RSPAMD_ACTION_RESULT_NO_THRESHOLD;
+			}
+			metric_res->actions_config[i].action = act;
 
 			i ++;
 		}
@@ -122,9 +126,14 @@ rspamd_pr_sort (const struct rspamd_passthrough_result *pra,
 	return prb->priority - pra->priority;
 }
 
-void
-rspamd_add_passthrough_result (struct rspamd_task *task, struct rspamd_action *action, guint priority,
-							   double target_score, const gchar *message, const gchar *module, guint flags,
+bool
+rspamd_add_passthrough_result (struct rspamd_task *task,
+							   struct rspamd_action *action,
+							   guint priority,
+							   double target_score,
+							   const gchar *message,
+							   const gchar *module,
+							   uint flags,
 							   struct rspamd_scan_result *scan_result)
 {
 	struct rspamd_passthrough_result *pr;
@@ -133,6 +142,29 @@ rspamd_add_passthrough_result (struct rspamd_task *task, struct rspamd_action *a
 		scan_result = task->result;
 	}
 
+	/* Find the specific action config */
+	struct rspamd_action_config *action_config = NULL;
+
+	for (unsigned int i = 0; i < scan_result->nactions; i ++) {
+		struct rspamd_action_config *cur = &scan_result->actions_config[i];
+
+		/* We assume that all action pointers are static */
+		if (cur->action == action) {
+			action_config = cur;
+			break;
+		}
+	}
+
+	if (action_config && (action_config->flags & RSPAMD_ACTION_RESULT_DISABLED)) {
+		msg_info_task ("<%s>: NOT set pre-result to '%s' %s(%.2f): '%s' from %s(%d); action is disabled",
+				MESSAGE_FIELD_CHECK (task, message_id), action->name,
+				flags & RSPAMD_PASSTHROUGH_LEAST ? "*least " : "",
+				target_score,
+				message, module, priority);
+
+		return false;
+	}
+
 	pr = rspamd_mempool_alloc (task->task_pool, sizeof (*pr));
 	pr->action = action;
 	pr->priority = priority;
@@ -160,6 +192,8 @@ rspamd_add_passthrough_result (struct rspamd_task *task, struct rspamd_action *a
 	}
 
 	scan_result->nresults ++;
+
+	return true;
 }
 
 static inline gdouble
@@ -194,9 +228,10 @@ insert_metric_result (struct rspamd_task *task,
 					  double weight,
 					  const gchar *opt,
 					  struct rspamd_scan_result *metric_res,
-					  enum rspamd_symbol_insert_flags flags)
+					  enum rspamd_symbol_insert_flags flags,
+					  bool *new_sym)
 {
-	struct rspamd_symbol_result *s = NULL;
+	struct rspamd_symbol_result *symbol_result = NULL;
 	gdouble final_score, *gr_score = NULL, next_gf = 1.0, diff;
 	struct rspamd_symbol *sdef;
 	struct rspamd_symbols_group *gr = NULL;
@@ -278,7 +313,7 @@ insert_metric_result (struct rspamd_task *task,
 	k = kh_get (rspamd_symbols_hash, metric_res->symbols, symbol);
 	if (k != kh_end (metric_res->symbols)) {
 		/* Existing metric score */
-		s = &kh_value (metric_res->symbols, k);
+		symbol_result = kh_value (metric_res->symbols, k);
 		if (single) {
 			max_shots = 1;
 		}
@@ -293,14 +328,14 @@ insert_metric_result (struct rspamd_task *task,
 
 		msg_debug_metric ("nshots: %d for symbol %s", max_shots, symbol);
 
-		if (!single && (max_shots > 0 && (s->nshots >= max_shots))) {
+		if (!single && (max_shots > 0 && (symbol_result->nshots >= max_shots))) {
 			single = TRUE;
 		}
 
-		s->nshots ++;
+		symbol_result->nshots ++;
 
 		if (opt) {
-			rspamd_task_add_result_option (task, s, opt, strlen (opt));
+			rspamd_task_add_result_option (task, symbol_result, opt, strlen (opt));
 		}
 
 		/* Adjust diff */
@@ -310,13 +345,13 @@ insert_metric_result (struct rspamd_task *task,
 					symbol, diff);
 		}
 		else {
-			if (fabs (s->score) < fabs (final_score) &&
-				signbit (s->score) == signbit (final_score)) {
+			if (fabs (symbol_result->score) < fabs (final_score) &&
+				signbit (symbol_result->score) == signbit (final_score)) {
 				/* Replace less significant weight with a more significant one */
-				diff = final_score - s->score;
+				diff = final_score - symbol_result->score;
 				msg_debug_metric ("symbol %s can be inserted single time;"
 					  " weight adjusted %.2f + %.2f",
-						symbol, s->score, diff);
+						symbol, symbol_result->score, diff);
 			}
 			else {
 				diff = 0;
@@ -379,23 +414,27 @@ insert_metric_result (struct rspamd_task *task,
 				if (single) {
 					msg_debug_metric ("final score for single symbol %s = %.2f; %.2f diff",
 							symbol, final_score, diff);
-					s->score = final_score;
+					symbol_result->score = final_score;
 				} else {
 					msg_debug_metric ("increase final score for multiple symbol %s += %.2f = %.2f",
-							symbol, s->score, diff);
-					s->score += diff;
+							symbol, symbol_result->score, diff);
+					symbol_result->score += diff;
 				}
 			}
 		}
 	}
 	else {
 		/* New result */
+		if (new_sym) {
+			*new_sym = true;
+		}
+
 		sym_cpy = rspamd_mempool_strdup (task->task_pool, symbol);
 		k = kh_put (rspamd_symbols_hash, metric_res->symbols,
 				sym_cpy, &ret);
 		g_assert (ret > 0);
-		s = &kh_value (metric_res->symbols, k);
-		memset (s, 0, sizeof (*s));
+		symbol_result = rspamd_mempool_alloc0 (task->task_pool, sizeof (*symbol_result));
+		kh_value (metric_res->symbols, k) = symbol_result;
 
 		/* Handle grow factor */
 		if (metric_res->grow_factor && final_score > 0) {
@@ -409,9 +448,9 @@ insert_metric_result (struct rspamd_task *task,
 		msg_debug_metric ("adjust grow factor to %.2f for symbol %s (%.2f final)",
 				next_gf, symbol, final_score);
 
-		s->name = sym_cpy;
-		s->sym = sdef;
-		s->nshots = 1;
+		symbol_result->name = sym_cpy;
+		symbol_result->sym = sdef;
+		symbol_result->nshots = 1;
 
 		if (sdef) {
 			/* Check group limits */
@@ -452,7 +491,7 @@ insert_metric_result (struct rspamd_task *task,
 
 			metric_res->score += final_score;
 			metric_res->grow_factor = next_gf;
-			s->score = final_score;
+			symbol_result->score = final_score;
 
 			if (final_score > epsilon) {
 				metric_res->npositive ++;
@@ -464,21 +503,21 @@ insert_metric_result (struct rspamd_task *task,
 			}
 		}
 		else {
-			s->score = 0;
+			symbol_result->score = 0;
 		}
 
 		if (opt) {
-			rspamd_task_add_result_option (task, s, opt, strlen (opt));
+			rspamd_task_add_result_option (task, symbol_result, opt, strlen (opt));
 		}
 	}
 
 	msg_debug_metric ("final insertion for symbol %s, score %.2f, factor: %f",
 			symbol,
-			s->score,
+			symbol_result->score,
 			final_score);
 	metric_res->nresults ++;
 
-	return s;
+	return symbol_result;
 }
 
 struct rspamd_symbol_result *
@@ -489,7 +528,7 @@ rspamd_task_insert_result_full (struct rspamd_task *task,
 								enum rspamd_symbol_insert_flags flags,
 								struct rspamd_scan_result *result)
 {
-	struct rspamd_symbol_result *s = NULL, *ret = NULL;
+	struct rspamd_symbol_result *symbol_result = NULL, *ret = NULL;
 	struct rspamd_scan_result *mres;
 
 	if (task->processed_stages & (RSPAMD_TASK_STAGE_IDEMPOTENT >> 1)) {
@@ -530,40 +569,50 @@ rspamd_task_insert_result_full (struct rspamd_task *task,
 				}
 			}
 
-			s = insert_metric_result (task,
+			bool new_symbol = false;
+
+			symbol_result = insert_metric_result (task,
 					symbol,
 					weight,
 					opt,
 					mres,
-					flags);
+					flags,
+					&new_symbol);
 
 			if (mres->name == NULL) {
 				/* Default result */
-				ret = s;
+				ret = symbol_result;
 
 				/* Process cache item */
-				if (s && task->cfg->cache && s->sym) {
+				if (symbol_result && task->cfg->cache && symbol_result->sym && symbol_result->nshots == 1) {
 					rspamd_symcache_inc_frequency (task->cfg->cache,
-							s->sym->cache_item);
+							symbol_result->sym->cache_item,
+							symbol_result->sym->name);
 				}
 			}
+			else if (new_symbol) {
+				/* O(N) but we normally don't have any shadow results */
+				LL_APPEND (ret, symbol_result);
+			}
 		}
 	}
 	else {
 		/* Specific insertion */
-		s = insert_metric_result (task,
+		symbol_result = insert_metric_result (task,
 				symbol,
 				weight,
 				opt,
 				result,
-				flags);
-		ret = s;
+				flags,
+				NULL);
+		ret = symbol_result;
 
 		if (result->name == NULL) {
 			/* Process cache item */
-			if (s && task->cfg->cache && s->sym) {
+			if (symbol_result && task->cfg->cache && symbol_result->sym && symbol_result->nshots == 1) {
 				rspamd_symcache_inc_frequency (task->cfg->cache,
-						s->sym->cache_item);
+						symbol_result->sym->cache_item,
+						symbol_result->sym->name);
 			}
 		}
 	}
@@ -679,54 +728,75 @@ rspamd_task_add_result_option (struct rspamd_task *task,
 	gsize cpy_len;
 	khiter_t k;
 	gint r;
+	struct rspamd_symbol_result *cur;
 
 	if (s && val) {
-		if (s->opts_len < 0) {
-			/* Cannot add more options, give up */
-			msg_debug_task ("cannot add more options to symbol %s when adding option %s",
-					s->name, val);
-			return FALSE;
-		}
-
-		if (!s->options) {
-			s->options = kh_init (rspamd_options_hash);
-		}
-
-		if (vlen + s->opts_len > task->cfg->max_opts_len) {
-			/* Add truncated option */
-			msg_info_task ("cannot add more options to symbol %s when adding option %s",
-					s->name, val);
-			val = "...";
-			vlen = 3;
-			s->opts_len = -1;
-		}
-
-		if (!(s->sym && (s->sym->flags & RSPAMD_SYMBOL_FLAG_ONEPARAM)) &&
-				kh_size (s->options) < task->cfg->default_max_shots) {
-			opt_cpy = rspamd_task_option_safe_copy (task, val, vlen, &cpy_len);
-			/* Append new options */
-			srch.option = (gchar *)opt_cpy;
-			srch.optlen = cpy_len;
-			k = kh_get (rspamd_options_hash, s->options, &srch);
-
-			if (k == kh_end (s->options)) {
-				opt = rspamd_mempool_alloc0 (task->task_pool, sizeof (*opt));
-				opt->optlen = cpy_len;
-				opt->option = opt_cpy;
-
-				kh_put (rspamd_options_hash, s->options, opt, &r);
-				DL_APPEND (s->opts_head, opt);
-
-				ret = TRUE;
+		/*
+		 * Here we assume that this function is all the time called with the
+		 * symbol from the default result, not some shadow result, or
+		 * the option insertion will be wrong
+		 */
+		LL_FOREACH (s, cur) {
+			if (cur->opts_len < 0) {
+				/* Cannot add more options, give up */
+				msg_debug_task ("cannot add more options to symbol %s when adding option %s",
+						cur->name, val);
+				ret = FALSE;
+				continue;
 			}
-		}
-		else {
-			/* Skip addition */
-			ret = FALSE;
-		}
 
-		if (ret && s->opts_len >= 0) {
-			s->opts_len += vlen;
+			if (!cur->options) {
+				cur->options = kh_init (rspamd_options_hash);
+			}
+
+			if (vlen + cur->opts_len > task->cfg->max_opts_len) {
+				/* Add truncated option */
+				msg_info_task ("cannot add more options to symbol %s when adding option %s",
+						cur->name, val);
+				val = "...";
+				vlen = 3;
+				cur->opts_len = -1;
+			}
+
+			if (!(cur->sym && (cur->sym->flags & RSPAMD_SYMBOL_FLAG_ONEPARAM)) &&
+				kh_size (cur->options) < task->cfg->default_max_shots) {
+
+				srch.option = (gchar *) val;
+				srch.optlen = vlen;
+				k = kh_get (rspamd_options_hash, cur->options, &srch);
+
+				if (k == kh_end (cur->options)) {
+					opt_cpy = rspamd_task_option_safe_copy (task, val, vlen, &cpy_len);
+					if (cpy_len != vlen) {
+						srch.option = (gchar *) opt_cpy;
+						srch.optlen = cpy_len;
+						k = kh_get (rspamd_options_hash, cur->options, &srch);
+					}
+					/* Append new options */
+					if (k == kh_end (cur->options)) {
+						opt = rspamd_mempool_alloc0 (task->task_pool, sizeof(*opt));
+						opt->optlen = cpy_len;
+						opt->option = opt_cpy;
+
+						kh_put (rspamd_options_hash, cur->options, opt, &r);
+						DL_APPEND (cur->opts_head, opt);
+
+						if (s == cur) {
+							ret = TRUE;
+						}
+					}
+				}
+			}
+			else {
+				/* Skip addition */
+				if (s == cur) {
+					ret = FALSE;
+				}
+			}
+
+			if (ret && cur->opts_len >= 0) {
+				cur->opts_len += vlen;
+			}
 		}
 	}
 	else if (!val) {
@@ -738,17 +808,31 @@ rspamd_task_add_result_option (struct rspamd_task *task,
 	return ret;
 }
 
+struct rspamd_action_config *
+rspamd_find_action_config_for_action (struct rspamd_scan_result *scan_result,
+									  struct rspamd_action *act)
+{
+	for (unsigned int i = 0; i < scan_result->nactions; i ++) {
+		struct rspamd_action_config *cur = &scan_result->actions_config[i];
+
+		if (act == cur->action) {
+			return cur;
+		}
+	}
+
+	return NULL;
+}
+
 struct rspamd_action *
 rspamd_check_action_metric (struct rspamd_task *task,
 							struct rspamd_passthrough_result **ppr,
 							struct rspamd_scan_result *scan_result)
 {
-	struct rspamd_action_result *action_lim,
+	struct rspamd_action_config *action_lim,
 			*noaction = NULL;
 	struct rspamd_action *selected_action = NULL, *least_action = NULL;
 	struct rspamd_passthrough_result *pr, *sel_pr = NULL;
 	double max_score = -(G_MAXDOUBLE), sc;
-	int i;
 	gboolean seen_least = FALSE;
 
 	if (scan_result == NULL) {
@@ -757,6 +841,14 @@ rspamd_check_action_metric (struct rspamd_task *task,
 
 	if (scan_result->passthrough_result != NULL)  {
 		DL_FOREACH (scan_result->passthrough_result, pr) {
+			struct rspamd_action_config *act_config =
+					rspamd_find_action_config_for_action (scan_result, pr->action);
+
+			/* Skip disabled actions */
+			if (act_config && (act_config->flags & RSPAMD_ACTION_RESULT_DISABLED)) {
+				continue;
+			}
+
 			if (!seen_least || !(pr->flags & RSPAMD_PASSTHROUGH_LEAST)) {
 				sc = pr->target_score;
 				selected_action = pr->action;
@@ -815,14 +907,18 @@ rspamd_check_action_metric (struct rspamd_task *task,
 	/*
 	 * Select result by score
 	 */
-	for (i = scan_result->nactions - 1; i >= 0; i--) {
-		action_lim = &scan_result->actions_limits[i];
+	for (size_t i = scan_result->nactions - 1; i != (size_t)-1; i--) {
+		action_lim = &scan_result->actions_config[i];
 		sc = action_lim->cur_limit;
 
 		if (action_lim->action->action_type == METRIC_ACTION_NOACTION) {
 			noaction = action_lim;
 		}
 
+		if ((action_lim->flags & (RSPAMD_ACTION_RESULT_DISABLED|RSPAMD_ACTION_RESULT_NO_THRESHOLD))) {
+			continue;
+		}
+
 		if (isnan (sc) ||
 			(action_lim->action->flags & (RSPAMD_ACTION_NO_THRESHOLD|RSPAMD_ACTION_HAM))) {
 			continue;
@@ -887,10 +983,10 @@ rspamd_task_find_symbol_result (struct rspamd_task *task, const char *sym,
 		result = task->result;
 	}
 
-	k = kh_get (rspamd_symbols_hash, result->symbols, sym);
+	k = kh_get(rspamd_symbols_hash, result->symbols, sym);
 
 	if (k != kh_end (result->symbols)) {
-		res = &kh_value (result->symbols, k);
+		res = kh_value (result->symbols, k);
 	}
 
 	return res;
@@ -912,7 +1008,7 @@ struct rspamd_symbol_result* rspamd_task_remove_symbol_result (
 	k = kh_get (rspamd_symbols_hash, result->symbols, symbol);
 
 	if (k != kh_end (result->symbols)) {
-		res = &kh_value (result->symbols, k);
+		res = kh_value (result->symbols, k);
 
 		if (!isnan (res->score)) {
 			/* Remove score from the result */
@@ -922,15 +1018,16 @@ struct rspamd_symbol_result* rspamd_task_remove_symbol_result (
 			if (result->sym_groups && res->sym) {
 				struct rspamd_symbol_group *gr;
 				gint i;
+				khiter_t k_groups;
 
 				PTR_ARRAY_FOREACH (res->sym->groups, i, gr) {
 					gdouble *gr_score;
 
-					k = kh_get (rspamd_symbols_group_hash,
+					k_groups = kh_get (rspamd_symbols_group_hash,
 							result->sym_groups, gr);
 
-					if (k != kh_end (result->sym_groups)) {
-						gr_score = &kh_value (result->sym_groups, k);
+					if (k_groups != kh_end (result->sym_groups)) {
+						gr_score = &kh_value (result->sym_groups, k_groups);
 
 						if (gr_score) {
 							*gr_score -= res->score;
@@ -955,7 +1052,7 @@ rspamd_task_symbol_result_foreach (struct rspamd_task *task,
 								   gpointer ud)
 {
 	const gchar *kk;
-	struct rspamd_symbol_result res;
+	struct rspamd_symbol_result *res;
 
 	if (result == NULL) {
 		/* Use default result */
@@ -964,7 +1061,7 @@ rspamd_task_symbol_result_foreach (struct rspamd_task *task,
 
 	if (func) {
 		kh_foreach (result->symbols, kk, res, {
-			func ((gpointer)kk, (gpointer)&res, ud);
+			func ((gpointer)kk, (gpointer)res, ud);
 		});
 	}
 }
@@ -989,4 +1086,4 @@ rspamd_find_metric_result (struct rspamd_task *task,
 	}
 
 	return NULL;
-}
\ No newline at end of file
+}
