@@ -27,10 +27,6 @@
 extern "C" {
 #endif
 
-int rspamd_smtp_received_parse (struct rspamd_task *task,
-								const char *data, size_t len,
-								struct rspamd_received_header *rh);
-
 int rspamd_smtp_addr_parse (const char *data, size_t len,
 							struct rspamd_email_address *addr);
 
