@@ -146,7 +146,7 @@ struct rspamd_symbol {
 	struct rspamd_symbols_group *gr; /* Main group */
 	GPtrArray *groups; /* Other groups */
 	guint flags;
-	struct rspamd_symcache_item *cache_item;
+	void *cache_item;
 	gint nshots;
 };
 
@@ -718,6 +718,14 @@ gboolean rspamd_config_maybe_disable_action (struct rspamd_config *cfg,
 gboolean rspamd_config_is_module_enabled (struct rspamd_config *cfg,
 										  const gchar *module_name);
 
+/**
+ * Verifies enabled/disabled combination in the specified object
+ * @param obj
+ * @return TRUE if there is no explicit disable in the object found
+ */
+gboolean rspamd_config_is_enabled_from_ucl (rspamd_mempool_t *pool,
+		const ucl_object_t *obj);
+
 /*
  * Get action from a string
  */
