@@ -36,8 +36,12 @@
 #include "contrib/libottery/ottery.h"
 #include "contrib/fastutf8/fastutf8.h"
 
-#define ZSTD_STATIC_LINKING_ONLY
-#include "contrib/zstd/zstd.h"
+#ifdef SYS_ZSTD
+#  include "zstd.h"
+#else
+#  define ZSTD_STATIC_LINKING_ONLY
+#  include "contrib/zstd/zstd.h"
+#endif
 
 #ifdef HAVE_OPENSSL
 #include <openssl/rand.h>
@@ -53,6 +57,7 @@
 #include <sys/resource.h>
 #endif
 #include <math.h>
+#include "libserver/composites/composites.h"
 
 #include "blas-config.h"
 
@@ -159,13 +164,14 @@ rspamd_config_new (enum rspamd_config_init_flags flags)
 	rspamd_mempool_t *pool;
 
 	pool = rspamd_mempool_new (8 * 1024 * 1024, "cfg", 0);
-	cfg = rspamd_mempool_alloc0 (pool, sizeof (*cfg));
+	cfg = rspamd_mempool_alloc0_type(pool, struct rspamd_config);
 	/* Allocate larger pool for cfg */
 	cfg->cfg_pool = pool;
 	cfg->dns_timeout = 1.0;
 	cfg->dns_retransmits = 5;
 	/* 16 sockets per DNS server */
 	cfg->dns_io_per_server = 16;
+	cfg->unknown_weight = NAN;
 
 	/* Add all internal actions to keep compatibility */
 	for (int i = METRIC_ACTION_REJECT; i < METRIC_ACTION_MAX; i ++) {
@@ -196,8 +202,7 @@ rspamd_config_new (enum rspamd_config_init_flags flags)
 
 
 	rspamd_config_init_metric (cfg);
-	cfg->composite_symbols =
-		g_hash_table_new (rspamd_str_hash, rspamd_str_equal);
+	cfg->composites_manager = rspamd_composites_manager_create(cfg);
 	cfg->classifiers_symbols = g_hash_table_new (rspamd_str_hash,
 			rspamd_str_equal);
 	cfg->cfg_params = g_hash_table_new (rspamd_str_hash, rspamd_str_equal);
@@ -244,6 +249,9 @@ rspamd_config_new (enum rspamd_config_init_flags flags)
 	cfg->lua_gc_step = DEFAULT_LUA_GC_STEP;
 	cfg->full_gc_iters = DEFAULT_GC_MAXITERS;
 
+	/* Default hyperscan cache */
+	cfg->hs_cache_dir = RSPAMD_DBDIR "/";
+
 	if (!(flags & RSPAMD_CONFIG_INIT_SKIP_LUA)) {
 		cfg->lua_state = rspamd_lua_init (flags & RSPAMD_CONFIG_INIT_WIPE_LUA_MEM);
 		cfg->own_lua_state = TRUE;
@@ -265,15 +273,15 @@ rspamd_config_new (enum rspamd_config_init_flags flags)
 	cfg->images_cache_size = 256;
 	cfg->monitored_ctx = rspamd_monitored_ctx_init ();
 	cfg->neighbours = ucl_object_typed_new (UCL_OBJECT);
-#ifdef WITH_HIREDIS
 	cfg->redis_pool = rspamd_redis_pool_init ();
-#endif
 	cfg->default_max_shots = DEFAULT_MAX_SHOTS;
 	cfg->max_sessions_cache = DEFAULT_MAX_SESSIONS;
 	cfg->maps_cache_dir = rspamd_mempool_strdup (cfg->cfg_pool, RSPAMD_DBDIR);
 	cfg->c_modules = g_ptr_array_new ();
 	cfg->heartbeat_interval = 10.0;
 
+	cfg->enable_css_parser = true;
+
 	REF_INIT_RETAIN (cfg, rspamd_config_free);
 
 	return cfg;
@@ -319,8 +327,6 @@ rspamd_config_free (struct rspamd_config *cfg)
 	ucl_object_unref (cfg->config_comments);
 	ucl_object_unref (cfg->doc_strings);
 	ucl_object_unref (cfg->neighbours);
-	g_hash_table_remove_all (cfg->composite_symbols);
-	g_hash_table_unref (cfg->composite_symbols);
 	g_hash_table_remove_all (cfg->cfg_params);
 	g_hash_table_unref (cfg->cfg_params);
 	g_hash_table_unref (cfg->classifiers_symbols);
@@ -330,24 +336,22 @@ rspamd_config_free (struct rspamd_config *cfg)
 	g_hash_table_unref (cfg->trusted_keys);
 
 	rspamd_re_cache_unref (cfg->re_cache);
-	rspamd_upstreams_library_unref (cfg->ups_ctx);
 	g_ptr_array_free (cfg->c_modules, TRUE);
 
+	if (cfg->monitored_ctx) {
+		rspamd_monitored_ctx_destroy (cfg->monitored_ctx);
+	}
+
 	if (cfg->lua_state && cfg->own_lua_state) {
 		lua_thread_pool_free (cfg->lua_thread_pool);
 		lua_close (cfg->lua_state);
 	}
 
-#ifdef WITH_HIREDIS
 	if (cfg->redis_pool) {
 		rspamd_redis_pool_destroy (cfg->redis_pool);
 	}
-#endif
-
-	if (cfg->monitored_ctx) {
-		rspamd_monitored_ctx_destroy (cfg->monitored_ctx);
-	}
 
+	rspamd_upstreams_library_unref (cfg->ups_ctx);
 	HASH_CLEAR (hh, cfg->actions);
 
 	rspamd_mempool_destructors_enforce (cfg->cfg_pool);
@@ -816,7 +820,7 @@ rspamd_config_post_load (struct rspamd_config *cfg,
 	rspamd_regexp_library_init (cfg);
 	rspamd_multipattern_library_init (cfg->hs_cache_dir);
 
-#if defined(WITH_HYPERSCAN) && !defined(__aarch64__)
+#if defined(WITH_HYPERSCAN) && !defined(__aarch64__) && !defined(__powerpc64__)
 	if (!cfg->disable_hyperscan) {
 		if (!(cfg->libs_ctx->crypto_ctx->cpu_config & CPUID_SSSE3)) {
 			msg_warn_config ("CPU doesn't have SSSE3 instructions set "
@@ -930,109 +934,17 @@ rspamd_config_post_load (struct rspamd_config *cfg,
 		ret = rspamd_symcache_validate (cfg->cache, cfg, FALSE) && ret;
 	}
 
-	if (opts & RSPAMD_CONFIG_INIT_PRELOAD_MAPS) {
-		rspamd_map_preload (cfg);
-	}
-
 	if (opts & RSPAMD_CONFIG_INIT_POST_LOAD_LUA) {
 		rspamd_lua_run_config_post_init (cfg->lua_state, cfg);
 	}
 
+	if (opts & RSPAMD_CONFIG_INIT_PRELOAD_MAPS) {
+		rspamd_map_preload (cfg);
+	}
+
 	return ret;
 }
 
-#if 0
-void
-parse_err (const gchar *fmt, ...)
-{
-	va_list aq;
-	gchar logbuf[BUFSIZ], readbuf[32];
-	gint r;
-
-	va_start (aq, fmt);
-	rspamd_strlcpy (readbuf, yytext, sizeof (readbuf));
-
-	r = snprintf (logbuf,
-			sizeof (logbuf),
-			"config file parse error! line: %d, text: %s, reason: ",
-			yylineno,
-			readbuf);
-	r += vsnprintf (logbuf + r, sizeof (logbuf) - r, fmt, aq);
-
-	va_end (aq);
-	g_critical ("%s", logbuf);
-}
-
-void
-parse_warn (const gchar *fmt, ...)
-{
-	va_list aq;
-	gchar logbuf[BUFSIZ], readbuf[32];
-	gint r;
-
-	va_start (aq, fmt);
-	rspamd_strlcpy (readbuf, yytext, sizeof (readbuf));
-
-	r = snprintf (logbuf,
-			sizeof (logbuf),
-			"config file parse warning! line: %d, text: %s, reason: ",
-			yylineno,
-			readbuf);
-	r += vsnprintf (logbuf + r, sizeof (logbuf) - r, fmt, aq);
-
-	va_end (aq);
-	g_warning ("%s", logbuf);
-}
-#endif
-
-void
-rspamd_config_unescape_quotes (gchar *line)
-{
-	gchar *c = line, *t;
-
-	while (*c) {
-		if (*c == '\\' && *(c + 1) == '"') {
-			t = c;
-			while (*t) {
-				*t = *(t + 1);
-				t++;
-			}
-		}
-		c++;
-	}
-}
-
-GList *
-rspamd_config_parse_comma_list (rspamd_mempool_t * pool, const gchar *line)
-{
-	GList *res = NULL;
-	const gchar *c, *p;
-	gchar *str;
-
-	c = line;
-	p = c;
-
-	while (*p) {
-		if (*p == ',' && *c != *p) {
-			str = rspamd_mempool_alloc (pool, p - c + 1);
-			rspamd_strlcpy (str, c, p - c + 1);
-			res = g_list_prepend (res, str);
-			/* Skip spaces */
-			while (g_ascii_isspace (*(++p))) ;
-			c = p;
-			continue;
-		}
-		p++;
-	}
-	if (res != NULL) {
-		rspamd_mempool_add_destructor (pool,
-			(rspamd_mempool_destruct_t) g_list_free,
-			res);
-	}
-
-	return res;
-}
-
 struct rspamd_classifier_config *
 rspamd_config_new_classifier (struct rspamd_config *cfg,
 	struct rspamd_classifier_config *c)
@@ -1197,7 +1109,7 @@ rspamd_include_map_handler (const guchar *data, gsize len,
  * $RUNDIR - local states directory
  * $DBDIR - databases dir
  * $LOGDIR - logs dir
- * $PLUGINSDIR - pluggins dir
+ * $PLUGINSDIR - plugins dir
  * $PREFIX - installation prefix
  * $VERSION - rspamd version
  */
@@ -1365,7 +1277,7 @@ rspamd_config_check_statfiles (struct rspamd_classifier_config *cf)
 		/* We have only one statfile */
 		return FALSE;
 	}
-	/* We have not detected any statfile that has different class, so turn on euristic based on symbol's name */
+	/* We have not detected any statfile that has different class, so turn on heuristic based on symbol's name */
 	has_other = FALSE;
 	cur = cf->statfiles;
 	while (cur) {
@@ -1658,11 +1570,11 @@ rspamd_config_new_symbol (struct rspamd_config *cfg, const gchar *symbol,
 	/* Search for symbol group */
 	if (group == NULL) {
 		group = "ungrouped";
-		sym_def->flags |= RSPAMD_SYMBOL_FLAG_UNGROUPPED;
+		sym_def->flags |= RSPAMD_SYMBOL_FLAG_UNGROUPED;
 	}
 	else {
 		if (strcmp (group, "ungrouped") == 0) {
-			sym_def->flags |= RSPAMD_SYMBOL_FLAG_UNGROUPPED;
+			sym_def->flags |= RSPAMD_SYMBOL_FLAG_UNGROUPED;
 		}
 	}
 
@@ -1675,7 +1587,7 @@ rspamd_config_new_symbol (struct rspamd_config *cfg, const gchar *symbol,
 	sym_def->gr = sym_group;
 	g_hash_table_insert (sym_group->symbols, sym_def->name, sym_def);
 
-	if (!(sym_def->flags & RSPAMD_SYMBOL_FLAG_UNGROUPPED)) {
+	if (!(sym_def->flags & RSPAMD_SYMBOL_FLAG_UNGROUPED)) {
 		g_ptr_array_add (sym_def->groups, sym_group);
 	}
 }
@@ -1713,7 +1625,7 @@ rspamd_config_add_symbol (struct rspamd_config *cfg,
 			}
 
 			if (!has_group) {
-				/* Non-empty group has a priority over non-groupped one */
+				/* Non-empty group has a priority over non-grouped one */
 				sym_group = g_hash_table_lookup (cfg->groups, group);
 
 				if (sym_group == NULL) {
@@ -1721,12 +1633,13 @@ rspamd_config_add_symbol (struct rspamd_config *cfg,
 					sym_group = rspamd_config_new_group (cfg, group);
 				}
 
-				if (!sym_def->gr) {
+				if ((!sym_def->gr) || (sym_def->flags & RSPAMD_SYMBOL_FLAG_UNGROUPED)) {
 					sym_def->gr = sym_group;
+					sym_def->flags &= ~RSPAMD_SYMBOL_FLAG_UNGROUPED;
 				}
 
 				g_hash_table_insert (sym_group->symbols, sym_def->name, sym_def);
-				sym_def->flags &= ~(RSPAMD_SYMBOL_FLAG_UNGROUPPED);
+				sym_def->flags &= ~(RSPAMD_SYMBOL_FLAG_UNGROUPED);
 				g_ptr_array_add (sym_def->groups, sym_group);
 			}
 		}
@@ -1845,7 +1758,7 @@ rspamd_config_add_symbol_group (struct rspamd_config *cfg,
 		}
 
 		if (!has_group) {
-			/* Non-empty group has a priority over non-groupped one */
+			/* Non-empty group has a priority over non-grouped one */
 			sym_group = g_hash_table_lookup (cfg->groups, group);
 
 			if (sym_group == NULL) {
@@ -1858,7 +1771,7 @@ rspamd_config_add_symbol_group (struct rspamd_config *cfg,
 			}
 
 			g_hash_table_insert (sym_group->symbols, sym_def->name, sym_def);
-			sym_def->flags &= ~(RSPAMD_SYMBOL_FLAG_UNGROUPPED);
+			sym_def->flags &= ~(RSPAMD_SYMBOL_FLAG_UNGROUPED);
 			g_ptr_array_add (sym_def->groups, sym_group);
 
 			return TRUE;
@@ -1868,13 +1781,74 @@ rspamd_config_add_symbol_group (struct rspamd_config *cfg,
 	return FALSE;
 }
 
+gboolean
+rspamd_config_is_enabled_from_ucl (rspamd_mempool_t *pool,
+		const ucl_object_t *obj)
+{
+	{
+		const ucl_object_t *enabled;
+
+		enabled = ucl_object_lookup(obj, "enabled");
+
+		if (enabled) {
+			if (ucl_object_type(enabled) == UCL_BOOLEAN) {
+				return ucl_object_toboolean(enabled);
+			}
+			else if (ucl_object_type(enabled) == UCL_STRING) {
+				gchar ret;
+
+				ret = rspamd_config_parse_flag(ucl_object_tostring(enabled), 0);
+
+				if (ret == 0) {
+					return FALSE;
+				}
+				else if (ret == -1) {
+
+					msg_info_pool_check ("wrong value for the `enabled` key");
+					return FALSE;
+				}
+				/* Default return is TRUE here */
+			}
+		}
+	}
+
+	{
+		const ucl_object_t *disabled;
+
+		disabled = ucl_object_lookup(obj, "disabled");
+
+		if (disabled) {
+			if (ucl_object_type(disabled) == UCL_BOOLEAN) {
+				return !ucl_object_toboolean(disabled);
+			}
+			else if (ucl_object_type(disabled) == UCL_STRING) {
+				gchar ret;
+
+				ret = rspamd_config_parse_flag(ucl_object_tostring(disabled), 0);
+
+				if (ret == 0) {
+					return TRUE;
+				}
+				else if (ret == -1) {
+
+					msg_info_pool_check ("wrong value for the `disabled` key");
+					return FALSE;
+				}
+
+				return FALSE;
+			}
+		}
+	}
+
+	return TRUE;
+}
 
 gboolean
 rspamd_config_is_module_enabled (struct rspamd_config *cfg,
 		const gchar *module_name)
 {
-	gboolean is_c = FALSE;
-	const ucl_object_t *conf, *enabled;
+	gboolean is_c = FALSE, enabled;
+	const ucl_object_t *conf;
 	GList *cur;
 	struct rspamd_symbols_group *gr;
 	lua_State *L = cfg->lua_state;
@@ -1933,45 +1907,16 @@ rspamd_config_is_module_enabled (struct rspamd_config *cfg,
 		}
 	}
 	else {
-		enabled = ucl_object_lookup (conf, "enabled");
+		enabled = rspamd_config_is_enabled_from_ucl (cfg->cfg_pool, conf);
 
-		if (enabled) {
-			if (ucl_object_type (enabled) == UCL_BOOLEAN) {
-				if (!ucl_object_toboolean (enabled)) {
-					rspamd_plugins_table_push_elt (L,
-							"disabled_explicitly", module_name);
+		if (!enabled) {
+			rspamd_plugins_table_push_elt (L,
+					"disabled_explicitly", module_name);
 
-					msg_info_config (
-							"%s module %s is disabled in the configuration",
-							is_c ? "internal" : "lua", module_name);
-					return FALSE;
-				}
-			}
-			else if (ucl_object_type (enabled) == UCL_STRING) {
-				gint ret;
-
-				ret = rspamd_config_parse_flag (ucl_object_tostring (enabled), 0);
-
-				if (ret == 0) {
-					rspamd_plugins_table_push_elt (L,
-							"disabled_explicitly", module_name);
-
-					msg_info_config (
-							"%s module %s is disabled in the configuration",
-							is_c ? "internal" : "lua", module_name);
-					return FALSE;
-				}
-				else if (ret == -1) {
-					rspamd_plugins_table_push_elt (L,
-							"disabled_failed", module_name);
-
-					msg_info_config (
-							"%s module %s has wrong enabled flag (%s) in the configuration",
-							is_c ? "internal" : "lua", module_name,
-							ucl_object_tostring (enabled));
-					return FALSE;
-				}
-			}
+			msg_info_config (
+					"%s module %s is disabled in the configuration",
+					is_c ? "internal" : "lua", module_name);
+			return FALSE;
 		}
 	}
 
@@ -2696,7 +2641,7 @@ rspamd_init_libs (void)
 	if ((ctx->crypto_ctx->cpu_config & CPUID_RDRAND) == 0) {
 		ottery_config_disable_entropy_sources (ottery_cfg,
 				OTTERY_ENTROPY_SRC_RDRAND);
-#if OPENSSL_VERSION_NUMBER >= 0x1000104fL && !defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER >= 0x1000104fL && OPENSSL_VERSION_NUMBER < 0x30000000L && !defined(LIBRESSL_VERSION_NUMBER)
 		RAND_set_rand_engine (NULL);
 #endif
 	}
@@ -2844,7 +2789,11 @@ rspamd_config_libs (struct rspamd_external_libs_ctx *ctx,
 
 			/* Toggle FIPS mode */
 			if (mode == 0) {
+#if defined(OPENSSL_VERSION_MAJOR) && (OPENSSL_VERSION_MAJOR >= 3)
+				if (EVP_set_default_properties (NULL, "fips=yes") != 1) {
+#else
 				if (FIPS_mode_set (1) != 1) {
+#endif
 					err = ERR_get_error ();
 				}
 			}
@@ -2853,7 +2802,11 @@ rspamd_config_libs (struct rspamd_external_libs_ctx *ctx,
 			}
 
 			if (err != (unsigned long)-1) {
+#if defined(OPENSSL_VERSION_MAJOR) && (OPENSSL_VERSION_MAJOR >= 3)
+				msg_err_config ("EVP_set_default_properties failed: %s",
+#else
 				msg_err_config ("FIPS_mode_set failed: %s",
+#endif
 						ERR_error_string (err, NULL));
 				ret = FALSE;
 			}
@@ -2910,7 +2863,7 @@ rspamd_libs_reset_decompression (struct rspamd_external_libs_ctx *ctx)
 		return FALSE;
 	}
 	else {
-		r = ZSTD_resetDStream (ctx->in_zstream);
+		r = ZSTD_DCtx_reset (ctx->in_zstream, ZSTD_reset_session_only);
 
 		if (ZSTD_isError (r)) {
 			msg_err ("cannot init decompression stream: %s",
@@ -2935,7 +2888,10 @@ rspamd_libs_reset_compression (struct rspamd_external_libs_ctx *ctx)
 	}
 	else {
 		/* Dictionary will be reused automatically if specified */
-		r = ZSTD_resetCStream (ctx->out_zstream, 0);
+		r = ZSTD_CCtx_reset (ctx->out_zstream, ZSTD_reset_session_only);
+		if (!ZSTD_isError (r)) {
+			r = ZSTD_CCtx_setPledgedSrcSize (ctx->out_zstream, ZSTD_CONTENTSIZE_UNKNOWN);
+		}
 
 		if (ZSTD_isError (r)) {
 			msg_err ("cannot init compression stream: %s",
