@@ -25,26 +25,26 @@
 #include <vector>
 #include <variant>
 #include "libutil/cxx/util.hxx"
-#include "contrib/robin-hood/robin_hood.h"
+#include "contrib/ankerl/unordered_dense.h"
 
 #include "composites_internal.hxx"
 
 #define msg_err_composites(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         "composites", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_composites(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         "composites", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_composites(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         "composites", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 #define msg_debug_composites(...)  rspamd_conditional_debug_fast (NULL, task->from_addr, \
         rspamd_composites_log_id, "composites", task->task_pool->tag.uid, \
-        __FUNCTION__, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(composites)
@@ -82,7 +82,7 @@ struct composites_data {
 	struct rspamd_task *task;
 	struct rspamd_composite *composite;
 	struct rspamd_scan_result *metric_res;
-	robin_hood::unordered_flat_map<std::string_view,
+	ankerl::unordered_dense::map<std::string_view,
 			std::vector<symbol_remove_data>> symbols_to_remove;
 	std::vector<bool> checked;
 
@@ -802,6 +802,8 @@ composites_foreach_callback(gpointer key, gpointer value, void *data)
 	cd->composite = comp;
 	task = cd->task;
 
+	msg_debug_composites("process composite %s", str_key);
+
 	if (!cd->checked[cd->composite->id * 2]) {
 		if (rspamd_symcache_is_checked(cd->task, cd->task->cfg->cache,
 				str_key)) {
