@@ -68,23 +68,23 @@ enum rspamd_dkim_param_type {
 
 #define msg_err_dkim(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         "dkim", ctx->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_dkim(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         "dkim", ctx->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_dkim(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         "dkim", ctx->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug_dkim(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_dkim_log_id, "dkim", ctx->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug_dkim_taskless(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_dkim_log_id, "dkim", "", \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(dkim)
@@ -151,6 +151,7 @@ struct rspamd_dkim_context_s {
 
 struct rspamd_dkim_key_s {
 	guint8 *keydata;
+	guint8 *raw_key;
 	gsize keylen;
 	gsize decoded_len;
 	gchar key_id[RSPAMD_DKIM_KEY_ID_LEN];
@@ -346,7 +347,7 @@ rspamd_dkim_parse_domain (rspamd_dkim_context_t * ctx,
 			g_set_error (err,
 					DKIM_ERROR,
 					DKIM_SIGERROR_INVALID_H,
-					"invalid dkim domain tag %*.s: idna failed",
+					"invalid dkim domain tag %.*s: idna failed",
 					(int)len, param);
 
 			return FALSE;
@@ -447,7 +448,7 @@ rspamd_dkim_parse_selector (rspamd_dkim_context_t * ctx,
 			g_set_error (err,
 					DKIM_ERROR,
 					DKIM_SIGERROR_INVALID_H,
-					"invalid dkim selector tag %*.s: idna failed",
+					"invalid dkim selector tag %.*s: idna failed",
 					(int)len, param);
 
 			return FALSE;
@@ -805,7 +806,7 @@ rspamd_create_dkim_context (const gchar *sig,
 							GError **err)
 {
 	const gchar *p, *c, *tag = NULL, *end;
-	gsize taglen;
+	gint taglen;
 	gint param = DKIM_PARAM_UNKNOWN;
 	const EVP_MD *md_alg;
 	time_t now;
@@ -859,7 +860,7 @@ rspamd_create_dkim_context (const gchar *sig,
 		switch (state) {
 		case DKIM_STATE_TAG:
 			if (g_ascii_isspace (*p)) {
-				taglen = p - c;
+				taglen = (int)(p - c);
 				while (*p && g_ascii_isspace (*p)) {
 					/* Skip spaces before '=' sign */
 					p++;
@@ -888,7 +889,17 @@ rspamd_create_dkim_context (const gchar *sig,
 			}
 			else {
 				taglen++;
-				p++;
+
+				if (taglen > G_MAXINT8) {
+					g_set_error (err,
+							DKIM_ERROR,
+							DKIM_SIGERROR_UNKNOWN,
+							"too long dkim tag");
+					state = DKIM_STATE_ERROR;
+				}
+				else {
+					p++;
+				}
 			}
 			break;
 		case DKIM_STATE_AFTER_TAG:
@@ -902,7 +913,7 @@ rspamd_create_dkim_context (const gchar *sig,
 				state = DKIM_STATE_ERROR;
 				break;
 			case 1:
-				/* Simple tags */
+				/* 1 character tags */
 				switch (*tag) {
 				case 'v':
 					if (type == RSPAMD_DKIM_NORMAL) {
@@ -972,16 +983,13 @@ rspamd_create_dkim_context (const gchar *sig,
 					param = DKIM_PARAM_IGNORE;
 					break;
 				default:
-					g_set_error (err,
-						DKIM_ERROR,
-						DKIM_SIGERROR_UNKNOWN,
-						"invalid dkim param: %c",
-						*tag);
-					state = DKIM_STATE_ERROR;
+					param = DKIM_PARAM_UNKNOWN;
+					msg_debug_dkim("unknown DKIM param %c, ignoring it", *tag);
 					break;
 				}
 				break;
 			case 2:
+				/* Two characters tags, e.g. `bh` */
 				if (tag[0] == 'b' && tag[1] == 'h') {
 					if (type == RSPAMD_DKIM_ARC_SEAL) {
 						g_set_error (err,
@@ -989,7 +997,6 @@ rspamd_create_dkim_context (const gchar *sig,
 								DKIM_SIGERROR_UNKNOWN,
 								"ARC seal must NOT have bh= tag");
 						state = DKIM_STATE_ERROR;
-						break;
 					}
 					else {
 						param = DKIM_PARAM_BODYHASH;
@@ -1002,31 +1009,23 @@ rspamd_create_dkim_context (const gchar *sig,
 								DKIM_SIGERROR_UNKNOWN,
 								"cv tag is valid for ARC-Seal only");
 						state = DKIM_STATE_ERROR;
-						break;
 					}
 					else {
 						param = DKIM_PARAM_CV;
 					}
 				}
 				else {
-					g_set_error (err,
-						DKIM_ERROR,
-						DKIM_SIGERROR_UNKNOWN,
-						"invalid dkim param: %c%c",
-						tag[0],
-						tag[1]);
-					state = DKIM_STATE_ERROR;
+					param = DKIM_PARAM_UNKNOWN;
+					msg_debug_dkim("unknown DKIM param %*s, ignoring it", taglen, tag);
 				}
 				break;
 			default:
-				g_set_error (err,
-					DKIM_ERROR,
-					DKIM_SIGERROR_UNKNOWN,
-					"invalid dkim param length: %zd",
-					taglen);
-				state = DKIM_STATE_ERROR;
+				/* Long and unknown (yet) DKIM tag */
+				param = DKIM_PARAM_UNKNOWN;
+				msg_debug_dkim("unknown DKIM param %*s, ignoring it", taglen, tag);
 				break;
 			}
+
 			if (state != DKIM_STATE_ERROR) {
 				/* Skip spaces */
 				state = DKIM_STATE_SKIP_SPACES;
@@ -1035,8 +1034,7 @@ rspamd_create_dkim_context (const gchar *sig,
 			break;
 		case DKIM_STATE_VALUE:
 			if (*p == ';') {
-				if (param == DKIM_PARAM_UNKNOWN ||
-					p - c == 0) {
+				if (p - c == 0 || c > p) {
 					state = DKIM_STATE_ERROR;
 				}
 				else {
@@ -1052,10 +1050,21 @@ rspamd_create_dkim_context (const gchar *sig,
 						tmp --;
 					}
 
-					if (!parser_funcs[param](ctx, c, tlen, err)) {
-						state = DKIM_STATE_ERROR;
+					if (param != DKIM_PARAM_UNKNOWN) {
+						if (!parser_funcs[param](ctx, c, tlen, err)) {
+							state = DKIM_STATE_ERROR;
+						}
+						else {
+							state = DKIM_STATE_SKIP_SPACES;
+							next_state = DKIM_STATE_TAG;
+							p++;
+							taglen = 0;
+						}
 					}
 					else {
+						/* Unknown param has been ignored */
+						msg_debug_dkim("ignored unknown tag parameter value: %*s = %*s",
+								taglen, tag, tlen, c);
 						state = DKIM_STATE_SKIP_SPACES;
 						next_state = DKIM_STATE_TAG;
 						p++;
@@ -1064,38 +1073,41 @@ rspamd_create_dkim_context (const gchar *sig,
 				}
 			}
 			else if (p == end) {
-				if (param == DKIM_PARAM_UNKNOWN) {
-					state = DKIM_STATE_ERROR;
-				}
-				else {
-					gint tlen = p - c;
-					const gchar *tmp = p - 1;
+				/* Last parameter with no `;` character */
+				gint tlen = p - c;
+				const gchar *tmp = p - 1;
 
-					while (tlen > 0) {
-						if (!g_ascii_isspace (*tmp)) {
-							break;
-						}
-						tlen --;
-						tmp --;
+				while (tlen > 0) {
+					if (!g_ascii_isspace (*tmp)) {
+						break;
 					}
+					tlen --;
+					tmp --;
+				}
 
+				if (param != DKIM_PARAM_UNKNOWN) {
 					if (!parser_funcs[param](ctx, c, tlen, err)) {
 						state = DKIM_STATE_ERROR;
 					}
-					if (state == DKIM_STATE_ERROR) {
-						/*
-						 * We need to return from here as state machine won't
-						 * do any more steps after p == end
-						 */
-						if (err) {
-							msg_info_dkim ("dkim parse failed: %e", *err);
-						}
-
-						return NULL;
-					}
-					/* Finish processing */
-					p++;
 				}
+				else {
+					msg_debug_dkim("ignored unknown tag parameter value: %*s: %*s",
+							taglen, tag, tlen, c);
+				}
+
+				if (state == DKIM_STATE_ERROR) {
+					/*
+					 * We need to return from here as state machine won't
+					 * do any more steps after p == end
+					 */
+					if (err) {
+						msg_info_dkim ("dkim parse failed: %e", *err);
+					}
+
+					return NULL;
+				}
+				/* Finish processing */
+				p++;
 			}
 			else {
 				p++;
@@ -1117,7 +1129,7 @@ rspamd_create_dkim_context (const gchar *sig,
 			}
 			else {
 				msg_info_dkim ("dkim parse failed: unknown error when parsing %c tag",
-						*tag);
+						tag ? *tag : '?');
 				return NULL;
 			}
 			break;
@@ -1254,11 +1266,11 @@ rspamd_create_dkim_context (const gchar *sig,
 	}
 
 	/* Now create dns key to request further */
-	taglen = strlen (ctx->domain) + strlen (ctx->selector) +
+	gsize dnslen = strlen (ctx->domain) + strlen (ctx->selector) +
 		sizeof (DKIM_DNSKEYNAME) + 2;
-	ctx->dns_key = rspamd_mempool_alloc (ctx->pool, taglen);
+	ctx->dns_key = rspamd_mempool_alloc (ctx->pool, dnslen);
 	rspamd_snprintf (ctx->dns_key,
-		taglen,
+			dnslen,
 		"%s.%s.%s",
 		ctx->selector,
 		DKIM_DNSKEYNAME,
@@ -1332,11 +1344,26 @@ rspamd_dkim_make_key (const gchar *keydata,
 	key = g_malloc0 (sizeof (rspamd_dkim_key_t));
 	REF_INIT_RETAIN (key, rspamd_dkim_key_free);
 	key->keydata = g_malloc0 (keylen + 1);
+	key->raw_key = g_malloc (keylen);
 	key->decoded_len = keylen;
-	key->keylen = keylen;
 	key->type = type;
 
-	if (!rspamd_cryptobox_base64_decode (keydata, keylen, key->keydata,
+	/* Copy key skipping all spaces and newlines */
+	const char *h = keydata;
+	guint8 *t = key->raw_key;
+
+	while (h - keydata < keylen) {
+		if (!g_ascii_isspace(*h)) {
+			*t++ = *h++;
+		}
+		else {
+			h++;
+		}
+	}
+
+	key->keylen = t - key->raw_key;
+
+	if (!rspamd_cryptobox_base64_decode (key->raw_key, key->keylen, key->keydata,
 			&key->decoded_len)) {
 		REF_RELEASE (key);
 		g_set_error (err,
@@ -1371,8 +1398,9 @@ rspamd_dkim_make_key (const gchar *keydata,
 			g_set_error (err,
 					DKIM_ERROR,
 					DKIM_SIGERROR_KEYFAIL,
-					"DKIM key is has invalid length %d for eddsa",
-					(gint)key->decoded_len);
+					"DKIM key is has invalid length %d for eddsa; expected %d",
+					(gint)key->decoded_len,
+					rspamd_cryptobox_pk_sig_bytes (RSPAMD_CRYPTOBOX_MODE_25519));
 			REF_RELEASE (key);
 
 			return NULL;
@@ -1469,6 +1497,7 @@ rspamd_dkim_key_free (rspamd_dkim_key_t *key)
 		BIO_free (key->key_bio);
 	}
 
+	g_free (key->raw_key);
 	g_free (key->keydata);
 	g_free (key);
 }
@@ -1502,10 +1531,13 @@ rspamd_dkim_parse_key (const gchar *txt, gsize *keylen, GError **err)
 	const gchar *c, *p, *end, *key = NULL, *alg = "rsa";
 	enum {
 		read_tag = 0,
+		read_tag_before_eqsign,
 		read_eqsign,
 		read_p_tag,
 		read_k_tag,
-	} state = read_tag;
+		ignore_value,
+		skip_spaces,
+	} state = read_tag, next_state;
 	gchar tag = '\0';
 	gsize klen = 0, alglen = 0;
 
@@ -1518,23 +1550,52 @@ rspamd_dkim_parse_key (const gchar *txt, gsize *keylen, GError **err)
 		case read_tag:
 			if (*p == '=') {
 				state = read_eqsign;
-			} else {
+			}
+			else if (g_ascii_isspace (*p)) {
+				state = skip_spaces;
+
+				if (tag != '\0') {
+					/* We had tag letter */
+					next_state = read_tag_before_eqsign;
+				}
+				else {
+					/* We had no tag letter, so we ignore empty tag */
+					next_state = read_tag;
+				}
+			}
+			else {
 				tag = *p;
 			}
 			p++;
 			break;
-		case read_eqsign:
-			if (tag == 'p') {
-				state = read_p_tag;
-				c = p;
-			} else if (tag == 'k') {
-				state = read_k_tag;
-				c = p;
-			} else {
-				/* Unknown tag, ignore */
+		case read_tag_before_eqsign:
+			/* Input: spaces before eqsign
+			 * Output: either read a next tag (previous had no value), or read value
+			 * p is moved forward
+			 */
+			if (*p == '=') {
+				state = read_eqsign;
+			}
+			else {
+				tag = *p;
 				state = read_tag;
+			}
+			p ++;
+			break;
+		case read_eqsign:
+			/* Always switch to skip spaces state and do not advance p */
+			state = skip_spaces;
+
+			if (tag == 'p') {
+				next_state = read_p_tag;
+			}
+			else if (tag == 'k') {
+				next_state = read_k_tag;
+			}
+			else {
+				/* Unknown tag, ignore */
+				next_state = ignore_value;
 				tag = '\0';
-				p++;
 			}
 			break;
 		case read_p_tag:
@@ -1543,8 +1604,11 @@ rspamd_dkim_parse_key (const gchar *txt, gsize *keylen, GError **err)
 				key = c;
 				state = read_tag;
 				tag = '\0';
+				p++;
+			}
+			else {
+				p ++;
 			}
-			p++;
 			break;
 		case read_k_tag:
 			if (*p == ';') {
@@ -1552,8 +1616,43 @@ rspamd_dkim_parse_key (const gchar *txt, gsize *keylen, GError **err)
 				alg = c;
 				state = read_tag;
 				tag = '\0';
+				p++;
+			}
+			else if (g_ascii_isspace (*p)) {
+				alglen = p - c;
+				alg = c;
+				state = skip_spaces;
+				next_state = read_tag;
+				tag = '\0';
+			}
+			else {
+				p ++;
+			}
+			break;
+		case ignore_value:
+			if (*p == ';') {
+				state = read_tag;
+				tag = '\0';
+				p ++;
+			}
+			else if (g_ascii_isspace (*p)) {
+				state = skip_spaces;
+				next_state = read_tag;
+				tag = '\0';
+			}
+			else {
+				p ++;
+			}
+			break;
+		case skip_spaces:
+			/* Skip spaces and switch to the next state if needed */
+			if (g_ascii_isspace(*p)) {
+				p ++;
+			}
+			else {
+				c = p;
+				state = next_state;
 			}
-			p++;
 			break;
 		default:
 			break;
@@ -1700,7 +1799,7 @@ rspamd_dkim_relaxed_body_step (struct rspamd_dkim_common_ctx *ctx, EVP_MD_CTX *c
 	gchar *t;
 	guint len, inlen;
 	gssize octets_remain;
-	gboolean got_sp;
+	gboolean got_sp, ret = TRUE;
 	gchar buf[1024];
 
 	len = size;
@@ -1710,7 +1809,7 @@ rspamd_dkim_relaxed_body_step (struct rspamd_dkim_common_ctx *ctx, EVP_MD_CTX *c
 	got_sp = FALSE;
 	octets_remain = *remain;
 
-	while (len > 0 && inlen > 0 && (octets_remain != 0)) {
+	while (len > 0 && inlen > 0 && (octets_remain > 0)) {
 
 		if (*h == '\r' || *h == '\n') {
 			if (got_sp) {
@@ -1765,6 +1864,16 @@ rspamd_dkim_relaxed_body_step (struct rspamd_dkim_common_ctx *ctx, EVP_MD_CTX *c
 		octets_remain --;
 	}
 
+	if (octets_remain < 0) {
+		/* Absurdic l tag value, but we still need to rewind the t pointer back */
+		while (t > buf && octets_remain < 0) {
+			t --;
+			octets_remain ++;
+		}
+
+		ret = FALSE;
+	}
+
 	*start = h;
 
 	if (t - buf > 0) {
@@ -1776,10 +1885,9 @@ rspamd_dkim_relaxed_body_step (struct rspamd_dkim_common_ctx *ctx, EVP_MD_CTX *c
 				"(%z size, %z -> %z remain)",
 						cklen, *remain, octets_remain);
 		*remain = octets_remain;
-
 	}
 
-	return ((len != 0) && (octets_remain != 0));
+	return ret && ((len > 0) && (octets_remain > 0));
 }
 
 static gboolean
@@ -2027,7 +2135,7 @@ rspamd_dkim_canonize_body (struct rspamd_dkim_common_ctx *ctx,
 	gboolean sign)
 {
 	const gchar *p;
-	gssize remain = ctx->len ? ctx->len : -1;
+	gssize remain = ctx->len ? ctx->len : G_MAXSSIZE;
 	guint total_len = end - start;
 	gboolean need_crlf = FALSE;
 
@@ -2317,15 +2425,21 @@ rspamd_dkim_canonize_header_relaxed (struct rspamd_dkim_common_ctx *ctx,
 
 static gboolean
 rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
-	struct rspamd_task *task,
-	const gchar *header_name,
-	gint count,
-	const gchar *dkim_header,
-	const gchar *dkim_domain)
+							 struct rspamd_task *task,
+							 const gchar *header_name,
+							 gint count,
+							 const gchar *dkim_header,
+							 const gchar *dkim_domain)
 {
 	struct rspamd_mime_header *rh, *cur, *sel = NULL;
 	gint hdr_cnt = 0;
-	bool use_idx = false;
+	bool use_idx = false, is_sign = ctx->is_sign;
+
+	/*
+	 * TODO:
+	 * Temporary hack to prevent linked list being misused until refactored
+	 */
+	const guint max_list_iters = 1000;
 
 	if (count < 0) {
 		use_idx = true;
@@ -2333,7 +2447,8 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 	}
 
 	if (dkim_header == NULL) {
-		rh = rspamd_message_get_header_array (task, header_name);
+		rh = rspamd_message_get_header_array (task, header_name,
+				is_sign);
 
 		if (rh) {
 			/* Check uniqueness of the header but we count from the bottom to top */
@@ -2345,7 +2460,7 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 
 					hdr_cnt++;
 
-					if (cur == rh) {
+					if (cur == rh || hdr_cnt >= max_list_iters) {
 						/* Cycle */
 						break;
 					}
@@ -2375,13 +2490,17 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 				}
 			}
 			else {
+				/*
+				 * This branch is used for ARC headers, and it orders them based on
+				 * i=<number> string and not their real order in the list of headers
+				 */
 				gchar idx_buf[16];
-				gint id_len;
+				gint id_len, i;
 
 				id_len = rspamd_snprintf (idx_buf, sizeof (idx_buf), "i=%d;",
 						count);
 
-				for (cur = rh->prev; ; cur = cur->prev) {
+				for (cur = rh->prev, i = 0; i < max_list_iters; cur = cur->prev, i ++) {
 					if (cur->decoded &&
 						rspamd_substring_search (cur->decoded, strlen (cur->decoded),
 								idx_buf, id_len) != -1) {
@@ -2412,7 +2531,7 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 						count, (gint)sel->raw_len, sel->raw_value);
 			}
 			else {
-				if (ctx->is_sign && (sel->flags & RSPAMD_HEADER_FROM)) {
+				if (is_sign && (sel->flags & RSPAMD_HEADER_FROM)) {
 					/* Special handling of the From handling when rewrite is done */
 					gboolean has_rewrite = FALSE;
 					guint i;
@@ -2450,7 +2569,7 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 		/* For signature check just use the saved dkim header */
 		if (ctx->header_canon_type == DKIM_CANON_SIMPLE) {
 			/* We need to find our own signature and use it */
-			rh = rspamd_message_get_header_array (task, header_name);
+			rh = rspamd_message_get_header_array (task, header_name, is_sign);
 
 			if (rh) {
 				/* We need to find our own signature */
@@ -2628,7 +2747,8 @@ rspamd_dkim_check (rspamd_dkim_context_t *ctx,
 	}
 
 
-	if (ctx->common.type != RSPAMD_DKIM_ARC_SEAL) {
+	/* Use cached BH for all but arc seal, if it is not NULL we are not in arc seal mode */
+	if (cached_bh != NULL) {
 		if (!cached_bh->digest_normal) {
 			/* Copy md_ctx to deal with broken CRLF at the end */
 			cpy_ctx = EVP_MD_CTX_create ();
@@ -2772,6 +2892,7 @@ rspamd_dkim_check (rspamd_dkim_context_t *ctx,
 		if (RSA_verify (nid, raw_digest, dlen, ctx->b, ctx->blen,
 				key->key.key_rsa) != 1) {
 			msg_debug_dkim ("headers rsa verify failed");
+			ERR_clear_error ();
 			res->rcode = DKIM_REJECT;
 			res->fail_reason = "headers rsa verify failed";
 
@@ -2799,6 +2920,7 @@ rspamd_dkim_check (rspamd_dkim_context_t *ctx,
 					RSPAMD_DKIM_KEY_ID_LEN, rspamd_dkim_key_id (key),
 					ctx->dkim_header);
 			msg_debug_dkim ("headers ecdsa verify failed");
+			ERR_clear_error ();
 			res->rcode = DKIM_REJECT;
 			res->fail_reason = "headers ecdsa verify failed";
 		}
@@ -3286,7 +3408,7 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 			/* Do oversigning */
 			guint count = 0;
 
-			rh = rspamd_message_get_header_array (task, dh->name);
+			rh = rspamd_message_get_header_array(task, dh->name, FALSE);
 
 			if (rh) {
 				DL_FOREACH (rh, cur) {
@@ -3314,7 +3436,7 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 			}
 		}
 		else {
-			rh = rspamd_message_get_header_array (task, dh->name);
+			rh = rspamd_message_get_header_array(task, dh->name, FALSE);
 
 			if (rh) {
 				if (hstat.s.count > 0) {
