@@ -89,6 +89,8 @@ rspamd_logger_t * rspamd_log_open_specific (rspamd_mempool_t *pool,
  * @param level
  */
 void rspamd_log_set_log_level (rspamd_logger_t *logger, gint level);
+gint rspamd_log_get_log_level (rspamd_logger_t *logger);
+const gchar *rspamd_get_log_severity_string(gint level_flags);
 /**
  * Set log flags (from enum rspamd_log_flags)
  * @param logger
@@ -148,19 +150,20 @@ bool rspamd_common_logv (rspamd_logger_t *logger, gint level_flags,
  * @param mod
  * @return
  */
-guint rspamd_logger_add_debug_module (const gchar *mod);
+gint rspamd_logger_add_debug_module (const gchar *mod);
 
 /*
  * Macro to use for faster debug modules
  */
 #define INIT_LOG_MODULE(mname) \
-    static guint rspamd_##mname##_log_id = (guint)-1; \
+    static gint rspamd_##mname##_log_id = -1; \
     RSPAMD_CONSTRUCTOR(rspamd_##mname##_log_init) { \
         rspamd_##mname##_log_id = rspamd_logger_add_debug_module(#mname); \
 }
 
+
 #define INIT_LOG_MODULE_PUBLIC(mname) \
-    guint rspamd_##mname##_log_id = (guint)-1; \
+    gint rspamd_##mname##_log_id = -1; \
     RSPAMD_CONSTRUCTOR(rspamd_##mname##_log_init) { \
         rspamd_##mname##_log_id = rspamd_logger_add_debug_module(#mname); \
 }
@@ -176,14 +179,17 @@ bool rspamd_conditional_debug (rspamd_logger_t *logger,
 
 bool rspamd_conditional_debug_fast (rspamd_logger_t *logger,
 									rspamd_inet_addr_t *addr,
-									guint mod_id,
+									gint mod_id,
 									const gchar *module, const gchar *id,
 									const gchar *function, const gchar *fmt, ...);
 bool rspamd_conditional_debug_fast_num_id (rspamd_logger_t *logger,
 									rspamd_inet_addr_t *addr,
-									guint mod_id,
+									gint mod_id,
 									const gchar *module, guint64 id,
 									const gchar *function, const gchar *fmt, ...);
+gboolean rspamd_logger_need_log (rspamd_logger_t *rspamd_log,
+								 GLogLevelFlags log_level,
+								 gint module_id);
 
 /**
  * Function with variable number of arguments support that uses static default logger
@@ -245,128 +251,145 @@ struct rspamd_logger_funcs* rspamd_logger_set_log_function (rspamd_logger_t *log
 /* Typical functions */
 
 extern guint rspamd_task_log_id;
+#ifdef __cplusplus
+#define RSPAMD_LOG_FUNC __func__
+#else
+#define RSPAMD_LOG_FUNC G_STRFUNC
+#endif
 
 /* Logging in postfix style */
 #define msg_err(...)    rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         NULL, NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         NULL, NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         NULL, NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_notice(...)   rspamd_default_log_function (G_LOG_LEVEL_MESSAGE, \
         NULL, NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
         NULL, NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 #define debug_task(...) rspamd_conditional_debug_fast (NULL, \
         task->from_addr, \
         rspamd_task_log_id, "task", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 /* Use the following macros if you have `task` in the function */
 #define msg_err_task(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         task->task_pool->tag.tagname, task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
+        __VA_ARGS__)
+#define msg_err_task_lambda(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
+        task->task_pool->tag.tagname, task->task_pool->tag.uid, \
+        log_func, \
         __VA_ARGS__)
 #define msg_warn_task(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         task->task_pool->tag.tagname, task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_notice_task(...)   rspamd_default_log_function (G_LOG_LEVEL_MESSAGE, \
         task->task_pool->tag.tagname, task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_task(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         task->task_pool->tag.tagname, task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
+        __VA_ARGS__)
+#define msg_info_task_lambda(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
+        task->task_pool->tag.tagname, task->task_pool->tag.uid, \
+        log_func, \
         __VA_ARGS__)
 #define msg_debug_task(...)  rspamd_conditional_debug_fast (NULL,  task->from_addr, \
         rspamd_task_log_id, "task", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
+        __VA_ARGS__)
+#define msg_debug_task_lambda(...)  rspamd_conditional_debug_fast (NULL,  task->from_addr, \
+        rspamd_task_log_id, "task", task->task_pool->tag.uid, \
+        log_func, \
         __VA_ARGS__)
 #define msg_err_task_encrypted(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL|RSPAMD_LOG_ENCRYPTED, \
         task->task_pool->tag.tagname, task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_task_encrypted(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING|RSPAMD_LOG_ENCRYPTED, \
         task->task_pool->tag.tagname, task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_notice_task_encrypted(...) rspamd_default_log_function (G_LOG_LEVEL_MESSAGE|RSPAMD_LOG_ENCRYPTED, \
         task->task_pool->tag.tagname, task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_task_encrypted(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO|RSPAMD_LOG_ENCRYPTED, \
         task->task_pool->tag.tagname, task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 /* Check for NULL pointer first */
 #define msg_err_task_check(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         task ? task->task_pool->tag.tagname : NULL, task ? task->task_pool->tag.uid : NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_task_check(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         task ? task->task_pool->tag.tagname : NULL, task ? task->task_pool->tag.uid : NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_task_check(...)   rspamd_default_log_function (G_LOG_LEVEL_MESSAGE, \
         task ? task->task_pool->tag.tagname : NULL, task ? task->task_pool->tag.uid : NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_notice_task_check(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         task ? task->task_pool->tag.tagname : NULL, task ? task->task_pool->tag.uid : NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug_task_check(...)  rspamd_conditional_debug_fast (NULL, \
         task ? task->from_addr : NULL, \
         rspamd_task_log_id, "task", task ? task->task_pool->tag.uid : NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 /* Use the following macros if you have `pool` in the function */
 #define msg_err_pool(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         pool->tag.tagname, pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_pool(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         pool->tag.tagname, pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_pool(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         pool->tag.tagname, pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug_pool(...)  rspamd_conditional_debug (NULL, NULL, \
         pool->tag.tagname, pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 /* Check for NULL pointer first */
 #define msg_err_pool_check(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         pool ? pool->tag.tagname : NULL, pool ? pool->tag.uid : NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_pool_check(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         pool ? pool->tag.tagname : NULL, pool ? pool->tag.uid : NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_pool_check(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         pool ? pool->tag.tagname : NULL, pool ? pool->tag.uid : NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug_pool_check(...)  rspamd_conditional_debug (NULL, NULL, \
         pool ? pool->tag.tagname : NULL, pool ? pool->tag.uid : NULL, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 #ifdef  __cplusplus
