@@ -23,10 +23,15 @@
 #include "libserver/http/http_connection.h"
 #include "libserver/http/http_private.h"
 #include "rspamd.h"
-#include "contrib/zstd/zstd.h"
 #include "contrib/libev/ev.h"
 #include "contrib/uthash/utlist.h"
 
+#ifdef SYS_ZSTD
+#  include "zstd.h"
+#else
+#  include "contrib/zstd/zstd.h"
+#endif
+
 #undef MAP_DEBUG_REFS
 #ifdef MAP_DEBUG_REFS
 #define MAP_RETAIN(x, t) do { \
@@ -86,11 +91,6 @@ write_http_request (struct http_callback_data *cbd)
 	struct rspamd_http_message *msg;
 
 	msg = rspamd_http_new_message (HTTP_REQUEST);
-
-	if (cbd->bk->protocol == MAP_PROTO_HTTPS) {
-		msg->flags |= RSPAMD_HTTP_FLAG_SSL;
-	}
-
 	if (cbd->check) {
 		msg->method = HTTP_HEAD;
 	}
@@ -243,8 +243,9 @@ rspamd_map_cache_cb (struct ev_loop *loop, ev_timer *w, int revents)
 		 * Important!: we do not set cache availability to zero here, as there
 		 * might be fresh cache
 		 */
-		msg_info_map ("cached data is now expired (gen mismatch %L != %L) for %s",
-				cache_cbd->gen, cache_cbd->data->gen, map->name);
+		msg_info_map ("cached data is now expired (gen mismatch %L != %L) for %s; shm name=%s; refcount=%d",
+				cache_cbd->gen, cache_cbd->data->gen, map->name, cache_cbd->shm->shm_name,
+				cache_cbd->shm->ref.refcount);
 		MAP_RELEASE (cache_cbd->shm, "rspamd_http_map_cached_cbdata");
 		ev_timer_stop (loop, &cache_cbd->timeout);
 		g_free (cache_cbd);
@@ -264,10 +265,13 @@ rspamd_map_cache_cb (struct ev_loop *loop, ev_timer *w, int revents)
 		}
 
 		if (w->repeat < 0) {
-			msg_info_map ("cached data for %s has skewed check time: %d last checked, %d poll timeout, %.2f diff",
+			msg_info_map ("cached data for %s has skewed check time: %d last checked, "
+						  "%d poll timeout, %.2f diff; shm name=%s; refcount=%d",
 					map->name, (int)cache_cbd->data->last_checked,
 					(int)cache_cbd->map->poll_timeout,
-					(rspamd_get_calendar_ticks () - cache_cbd->data->last_checked));
+					(rspamd_get_calendar_ticks () - cache_cbd->data->last_checked),
+					cache_cbd->shm->shm_name,
+					cache_cbd->shm->ref.refcount);
 			w->repeat = 0.0;
 		}
 
@@ -278,8 +282,11 @@ rspamd_map_cache_cb (struct ev_loop *loop, ev_timer *w, int revents)
 	else {
 		data->cur_cache_cbd = NULL;
 		g_atomic_int_set (&data->cache->available, 0);
+		msg_info_map ("cached data is now expired for %s; shm name=%s; refcount=%d",
+				map->name,
+				cache_cbd->shm->shm_name,
+				cache_cbd->shm->ref.refcount);
 		MAP_RELEASE (cache_cbd->shm, "rspamd_http_map_cached_cbdata");
-		msg_info_map ("cached data is now expired for %s", map->name);
 		ev_timer_stop (loop, &cache_cbd->timeout);
 		g_free (cache_cbd);
 	}
@@ -366,6 +373,10 @@ http_map_finish (struct rspamd_http_connection *conn,
 
 				map->next_check = hdate;
 			}
+			else {
+				msg_info_map ("invalid expires header: %T, ignore it", expires_hdr);
+				map->next_check = 0;
+			}
 		}
 
 		/* Check for etag */
@@ -407,6 +418,8 @@ http_map_finish (struct rspamd_http_connection *conn,
 		cache_cbd->last_checked = cbd->data->last_checked;
 		cache_cbd->gen = cbd->data->gen;
 		MAP_RETAIN (cache_cbd->shm, "shmem_data");
+		msg_info_map ("stored map data in a shared memory cache: %s",
+				cache_cbd->shm->shm_name);
 
 		ev_timer_init (&cache_cbd->timeout, rspamd_map_cache_cb, cached_timeout,
 				0.0);
@@ -514,6 +527,10 @@ http_map_finish (struct rspamd_http_connection *conn,
 			if (hdate != (time_t)-1 && hdate > msg->date) {
 				map->next_check = hdate;
 			}
+			else {
+				msg_info_map ("invalid expires header: %T, ignore it", expires_hdr);
+				map->next_check = 0;
+			}
 		}
 
 		etag_hdr = rspamd_http_message_find_header (msg, "ETag");
@@ -531,8 +548,8 @@ http_map_finish (struct rspamd_http_connection *conn,
 			rspamd_http_date_format (next_check_date, sizeof (next_check_date),
 					map->next_check);
 			msg_info_map ("data is not modified for server %s, next check at %s "
-						  "(http cache based)",
-					cbd->data->host, next_check_date);
+						  "(http cache based: %T)",
+					cbd->data->host, next_check_date, expires_hdr);
 		}
 		else {
 			rspamd_http_date_format (next_check_date, sizeof (next_check_date),
@@ -583,6 +600,8 @@ read_map_file_chunks (struct rspamd_map *map, struct map_cb_data *cbdata,
 	if (lseek (fd, off, SEEK_SET) == -1) {
 		msg_err_map ("can't seek in map to pos %d for buffered reading %s: %s",
 				(gint)off, fname, strerror (errno));
+		close (fd);
+
 		return FALSE;
 	}
 
@@ -973,8 +992,8 @@ rspamd_map_periodic_dtor (struct map_periodic_cbdata *periodic)
 	map = periodic->map;
 	msg_debug_map ("periodic dtor %p", periodic);
 
-	if (periodic->need_modify) {
-		/* We are done */
+	if (periodic->need_modify || periodic->cbdata.errored) {
+		/* Need to notify the real data structure */
 		periodic->map->fin_callback (&periodic->cbdata, periodic->map->user_data);
 	}
 	else {
@@ -1036,6 +1055,7 @@ rspamd_map_schedule_periodic (struct rspamd_map *map, int how)
 
 	if (map->non_trivial && map->next_check != 0) {
 		timeout = map->next_check - rspamd_get_calendar_ticks ();
+		map->next_check = 0;
 
 		if (timeout > 0 && timeout < map->poll_timeout) {
 			/* Early check case, jitter */
@@ -1058,8 +1078,15 @@ rspamd_map_schedule_periodic (struct rspamd_map *map, int how)
 		}
 		else if (timeout <= 0) {
 			/* Data is already expired, need to check */
-			jittered_sec = 0.0;
-			reason = "expired non-trivial data";
+			if (how & RSPAMD_MAP_SCHEDULE_ERROR) {
+				/* In case of error we still need to increase delay */
+				jittered_sec = map->poll_timeout * error_mult;
+				reason = "expired non-trivial data (after error)";
+			}
+			else {
+				jittered_sec = 0.0;
+				reason = "expired non-trivial data";
+			}
 		}
 		else {
 			/* No need to check now, wait till next_check */
@@ -1068,6 +1095,7 @@ rspamd_map_schedule_periodic (struct rspamd_map *map, int how)
 		}
 	}
 	else {
+		/* No valid information when to check a map, plan a timer based check */
 		timeout = map->poll_timeout;
 
 		if (how & RSPAMD_MAP_SCHEDULE_INIT) {
@@ -1100,7 +1128,7 @@ rspamd_map_schedule_periodic (struct rspamd_map *map, int how)
 
 	/* Now, we do some sanity checks for jittered seconds */
 	if (!(how & RSPAMD_MAP_SCHEDULE_INIT)) {
-		/* Never allow too low interval between timer checks, it is epxensive */
+		/* Never allow too low interval between timer checks, it is expensive */
 		if (jittered_sec < min_timer_interval) {
 			jittered_sec = rspamd_time_jitter (min_timer_interval, 0);
 		}
@@ -1119,7 +1147,6 @@ rspamd_map_schedule_periodic (struct rspamd_map *map, int how)
 	}
 
 	cbd = g_malloc0 (sizeof (*cbd));
-	cbd->cbdata.state = 0;
 	cbd->cbdata.prev_data = *map->user_data;
 	cbd->cbdata.cur_data = NULL;
 	cbd->cbdata.map = map;
@@ -1190,7 +1217,7 @@ rspamd_map_dns_callback (struct rdns_reply *reply, void *arg)
 	if (reply->code == RDNS_RC_NOERROR) {
 		DL_FOREACH (reply->entries, cur_rep) {
 			rspamd_inet_addr_t *addr;
-			addr = rspamd_inet_address_from_rnds (reply->entries);
+			addr = rspamd_inet_address_from_rnds (cur_rep);
 
 			if (addr != NULL) {
 				rspamd_inet_address_set_port (addr, cbd->data->port);
@@ -1244,6 +1271,9 @@ rspamd_map_dns_callback (struct rdns_reply *reply, void *arg)
 retry:
 		msg_debug_map ("try open http connection to %s",
 				rspamd_inet_address_to_string_pretty (cbd->addr));
+		if (cbd->bk->protocol == MAP_PROTO_HTTPS) {
+			flags |= RSPAMD_HTTP_CLIENT_SSL;
+		}
 		cbd->conn = rspamd_http_connection_new_client (NULL,
 				NULL,
 				http_map_error,
@@ -1289,13 +1319,13 @@ static gboolean
 rspamd_map_read_cached (struct rspamd_map *map, struct rspamd_map_backend *bk,
 		struct map_periodic_cbdata *periodic, const gchar *host)
 {
-	gsize len;
+	gsize mmap_len, len;
 	gpointer in;
 	struct http_map_data *data;
 
 	data = bk->data.hd;
 
-	in = rspamd_shmem_xmap (data->cache->shmem_name, PROT_READ, &len);
+	in = rspamd_shmem_xmap (data->cache->shmem_name, PROT_READ, &mmap_len);
 
 	if (in == NULL) {
 		msg_err ("cannot map cache from %s: %s", data->cache->shmem_name,
@@ -1303,15 +1333,24 @@ rspamd_map_read_cached (struct rspamd_map *map, struct rspamd_map_backend *bk,
 		return FALSE;
 	}
 
-	if (len < data->cache->len) {
-		msg_err ("cannot map cache from %s: bad length %z, %z expected",
+	if (mmap_len < data->cache->len) {
+		msg_err ("cannot map cache from %s: truncated length %z, %z expected",
 				data->cache->shmem_name,
-				len, data->cache->len);
-		munmap (in, len);
+				mmap_len, data->cache->len);
+		munmap (in, mmap_len);
 
 		return FALSE;
 	}
 
+	/*
+	 * Len is taken from the shmem file size that can be larger than the
+	 * actual data length, as we use shared memory as a growing buffer for the
+	 * HTTP input.
+	 * Hence, we need to use len from the saved cache data, counting that it is
+	 * at least not more than the cached file length (this is checked above).
+	 */
+	len = data->cache->len;
+
 	if (bk->is_compressed) {
 		ZSTD_DStream *zstream;
 		ZSTD_inBuffer zin;
@@ -1345,7 +1384,7 @@ rspamd_map_read_cached (struct rspamd_map *map, struct rspamd_map_backend *bk,
 						ZSTD_getErrorName (r));
 				ZSTD_freeDStream (zstream);
 				g_free (out);
-				munmap (in, len);
+				munmap (in, mmap_len);
 				return FALSE;
 			}
 
@@ -1365,12 +1404,11 @@ rspamd_map_read_cached (struct rspamd_map *map, struct rspamd_map_backend *bk,
 		g_free (out);
 	}
 	else {
-		msg_info_map ("%s: read map data cached %z bytes", bk->uri,
-				len);
+		msg_info_map ("%s: read map data cached %z bytes", bk->uri, len);
 		map->read_callback (in, len, &periodic->cbdata, TRUE);
 	}
 
-	munmap (in, len);
+	munmap (in, mmap_len);
 
 	return TRUE;
 }
@@ -1610,7 +1648,15 @@ rspamd_map_read_http_cached_file (struct rspamd_map *map,
 		return FALSE;
 	}
 
-	map->next_check = header.next_check;
+	double now = rspamd_get_calendar_ticks ();
+
+	if (header.next_check > now) {
+		map->next_check = header.next_check;
+	}
+	else {
+		map->next_check = now;
+	}
+
 	htdata->last_modified = header.mtime;
 
 	if (header.etag_len > 0) {
@@ -1760,6 +1806,11 @@ check:
 			strlen (data->host), RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
 		rspamd_inet_address_set_port (addr, cbd->data->port);
 		g_ptr_array_add (cbd->addrs, (void *)addr);
+
+		if (bk->protocol == MAP_PROTO_HTTPS) {
+			flags |= RSPAMD_HTTP_CLIENT_SSL;
+		}
+
 		cbd->conn = rspamd_http_connection_new_client (
 				NULL,
 				NULL,
@@ -1965,7 +2016,7 @@ rspamd_map_process_periodic (struct map_periodic_cbdata *cbd)
 	}
 
 	if (cbd->errored) {
-		/* We should not check other backends if some backend has failed */
+		/* We should not check other backends if some backend has failed*/
 		rspamd_map_schedule_periodic (cbd->map, RSPAMD_MAP_SCHEDULE_ERROR);
 
 		if (cbd->locked) {
@@ -1973,6 +2024,9 @@ rspamd_map_process_periodic (struct map_periodic_cbdata *cbd)
 			cbd->locked = FALSE;
 		}
 
+		/* Also set error flag for the map consumer */
+		cbd->cbdata.errored = true;
+
 		msg_debug_map ("unlocked map %s, refcount=%d", cbd->map->name,
 				cbd->ref.refcount);
 		MAP_RELEASE (cbd, "periodic");
@@ -2123,6 +2177,15 @@ rspamd_map_watch (struct rspamd_config *cfg,
 
 				data = bk->data.fd;
 
+				if (map->user_data == NULL || *map->user_data == NULL) {
+					/* Map has not been read, init it's reading if possible */
+					struct stat st;
+
+					if (stat (data->filename, &st) != -1) {
+						data->need_modify = TRUE;
+					}
+				}
+
 				ev_stat_init (&data->st_ev, rspamd_map_on_stat,
 						data->filename, map->poll_timeout * cfg->map_file_watch_multiplier);
 				data->st_ev.data = map;
@@ -2432,8 +2495,6 @@ rspamd_map_is_map (const gchar *map_line)
 static void
 rspamd_map_backend_dtor (struct rspamd_map_backend *bk)
 {
-	g_free (bk->uri);
-
 	switch (bk->protocol) {
 	case MAP_PROTO_FILE:
 		if (bk->data.fd) {
@@ -2468,17 +2529,25 @@ rspamd_map_backend_dtor (struct rspamd_map_backend *bk)
 				rspamd_fstring_free (data->etag);
 			}
 
-			if (g_atomic_int_compare_and_exchange (&data->cache->available, 1, 0)) {
-				if (data->cur_cache_cbd) {
-					MAP_RELEASE (data->cur_cache_cbd->shm,
-							"rspamd_http_map_cached_cbdata");
-					ev_timer_stop (data->cur_cache_cbd->event_loop,
-							&data->cur_cache_cbd->timeout);
-					g_free (data->cur_cache_cbd);
-					data->cur_cache_cbd = NULL;
+			/*
+			 * Clear cached file, but check if a worker is an active http worker
+			 * as cur_cache_cbd is meaningful merely for active worker, who actually
+			 * owns the cache
+			 */
+			if (bk->map && bk->map->active_http) {
+				if (g_atomic_int_compare_and_exchange (&data->cache->available, 1, 0)) {
+					if (data->cur_cache_cbd) {
+						msg_info ("clear shared memory cache for a map in %s as backend \"%s\" is closing",
+								data->cur_cache_cbd->shm->shm_name,
+								bk->uri);
+						MAP_RELEASE (data->cur_cache_cbd->shm,
+								"rspamd_http_map_cached_cbdata");
+						ev_timer_stop (data->cur_cache_cbd->event_loop,
+								&data->cur_cache_cbd->timeout);
+						g_free (data->cur_cache_cbd);
+						data->cur_cache_cbd = NULL;
+					}
 				}
-
-				unlink (data->cache->shmem_name);
 			}
 
 			g_free (bk->data.hd);
@@ -2490,6 +2559,7 @@ rspamd_map_backend_dtor (struct rspamd_map_backend *bk)
 		rspamd_pubkey_unref (bk->trusted_pubkey);
 	}
 
+	g_free (bk->uri);
 	g_free (bk);
 }
 
@@ -2629,6 +2699,14 @@ err:
 		g_free (hdata);
 	}
 
+	if (fdata) {
+		g_free (fdata);
+	}
+
+	if (sdata) {
+		g_free (sdata);
+	}
+
 	return NULL;
 }
 
@@ -2737,6 +2815,7 @@ rspamd_map_add (struct rspamd_config *cfg,
 
 	rspamd_map_calculate_hash (map);
 	msg_info_map ("added map %s", bk->uri);
+	bk->map = map;
 
 	cfg->maps = g_list_prepend (cfg->maps, map);
 
@@ -2777,6 +2856,8 @@ rspamd_map_add_backend (struct rspamd_map *map, struct rspamd_map_backend *bk)
 	else {
 		g_ptr_array_add (map->backends, bk);
 	}
+
+	bk->map = map;
 }
 
 struct rspamd_map*
