@@ -32,19 +32,19 @@ typedef void (*rspamd_map_tmp_dtor) (gpointer p);
 extern guint rspamd_map_log_id;
 #define msg_err_map(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         "map", map->tag, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_map(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         "map", map->tag, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_map(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         "map", map->tag, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug_map(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_map_log_id, "map", map->tag, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 enum fetch_proto {
