@@ -32,19 +32,19 @@
 
 #define msg_err_milter(...) rspamd_default_log_function(G_LOG_LEVEL_CRITICAL, \
         "milter", priv->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_milter(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         "milter", priv->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_milter(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         "milter", priv->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug_milter(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_milter_log_id, "milter", priv->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(milter)
@@ -1652,17 +1652,21 @@ rspamd_milter_remove_header_safe (struct rspamd_milter_session *session,
 		hname = g_string_new (key);
 		hvalue = g_string_new ("");
 
-		if (nhdr >= 1) {
-			rspamd_milter_send_action (session,
-					RSPAMD_MILTER_CHGHEADER,
-					nhdr, hname, hvalue);
+		if (nhdr > 0) {
+			if (ar->len >= nhdr) {
+				rspamd_milter_send_action (session,
+						RSPAMD_MILTER_CHGHEADER,
+						nhdr, hname, hvalue);
+				priv->cur_hdr --;
+			}
 		}
-		else if (nhdr == 0 && ar->len > 0) {
+		else if (nhdr == 0) {
 			/* We need to clear all headers */
 			for (i = ar->len; i > 0; i --) {
 				rspamd_milter_send_action (session,
 						RSPAMD_MILTER_CHGHEADER,
 						i, hname, hvalue);
+				priv->cur_hdr --;
 			}
 		}
 		else {
@@ -1671,11 +1675,17 @@ rspamd_milter_remove_header_safe (struct rspamd_milter_session *session,
 				rspamd_milter_send_action (session,
 						RSPAMD_MILTER_CHGHEADER,
 						ar->len + nhdr + 1, hname, hvalue);
+				priv->cur_hdr --;
 			}
 		}
 
 		g_string_free (hname, TRUE);
 		g_string_free (hvalue, TRUE);
+
+		if (priv->cur_hdr < 0) {
+			msg_err_milter("negative header count after removing %s", key);
+			priv->cur_hdr = 0;
+		}
 	}
 }
 
@@ -1697,7 +1707,7 @@ rspamd_milter_extract_single_header (struct rspamd_milter_session *session,
 		idx_obj = ucl_object_lookup_any (obj, "order",
 				"index", NULL);
 
-		if (idx_obj) {
+		if (idx_obj && (ucl_object_type (idx_obj) == UCL_INT || ucl_object_type (idx_obj) == UCL_FLOAT)) {
 			idx = ucl_object_toint (idx_obj);
 			has_idx = TRUE;
 		}
@@ -1715,10 +1725,19 @@ rspamd_milter_extract_single_header (struct rspamd_milter_session *session,
 			else {
 				/* Calculate negative offset */
 
-				if (-idx <= priv->cur_hdr) {
+				if (idx == -1) {
+					rspamd_milter_send_action (session,
+							RSPAMD_MILTER_ADDHEADER,
+							hname, hvalue);
+				}
+				else if (-idx <= priv->cur_hdr) {
+					/*
+					 * Note: We should account MTA's own "Received:" field
+					 * which wasn't passed by Milter's header command.
+					 */
 					rspamd_milter_send_action (session,
 							RSPAMD_MILTER_INSHEADER,
-							priv->cur_hdr + idx + 1,
+							priv->cur_hdr + idx + 2,
 							hname, hvalue);
 				}
 				else {
@@ -1735,6 +1754,8 @@ rspamd_milter_extract_single_header (struct rspamd_milter_session *session,
 					hname, hvalue);
 		}
 
+		priv->cur_hdr ++;
+
 		g_string_free (hname, TRUE);
 		g_string_free (hvalue, TRUE);
 	}
@@ -1747,7 +1768,7 @@ static gboolean
 rspamd_milter_process_milter_block (struct rspamd_milter_session *session,
 		const ucl_object_t *obj, struct rspamd_action *action)
 {
-	const ucl_object_t *elt, *cur, *cur_elt;
+	const ucl_object_t *elt, *cur;
 	ucl_object_iter_t it;
 	struct rspamd_milter_private *priv = session->priv;
 	GString *hname, *hvalue;
@@ -1781,43 +1802,61 @@ rspamd_milter_process_milter_block (struct rspamd_milter_session *session,
 			it = NULL;
 
 			while ((cur = ucl_object_iterate (elt, &it, true)) != NULL) {
-				ucl_object_iter_t *elt_it;
 
-				elt_it = ucl_object_iterate_new (cur);
+				const char *key_name = ucl_object_key (cur);
 
-				while ((cur_elt = ucl_object_iterate_safe (elt_it, true)) != NULL) {
-					if (ucl_object_type (cur_elt) == UCL_STRING) {
-						hname = g_string_new (ucl_object_key (cur));
-						hvalue = g_string_new (ucl_object_tostring (cur_elt));
+				if (ucl_object_type (cur) == UCL_STRING) {
+					/*
+					 * Legacy support of {"name": "value", ... } with
+					 * multiple names under the same name
+					 */
+					ucl_object_iter_t *elt_it;
+					const ucl_object_t *cur_elt;
 
-						rspamd_milter_send_action (session,
-								RSPAMD_MILTER_ADDHEADER,
-								hname, hvalue);
-						g_string_free (hname, TRUE);
-						g_string_free (hvalue, TRUE);
+					elt_it = ucl_object_iterate_new (cur);
+					while ((cur_elt = ucl_object_iterate_safe (elt_it, false)) != NULL) {
+						if (ucl_object_type (cur_elt) == UCL_STRING) {
+							hname = g_string_new (key_name);
+							hvalue = g_string_new (ucl_object_tostring (cur_elt));
+
+							rspamd_milter_send_action (session,
+									RSPAMD_MILTER_ADDHEADER,
+									hname, hvalue);
+							g_string_free (hname, TRUE);
+							g_string_free (hvalue, TRUE);
+						}
+						else {
+							msg_warn_milter("legacy header with name %s, that has not a string value: %s",
+									key_name, ucl_object_type_to_string(cur_elt->type));
+						}
 					}
-					else if (ucl_object_type (cur_elt) == UCL_OBJECT) {
+					ucl_object_iterate_free (elt_it);
+				}
+				else {
+					if (ucl_object_type (cur) == UCL_OBJECT) {
 						rspamd_milter_extract_single_header (session,
-								ucl_object_key (cur), cur_elt);
+								key_name, cur);
 					}
-					else if (ucl_object_type (cur_elt) == UCL_ARRAY) {
+					else if (ucl_object_type (cur) == UCL_ARRAY) {
 						/* Multiple values for the same key */
 						ucl_object_iter_t *array_it;
 						const ucl_object_t *array_elt;
 
-						array_it = ucl_object_iterate_new (cur_elt);
+						array_it = ucl_object_iterate_new (cur);
 
 						while ((array_elt = ucl_object_iterate_safe (array_it,
 								true)) != NULL) {
 							rspamd_milter_extract_single_header (session,
-									ucl_object_key (cur), array_elt);
+									key_name, array_elt);
 						}
 
 						ucl_object_iterate_free (array_it);
 					}
+					else {
+						msg_warn_milter("non-legacy header with name %s, that has unsupported value type: %s",
+								key_name, ucl_object_type_to_string(cur->type));
+					}
 				}
-
-				ucl_object_iterate_free (elt_it);
 			}
 		}
 
