@@ -30,6 +30,7 @@
 
 #ifdef WITH_HYPERSCAN
 #include "hs.h"
+#include "hyperscan_tools.h"
 #endif
 
 #include "unix-std.h"
@@ -52,24 +53,24 @@
 
 #define msg_err_re_cache(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         "re_cache", cache->hash, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_re_cache(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         "re_cache", cache->hash, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_re_cache(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         "re_cache", cache->hash, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 #define msg_debug_re_task(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_re_cache_log_id, "re_cache", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug_re_cache(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_re_cache_log_id, "re_cache", cache->hash, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(re_cache)
@@ -93,7 +94,7 @@ struct rspamd_re_class {
 	gchar hash[rspamd_cryptobox_HASHBYTES + 1];
 
 #ifdef WITH_HYPERSCAN
-	hs_database_t *hs_db;
+	rspamd_hyperscan_t *hs_db;
 	hs_scratch_t *hs_scratch;
 	gint *hs_ids;
 	guint nhs;
@@ -127,7 +128,6 @@ struct rspamd_re_cache {
 #ifdef WITH_HYPERSCAN
 	enum rspamd_hyperscan_status hyperscan_loaded;
 	gboolean disable_hyperscan;
-	gboolean vectorized_hyperscan;
 	hs_platform_info_t plt;
 #endif
 };
@@ -196,7 +196,7 @@ rspamd_re_cache_destroy (struct rspamd_re_cache *cache)
 
 #ifdef WITH_HYPERSCAN
 		if (re_class->hs_db) {
-			hs_free_database (re_class->hs_db);
+			rspamd_hyperscan_free(re_class->hs_db);
 		}
 		if (re_class->hs_scratch) {
 			hs_free_scratch (re_class->hs_scratch);
@@ -482,7 +482,6 @@ rspamd_re_cache_init (struct rspamd_re_cache *cache, struct rspamd_config *cfg)
 	rspamd_fstring_t *features = rspamd_fstring_new ();
 
 	cache->disable_hyperscan = cfg->disable_hyperscan;
-	cache->vectorized_hyperscan = cfg->vectorized_hyperscan;
 
 	g_assert (hs_populate_platform (&cache->plt) == HS_SUCCESS);
 
@@ -803,37 +802,18 @@ rspamd_re_cache_process_regexp_data (struct rspamd_re_runtime *rt,
 		g_assert (re_class->hs_db != NULL);
 
 		/* Go through hyperscan API */
-		if (!rt->cache->vectorized_hyperscan) {
-			for (i = 0; i < count; i++) {
-				cbdata.ins = &in[i];
-				cbdata.re = re;
-				cbdata.rt = rt;
-				cbdata.lens = &lens[i];
-				cbdata.count = 1;
-				cbdata.task = task;
-
-				if ((hs_scan (re_class->hs_db, in[i], lens[i], 0,
-						re_class->hs_scratch,
-						rspamd_re_cache_hyperscan_cb, &cbdata)) != HS_SUCCESS) {
-					ret = 0;
-				}
-				else {
-					ret = rt->results[re_id];
-					*processed_hyperscan = TRUE;
-				}
-			}
-		}
-		else {
-			cbdata.ins = in;
+		for (i = 0; i < count; i++) {
+			cbdata.ins = &in[i];
 			cbdata.re = re;
 			cbdata.rt = rt;
-			cbdata.lens = lens;
+			cbdata.lens = &lens[i];
 			cbdata.count = 1;
 			cbdata.task = task;
 
-			if ((hs_scan_vector (re_class->hs_db, (const char **)in, lens, count, 0,
-					re_class->hs_scratch,
-					rspamd_re_cache_hyperscan_cb, &cbdata)) != HS_SUCCESS) {
+			if ((hs_scan (rspamd_hyperscan_get_database(re_class->hs_db),
+				in[i], lens[i], 0,
+				re_class->hs_scratch,
+				rspamd_re_cache_hyperscan_cb, &cbdata)) != HS_SUCCESS) {
 				ret = 0;
 			}
 			else {
@@ -1178,6 +1158,12 @@ rspamd_re_cache_exec_re (struct rspamd_task *task,
 		break;
 	case RSPAMD_RE_MIMEHEADER:
 		PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, parts), i, mime_part) {
+			if (mime_part->parent_part == NULL ||
+				!IS_PART_MULTIPART(mime_part->parent_part) ||
+				IS_PART_MESSAGE(mime_part)) {
+				/* We filter parts that have no multipart parent or are a messages here */
+				continue;
+			}
 			rh = rspamd_message_get_header_from_hash(mime_part->raw_headers,
 					re_class->type_data, FALSE);
 
@@ -1787,7 +1773,7 @@ rspamd_re_cache_is_finite (struct rspamd_re_cache *cache,
 
 		if (hs_compile (pat,
 				flags | HS_FLAG_PREFILTER,
-				cache->vectorized_hyperscan ? HS_MODE_VECTORED : HS_MODE_BLOCK,
+				HS_MODE_BLOCK,
 				&cache->plt,
 				&test_db,
 				&hs_errors) != HS_SUCCESS) {
@@ -1920,7 +1906,7 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 	rspamd_snprintf (path, sizeof (path), "%s%c%s.hs", cbdata->cache_dir,
 			G_DIR_SEPARATOR, re_class->hash);
 
-	if (rspamd_re_cache_is_valid_hyperscan_file (cache, path, TRUE, TRUE)) {
+	if (rspamd_re_cache_is_valid_hyperscan_file (cache, path, TRUE, TRUE, NULL)) {
 
 		fd = open (path, O_RDONLY, 00600);
 
@@ -2021,7 +2007,7 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 
 		if (hs_compile (pat,
 				hs_flags[i],
-				cache->vectorized_hyperscan ? HS_MODE_VECTORED : HS_MODE_BLOCK,
+				HS_MODE_BLOCK,
 				&cache->plt,
 				&test_db,
 				&hs_errors) != HS_SUCCESS) {
@@ -2076,7 +2062,7 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 				hs_ids,
 				hs_exts,
 				n,
-				cache->vectorized_hyperscan ? HS_MODE_VECTORED : HS_MODE_BLOCK,
+				HS_MODE_BLOCK,
 				&cache->plt,
 				&test_db,
 				&hs_errors) != HS_SUCCESS) {
@@ -2124,13 +2110,8 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 				hs_serialized, serialized_len);
 		crc = rspamd_cryptobox_fast_hash_final (&crc_st);
 
-		if (cache->vectorized_hyperscan) {
-			iov[0].iov_base = (void *) rspamd_hs_magic_vector;
-		}
-		else {
-			iov[0].iov_base = (void *) rspamd_hs_magic;
-		}
 
+		iov[0].iov_base = (void *) rspamd_hs_magic;
 		iov[0].iov_len = RSPAMD_HS_MAGIC_LEN;
 		iov[1].iov_base = &cache->plt;
 		iov[1].iov_len = sizeof (cache->plt);
@@ -2260,7 +2241,7 @@ rspamd_re_cache_compile_hyperscan (struct rspamd_re_cache *cache,
 
 gboolean
 rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
-		const char *path, gboolean silent, gboolean try_load)
+		const char *path, gboolean silent, gboolean try_load, GError **err)
 {
 	g_assert (cache != NULL);
 	g_assert (path != NULL);
@@ -2289,6 +2270,9 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 			msg_err_re_cache ("cannot open hyperscan cache file %s: too short filename",
 					path);
 		}
+		g_set_error(err, rspamd_re_cache_quark(), 0,
+				"too short filename");
+
 		return FALSE;
 	}
 
@@ -2297,6 +2281,8 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 			msg_err_re_cache ("cannot open hyperscan cache file %s: not ending with .hs",
 					path);
 		}
+		g_set_error(err, rspamd_re_cache_quark(), 0,
+				"not ending with .hs");
 		return FALSE;
 	}
 
@@ -2317,6 +2303,9 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 					msg_err_re_cache ("cannot open hyperscan cache file %s: %s",
 							path, strerror (errno));
 				}
+				g_set_error(err, rspamd_re_cache_quark(), 0,
+						"%s",
+						strerror (errno));
 				return FALSE;
 			}
 
@@ -2325,22 +2314,24 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 					msg_err_re_cache ("cannot read magic from hyperscan "
 									  "cache file %s: %s",
 							path, strerror (errno));
+					g_set_error(err, rspamd_re_cache_quark(), 0,
+							"cannot read magic: %s",
+							strerror (errno));
 				}
 				else {
 					msg_err_re_cache ("truncated read magic from hyperscan "
 									  "cache file %s: %z, %z wanted",
 							path, r, (gsize)sizeof (magicbuf));
+					g_set_error(err, rspamd_re_cache_quark(), 0,
+							"truncated read magic %zd, %zd wanted",
+							r, (gsize)sizeof (magicbuf));
 				}
+
 				close (fd);
 				return FALSE;
 			}
 
-			if (cache->vectorized_hyperscan) {
-				mb = rspamd_hs_magic_vector;
-			}
-			else {
-				mb = rspamd_hs_magic;
-			}
+			mb = rspamd_hs_magic;
 
 			if (memcmp (magicbuf, mb, sizeof (magicbuf)) != 0) {
 				msg_err_re_cache ("cannot open hyperscan cache file %s: "
@@ -2349,6 +2340,7 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 						(int) RSPAMD_HS_MAGIC_LEN, mb);
 
 				close (fd);
+				g_set_error(err, rspamd_re_cache_quark(), 0, "invalid magic");
 				return FALSE;
 			}
 
@@ -2364,14 +2356,19 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 							path, r, (gsize)sizeof (magicbuf));
 				}
 
+				g_set_error(err, rspamd_re_cache_quark(), 0,
+						"cannot read platform data: %s", strerror (errno));
+
 				close (fd);
 				return FALSE;
 			}
 
-			if (memcmp (&test_plt, &cache->plt, sizeof (test_plt)) != 0) {
+			if (test_plt.cpu_features != cache->plt.cpu_features) {
 				msg_err_re_cache ("cannot open hyperscan cache file %s: "
 						"compiled for a different platform",
 						path);
+				g_set_error(err, rspamd_re_cache_quark(), 0,
+						"compiled for a different platform");
 
 				close (fd);
 				return FALSE;
@@ -2386,6 +2383,8 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 					msg_err_re_cache ("cannot mmap hyperscan cache file %s: "
 							"%s",
 							path, strerror (errno));
+					g_set_error(err, rspamd_re_cache_quark(), 0,
+							"mmap error: %s", strerror(errno));
 					return FALSE;
 				}
 
@@ -2401,6 +2400,8 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 					/* Some wrong amount of regexps */
 					msg_err_re_cache ("bad number of expressions in %s: %d",
 							path, n);
+					g_set_error(err, rspamd_re_cache_quark(), 0,
+							"bad number of expressions: %d", n);
 					munmap (map, len);
 					return FALSE;
 				}
@@ -2430,6 +2431,8 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 				if (crc != valid_crc) {
 					msg_warn_re_cache ("outdated or invalid hs database in %s: "
 							"crc read %xL, crc expected %xL", path, crc, valid_crc);
+					g_set_error(err, rspamd_re_cache_quark(), 0,
+							"outdated or invalid hs database, crc check failure");
 					munmap (map, len);
 
 					return FALSE;
@@ -2438,6 +2441,8 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 				if ((ret = hs_deserialize_database (p, end - p, &test_db))
 						!= HS_SUCCESS) {
 					msg_err_re_cache ("bad hs database in %s: %d", path, ret);
+					g_set_error(err, rspamd_re_cache_quark(), 0,
+							"deserialize error: %d", ret);
 					munmap (map, len);
 
 					return FALSE;
@@ -2456,6 +2461,9 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 		msg_warn_re_cache ("unknown hyperscan cache file %s", path);
 	}
 
+	g_set_error(err, rspamd_re_cache_quark(), 0,
+			"unknown hyperscan file");
+
 	return FALSE;
 #endif
 }
@@ -2475,7 +2483,7 @@ rspamd_re_cache_load_hyperscan (struct rspamd_re_cache *cache,
 	gint fd, i, n, *hs_ids = NULL, *hs_flags = NULL, total = 0, ret;
 	GHashTableIter it;
 	gpointer k, v;
-	guint8 *map, *p, *end;
+	guint8 *map, *p;
 	struct rspamd_re_class *re_class;
 	struct rspamd_re_cache_elt *elt;
 	struct stat st;
@@ -2488,7 +2496,7 @@ rspamd_re_cache_load_hyperscan (struct rspamd_re_cache *cache,
 		rspamd_snprintf (path, sizeof (path), "%s%c%s.hs", cache_dir,
 				G_DIR_SEPARATOR, re_class->hash);
 
-		if (rspamd_re_cache_is_valid_hyperscan_file (cache, path, try_load, FALSE)) {
+		if (rspamd_re_cache_is_valid_hyperscan_file (cache, path, try_load, FALSE, NULL)) {
 			msg_debug_re_cache ("load hyperscan database from '%s'",
 					re_class->hash);
 
@@ -2514,7 +2522,6 @@ rspamd_re_cache_load_hyperscan (struct rspamd_re_cache *cache,
 			}
 
 			close (fd);
-			end = map + st.st_size;
 			p = map + RSPAMD_HS_MAGIC_LEN + sizeof (cache->plt);
 			n = *(gint *)p;
 
@@ -2554,7 +2561,7 @@ rspamd_re_cache_load_hyperscan (struct rspamd_re_cache *cache,
 			}
 
 			if (re_class->hs_db != NULL) {
-				hs_free_database (re_class->hs_db);
+				rspamd_hyperscan_free (re_class->hs_db);
 			}
 
 			if (re_class->hs_ids) {
@@ -2564,16 +2571,16 @@ rspamd_re_cache_load_hyperscan (struct rspamd_re_cache *cache,
 			re_class->hs_ids = NULL;
 			re_class->hs_scratch = NULL;
 			re_class->hs_db = NULL;
+			munmap (map, st.st_size);
 
-			if ((ret = hs_deserialize_database (p, end - p, &re_class->hs_db))
-					!= HS_SUCCESS) {
+			re_class->hs_db = rspamd_hyperscan_maybe_load(path, p - map);
+			if (re_class->hs_db == NULL) {
 				if (!try_load) {
-					msg_err_re_cache ("bad hs database in %s: %d", path, ret);
+					msg_err_re_cache ("bad hs database in %s", path);
 				}
 				else {
-					msg_debug_re_cache ("bad hs database in %s: %d", path, ret);
+					msg_debug_re_cache ("bad hs database in %s", path);
 				}
-				munmap (map, st.st_size);
 				g_free (hs_ids);
 				g_free (hs_flags);
 
@@ -2585,10 +2592,11 @@ rspamd_re_cache_load_hyperscan (struct rspamd_re_cache *cache,
 				continue;
 			}
 
-			munmap (map, st.st_size);
-
-			g_assert (hs_alloc_scratch (re_class->hs_db,
-					&re_class->hs_scratch) == HS_SUCCESS);
+			if ((ret = hs_alloc_scratch (rspamd_hyperscan_get_database(re_class->hs_db),
+					&re_class->hs_scratch)) != HS_SUCCESS) {
+				msg_err_re_cache ("fatal error: cannot allocate scratch for %s: %d", path, ret);
+				g_abort();
+			}
 
 			/*
 			 * Now find hyperscan elts that are successfully compiled and
