@@ -189,7 +189,8 @@ gint rspamd_re_cache_compile_hyperscan (struct rspamd_re_cache *cache,
 gboolean rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 												  const char *path,
 												  gboolean silent,
-												  gboolean try_load);
+												  gboolean try_load,
+												  GError **err);
 
 /**
  * Loads all hyperscan regexps precompiled
