@@ -100,23 +100,23 @@ struct rspamd_spf_library_ctx *spf_lib_ctx = NULL;
 
 #define msg_err_spf(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         "spf", rec->task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_spf(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         "spf", rec->task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_spf(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         "spf", rec->task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug_spf(...)  rspamd_conditional_debug_fast (NULL, rec->task->from_addr, \
         rspamd_spf_log_id, "spf", rec->task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_debug_spf_flatten(...)  rspamd_conditional_debug_fast_num_id (NULL, NULL, \
         rspamd_spf_log_id, "spf", (flat)->digest, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(spf)
@@ -207,12 +207,6 @@ spf_library_config (const ucl_object_t *obj)
 		}
 	}
 
-	if ((value = ucl_object_find_key (obj, "disable_ipv6")) != NULL) {
-		if (ucl_object_toboolean_safe (value, &bval)) {
-			spf_lib_ctx->disable_ipv6 = bval;
-		}
-	}
-
 	if (spf_lib_ctx->spf_hash) {
 		rspamd_lru_hash_destroy (spf_lib_ctx->spf_hash);
 		spf_lib_ctx->spf_hash = NULL;
@@ -882,6 +876,11 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 				rec->ttl = elt_data->ttl;
 			}
 
+			if (elt_data->type == RDNS_REQUEST_CNAME) {
+				/* Skip cname aliases - it must be handled by a recursor */
+				continue;
+			}
+
 			switch (cb->cur_action) {
 				case SPF_RESOLVE_MX:
 					if (elt_data->type == RDNS_REQUEST_MX) {
@@ -975,14 +974,19 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 					break;
 				case SPF_RESOLVE_INCLUDE:
 					if (elt_data->type == RDNS_REQUEST_TXT) {
-						if (reply->entries) {
-							msg_debug_spf ("got include record for %s: '%s'",
-									req_name->name,
-									reply->entries[0].content.txt.data);
-						}
+						struct rdns_reply_entry *selected = NULL;
 
 						cb->addr->flags |= RSPAMD_SPF_FLAG_RESOLVED;
-						spf_process_txt_record (rec, cb->resolved, reply, NULL);
+						spf_process_txt_record (rec, cb->resolved, reply, &selected);
+						if (selected) {
+							msg_debug_spf ("got include record for %s: '%s'",
+									req_name->name,
+									selected->content.txt.data);
+						}
+						else {
+							msg_debug_spf ("no include record for %s",
+									req_name->name);
+						}
 					}
 					goto end;
 
