@@ -70,7 +70,7 @@ struct rspamd_ssl_connection {
 
 #define msg_debug_ssl(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_ssl_log_id, "ssl", conn->log_tag, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 static void rspamd_ssl_event_handler (gint fd, short what, gpointer ud);
@@ -356,7 +356,7 @@ rspamd_ssl_peer_verify (struct rspamd_ssl_connection *c)
 	ver_err = SSL_get_verify_result (c->ssl);
 
 	if (ver_err != X509_V_OK) {
-		g_set_error (&err, rspamd_ssl_quark (), ver_err, "certificate validation "
+		g_set_error (&err, rspamd_ssl_quark (), 400, "certificate validation "
 				"failed: %s", X509_verify_cert_error_string (ver_err));
 		c->err_handler (c->handler_data, err);
 		g_error_free (err);
@@ -367,7 +367,7 @@ rspamd_ssl_peer_verify (struct rspamd_ssl_connection *c)
 	/* Get server's certificate */
 	server_cert =  SSL_get_peer_certificate (c->ssl);
 	if (server_cert == NULL) {
-		g_set_error (&err, rspamd_ssl_quark (), ver_err, "peer certificate is absent");
+		g_set_error (&err, rspamd_ssl_quark (), 401, "peer certificate is absent");
 		c->err_handler (c->handler_data, err);
 		g_error_free (err);
 
@@ -377,7 +377,7 @@ rspamd_ssl_peer_verify (struct rspamd_ssl_connection *c)
 	if (c->hostname) {
 		if (!rspamd_tls_check_name (server_cert, c->hostname)) {
 			X509_free (server_cert);
-			g_set_error (&err, rspamd_ssl_quark (), ver_err, "peer certificate fails "
+			g_set_error (&err, rspamd_ssl_quark (), 403, "peer certificate fails "
 					"hostname verification for %s", c->hostname);
 			c->err_handler (c->handler_data, err);
 			g_error_free (err);
@@ -396,22 +396,21 @@ rspamd_tls_set_error (gint retcode, const gchar *stage, GError **err)
 {
 	GString *reason;
 	gchar buf[120];
-	gint err_code = 0, last_err = 0;
+	gint err_code = 0;
 
 	reason = g_string_sized_new (sizeof (buf));
 
 	if (retcode == SSL_ERROR_SYSCALL) {
 		rspamd_printf_gstring (reason, "syscall fail: %s", strerror (errno));
-		err_code = errno;
+		err_code = 500;
 	}
 	else {
 		while ((err_code = ERR_get_error()) != 0) {
-			last_err = err_code;
 			ERR_error_string (err_code, buf);
 			rspamd_printf_gstring (reason, "ssl error: %s,", buf);
 		}
 
-		err_code = last_err;
+		err_code = 400;
 
 		if (reason->len > 0 && reason->str[reason->len - 1] == ',') {
 			reason->str[reason->len - 1] = '\0';
@@ -435,6 +434,13 @@ rspamd_ssl_connection_dtor (struct rspamd_ssl_connection *conn)
 		g_free (conn->hostname);
 	}
 
+	/*
+	 * Try to workaround for the race between timeout and ssl error
+	 */
+	if (conn->shut_ev != conn->ev && ev_can_stop (&conn->ev->tm)) {
+		rspamd_ev_watcher_stop (conn->event_loop, conn->ev);
+	}
+
 	if (conn->shut_ev) {
 		rspamd_ev_watcher_stop (conn->event_loop, conn->shut_ev);
 		g_free (conn->shut_ev);
@@ -544,7 +550,7 @@ rspamd_ssl_event_handler (gint fd, short what, gpointer ud)
 		else {
 			conn->shut = ssl_shut_unclean;
 			rspamd_ev_watcher_stop (conn->event_loop, conn->ev);
-			g_set_error (&err, rspamd_ssl_quark (), ETIMEDOUT,
+			g_set_error (&err, rspamd_ssl_quark (), 408,
 					"ssl connection timed out");
 			conn->err_handler (conn->handler_data, err);
 			g_error_free (err);
@@ -616,7 +622,7 @@ rspamd_ssl_event_handler (gint fd, short what, gpointer ud)
 		break;
 	default:
 		rspamd_ev_watcher_stop (conn->event_loop, conn->ev);
-		g_set_error (&err, rspamd_ssl_quark (), EINVAL,
+		g_set_error (&err, rspamd_ssl_quark (), 500,
 				"ssl bad state error: %d", conn->state);
 		conn->err_handler (conn->handler_data, err);
 		g_error_free (err);
@@ -638,7 +644,7 @@ rspamd_ssl_connection_new (gpointer ssl_ctx, struct ev_loop *ev_base,
 	conn->verify_peer = verify_peer;
 
 	if (log_tag) {
-		rspamd_strlcpy (conn->log_tag, log_tag, sizeof (log_tag));
+		rspamd_strlcpy (conn->log_tag, log_tag, sizeof (conn->log_tag));
 	}
 	else {
 		rspamd_random_hex (conn->log_tag, sizeof (log_tag) - 1);
@@ -660,6 +666,8 @@ rspamd_ssl_connect_fd (struct rspamd_ssl_connection *conn, gint fd,
 
 	g_assert (conn != NULL);
 
+	/* Ensure that we start from the empty SSL errors stack */
+	ERR_clear_error ();
 	conn->ssl = SSL_new (conn->ssl_ctx->s);
 
 	if (hostname) {
@@ -748,6 +756,22 @@ rspamd_ssl_connect_fd (struct rspamd_ssl_connection *conn, gint fd,
 	return TRUE;
 }
 
+void
+rspamd_ssl_connection_restore_handlers (struct rspamd_ssl_connection *conn,
+										rspamd_ssl_handler_t handler,
+										rspamd_ssl_error_handler_t err_handler,
+										gpointer handler_data,
+										short ev_what)
+{
+	conn->handler = handler;
+	conn->err_handler = err_handler;
+	conn->handler_data = handler_data;
+
+	rspamd_ev_watcher_stop (conn->event_loop, conn->ev);
+	rspamd_ev_watcher_init (conn->ev, conn->fd, ev_what, rspamd_ssl_event_handler, conn);
+	rspamd_ev_watcher_start (conn->event_loop, conn->ev, conn->ev->timeout);
+}
+
 gssize
 rspamd_ssl_read (struct rspamd_ssl_connection *conn, gpointer buf,
 		gsize buflen)
@@ -760,7 +784,7 @@ rspamd_ssl_read (struct rspamd_ssl_connection *conn, gpointer buf,
 
 	if (conn->state != ssl_conn_connected && conn->state != ssl_next_read) {
 		errno = EINVAL;
-		g_set_error (&err, rspamd_ssl_quark (), ECONNRESET,
+		g_set_error (&err, rspamd_ssl_quark (), 400,
 				"ssl state error: cannot read data");
 		conn->shut = ssl_shut_unclean;
 		conn->err_handler (conn->handler_data, err);
@@ -1047,7 +1071,7 @@ rspamd_openssl_maybe_init (void)
 		OpenSSL_add_all_digests ();
 		OpenSSL_add_all_ciphers ();
 
-#if OPENSSL_VERSION_NUMBER >= 0x1000104fL && !defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER >= 0x1000104fL && OPENSSL_VERSION_NUMBER < 0x30000000L && !defined(LIBRESSL_VERSION_NUMBER)
 		ENGINE_load_builtin_engines ();
 #endif
 #if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
