@@ -42,7 +42,9 @@ struct rspamd_stat_backend {
 					  struct rspamd_statfile *st);
 
 	gpointer (*runtime) (struct rspamd_task *task,
-						 struct rspamd_statfile_config *stcf, gboolean learn, gpointer ctx);
+						 struct rspamd_statfile_config *stcf,
+						 gboolean learn, gpointer ctx,
+						 gint id);
 
 	gboolean (*process_tokens) (struct rspamd_task *task, GPtrArray *tokens,
 								gint id,
@@ -81,16 +83,16 @@ struct rspamd_stat_backend {
             struct rspamd_config *cfg, struct rspamd_statfile *st); \
         gpointer rspamd_##name##_runtime (struct rspamd_task *task, \
                 struct rspamd_statfile_config *stcf, \
-                gboolean learn, gpointer ctx); \
+                gboolean learn, gpointer ctx, gint id); \
         gboolean rspamd_##name##_process_tokens (struct rspamd_task *task, \
                 GPtrArray *tokens, gint id, \
-                gpointer ctx); \
+                gpointer runtime); \
         gboolean rspamd_##name##_finalize_process (struct rspamd_task *task, \
                 gpointer runtime, \
                 gpointer ctx); \
         gboolean rspamd_##name##_learn_tokens (struct rspamd_task *task, \
                 GPtrArray *tokens, gint id, \
-                gpointer ctx); \
+                gpointer runtime); \
         gboolean rspamd_##name##_finalize_learn (struct rspamd_task *task, \
                 gpointer runtime, \
                 gpointer ctx, GError **err); \
@@ -116,6 +118,7 @@ RSPAMD_STAT_BACKEND_DEF(mmaped_file);
 RSPAMD_STAT_BACKEND_DEF(sqlite3);
 RSPAMD_STAT_BACKEND_DEF(cdb);
 RSPAMD_STAT_BACKEND_DEF(redis);
+RSPAMD_STAT_BACKEND_DEF(http);
 
 #ifdef  __cplusplus
 }
