@@ -25,7 +25,7 @@
 
 #define msg_debug_stat_redis(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_stat_redis_log_id, "stat_redis", task->task_pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(stat_redis)
@@ -1076,6 +1076,14 @@ rspamd_redis_async_stat_fin (struct rspamd_stat_async_elt *elt, gpointer d)
 		rspamd_redis_async_cbdata_cleanup (redis_elt->cbdata);
 		redis_elt->cbdata = NULL;
 	}
+
+	/* Clear the static elements */
+	if (redis_elt->stat) {
+		ucl_object_unref(redis_elt->stat);
+		redis_elt->stat = NULL;
+	}
+
+	g_free (redis_elt);
 }
 
 /* Called on connection termination */
@@ -1650,7 +1658,7 @@ rspamd_stat_redis_on_connect(const struct redisAsyncContext *ac, int status)
 gpointer
 rspamd_redis_runtime (struct rspamd_task *task,
 		struct rspamd_statfile_config *stcf,
-		gboolean learn, gpointer c)
+		gboolean learn, gpointer c, gint _id)
 {
 	struct redis_stat_ctx *ctx = REDIS_CTX (c);
 	struct redis_stat_runtime *rt;
