@@ -141,7 +141,11 @@ rspamd_stat_cache_redis_get (redisAsyncContext *c, gpointer r, gpointer priv)
 				rspamd_strtol (reply->str, reply->len, &val);
 			}
 			else {
-				if (reply->type != REDIS_REPLY_NIL) {
+				if (reply->type == REDIS_REPLY_ERROR) {
+					msg_err_task ("cannot learn %s: redis error: \"%s\"",
+							rt->ctx->stcf->symbol, reply->str);
+				}
+				else if (reply->type != REDIS_REPLY_NIL) {
 					msg_err_task ("bad learned type for %s: %d",
 							rt->ctx->stcf->symbol, reply->type);
 				}
@@ -445,13 +449,13 @@ rspamd_stat_cache_redis_check (struct rspamd_task *task,
 	gchar *h;
 
 	if (rspamd_session_blocked (task->s)) {
-		return RSPAMD_LEARN_INGORE;
+		return RSPAMD_LEARN_IGNORE;
 	}
 
 	h = rspamd_mempool_get_variable (task->task_pool, "words_hash");
 
 	if (h == NULL) {
-		return RSPAMD_LEARN_INGORE;
+		return RSPAMD_LEARN_IGNORE;
 	}
 
 	if (redisAsyncCommand (rt->redis, rspamd_stat_cache_redis_get, rt,
@@ -479,7 +483,7 @@ rspamd_stat_cache_redis_learn (struct rspamd_task *task,
 	gint flag;
 
 	if (rt == NULL || rt->ctx == NULL || rspamd_session_blocked (task->s)) {
-		return RSPAMD_LEARN_INGORE;
+		return RSPAMD_LEARN_IGNORE;
 	}
 
 	h = rspamd_mempool_get_variable (task->task_pool, "words_hash");
