@@ -60,7 +60,7 @@ void *rspamd_inet_library_get_lib_ctx (void);
 void rspamd_inet_library_destroy (void);
 
 /**
- * Create new inet address structure based on the address familiy and opaque init pointer
+ * Create new inet address structure based on the address family and opaque init pointer
  * @param af
  * @param init
  * @return new inet addr
@@ -322,7 +322,7 @@ gint rspamd_inet_address_compare_ptr (gconstpointer a1,
  * @param addr
  * @return
  */
-rspamd_inet_addr_t *rspamd_inet_address_copy (const rspamd_inet_addr_t *addr);
+rspamd_inet_addr_t *rspamd_inet_address_copy(const rspamd_inet_addr_t *addr, rspamd_mempool_t *pool);
 
 /**
  * Returns hash for inet address (ignoring port)
