@@ -23,6 +23,7 @@
 #include "logger.h"
 #include "unix-std.h"
 #include "hs.h"
+#include "libserver/hyperscan_tools.h"
 #endif
 #include "acism.h"
 #include "libutil/regexp.h"
@@ -43,7 +44,7 @@ static enum rspamd_hs_check_state hs_suitable_cpu = RSPAMD_HS_UNCHECKED;
 struct RSPAMD_ALIGNED(64) rspamd_multipattern {
 #ifdef WITH_HYPERSCAN
 	rspamd_cryptobox_hash_state_t hash_state;
-	hs_database_t *db;
+	rspamd_hyperscan_t *hs_db;
 	hs_scratch_t *scratch[MAX_SCRATCH];
 	GArray *hs_pats;
 	GArray *hs_ids;
@@ -400,8 +401,6 @@ rspamd_multipattern_try_load_hs (struct rspamd_multipattern *mp,
 		const guchar *hash)
 {
 	gchar fp[PATH_MAX];
-	gpointer map;
-	gsize len;
 
 	if (hs_cache_dir == NULL) {
 		return FALSE;
@@ -409,19 +408,9 @@ rspamd_multipattern_try_load_hs (struct rspamd_multipattern *mp,
 
 	rspamd_snprintf (fp, sizeof (fp), "%s/%*xs.hsmp", hs_cache_dir,
 			(gint)rspamd_cryptobox_HASHBYTES / 2, hash);
+	mp->hs_db = rspamd_hyperscan_maybe_load(fp, 0);
 
-	if ((map = rspamd_file_xmap (fp, PROT_READ, &len, TRUE)) != NULL) {
-		if (hs_deserialize_database (map, len, &mp->db) == HS_SUCCESS) {
-			munmap (map, len);
-			return TRUE;
-		}
-
-		munmap (map, len);
-		/* Remove stale file */
-		(void)unlink (fp);
-	}
-
-	return FALSE;
+	return mp->hs_db != NULL;
 }
 
 static void
@@ -441,7 +430,7 @@ rspamd_multipattern_try_save_hs (struct rspamd_multipattern *mp,
 			(gint)rspamd_cryptobox_HASHBYTES / 2, hash);
 
 	if ((fd = rspamd_file_xopen (fp, O_WRONLY | O_CREAT | O_EXCL, 00644, 0)) != -1) {
-		if (hs_serialize_database (mp->db, &bytes, &len) == HS_SUCCESS) {
+		if (hs_serialize_database (rspamd_hyperscan_get_database(mp->hs_db), &bytes, &len) == HS_SUCCESS) {
 			if (write (fd, bytes, len) == -1) {
 				msg_warn ("cannot write hyperscan cache to %s: %s",
 						fp, strerror (errno));
@@ -460,6 +449,9 @@ rspamd_multipattern_try_save_hs (struct rspamd_multipattern *mp,
 							fp, np, strerror (errno));
 					unlink (fp);
 				}
+				else {
+					rspamd_hyperscan_notice_known(np);
+				}
 			}
 		}
 		else {
@@ -493,13 +485,15 @@ rspamd_multipattern_compile (struct rspamd_multipattern *mp, GError **err)
 			rspamd_cryptobox_hash_final (&mp->hash_state, hash);
 
 			if (!rspamd_multipattern_try_load_hs (mp, hash)) {
+				hs_database_t *db = NULL;
+
 				if (hs_compile_multi ((const char *const *)mp->hs_pats->data,
 						(const unsigned int *)mp->hs_flags->data,
 						(const unsigned int *)mp->hs_ids->data,
 						mp->cnt,
 						HS_MODE_BLOCK,
 						&plt,
-						&mp->db,
+						&db,
 						&hs_errors) != HS_SUCCESS) {
 
 					g_set_error (err, rspamd_multipattern_quark (), EINVAL,
@@ -510,12 +504,17 @@ rspamd_multipattern_compile (struct rspamd_multipattern *mp, GError **err)
 
 					return FALSE;
 				}
+				mp->hs_db = rspamd_hyperscan_from_raw_db(db);
 			}
 
 			rspamd_multipattern_try_save_hs (mp, hash);
 
 			for (i = 0; i < MAX_SCRATCH; i ++) {
-				g_assert (hs_alloc_scratch (mp->db, &mp->scratch[i]) == HS_SUCCESS);
+				int ret;
+				if ((ret = hs_alloc_scratch (rspamd_hyperscan_get_database(mp->hs_db), &mp->scratch[i])) != HS_SUCCESS) {
+					msg_err("fatal error: cannot allocate scratch space for hyperscan: %d", ret);
+					g_abort();
+				}
 			}
 		}
 
@@ -652,7 +651,7 @@ rspamd_multipattern_lookup (struct rspamd_multipattern *mp,
 
 		g_assert (scr != NULL);
 
-		ret = hs_scan (mp->db, in, len, 0, scr,
+		ret = hs_scan (rspamd_hyperscan_get_database(mp->hs_db), in, len, 0, scr,
 				rspamd_multipattern_hs_cb, &cbd);
 
 		mp->scratch_used &= ~(1 << i);
@@ -728,7 +727,9 @@ rspamd_multipattern_destroy (struct rspamd_multipattern *mp)
 					hs_free_scratch (mp->scratch[i]);
 				}
 
-				hs_free_database (mp->db);
+				if (mp->hs_db) {
+					rspamd_hyperscan_free(mp->hs_db);
+				}
 			}
 
 			for (i = 0; i < mp->cnt; i ++) {
