@@ -349,14 +349,44 @@ rspamd_uint64_print (guint64 in, gchar *out)
 	return ndigits;
 }
 
+static inline int
+rspamd_ffsll(long long n)
+{
+#ifdef __has_builtin
+# if __has_builtin(__builtin_ffsll)
+	return __builtin_ffsll(n);
+# elif __has_builtin(__builtin_ctzll)
+	if (n == 0) {
+		return 0;
+	}
+
+	return __builtin_ctzll(n) + 1;
+# endif
+#endif /* __has_builtin */
+
+#ifdef HAVE_FFSL
+	return ffsl(n);
+#else
+	if (n == 0) {
+		return 0;
+	}
+
+	int bit;
+	for (bit = 1; !(n & 1); bit++) {
+		n = ((unsigned long long) n) >> 1;
+	}
+	return bit;
+#endif
+}
+
 static gchar *
 rspamd_sprintf_num (gchar *buf, gchar *last, guint64 ui64, gchar zero,
-					  guint hexadecimal, guint width)
+					  guint hexadecimal, guint binary, guint width)
 {
-	gchar *p, temp[sizeof ("18446744073709551615")];
-	size_t len;
+	gchar *p, temp[64];
+	size_t len = 0;
 
-	if (hexadecimal == 0) {
+	if (G_LIKELY(hexadecimal == 0 && binary == 0)) {
 		p = temp;
 
 		if (ui64 < G_MAXUINT32) {
@@ -374,7 +404,7 @@ rspamd_sprintf_num (gchar *buf, gchar *last, guint64 ui64, gchar zero,
 
 		len = (temp + sizeof (temp)) - p;
 	}
-	else { /* hexadecimal == 2 */
+	else if (hexadecimal == 2) { /* hexadecimal == 2 */
 		p = temp + sizeof(temp);
 		do {
 			*--p = _HEX[(guint32) (ui64 & 0xf)];
@@ -382,6 +412,16 @@ rspamd_sprintf_num (gchar *buf, gchar *last, guint64 ui64, gchar zero,
 
 		len = (temp + sizeof (temp)) - p;
 	}
+	else if (binary > 0) {
+		int first_bit = MIN(sizeof(temp), rspamd_ffsll(ui64));
+
+		p = temp + sizeof(temp);
+		for (int i = 0; i <= first_bit; i ++, ui64 >>= 1) {
+			*--p = '0' + (ui64 & 0x1);
+		}
+
+		len = (temp + sizeof (temp)) - p;
+	}
 
 	/* zero or space padding */
 
@@ -1040,7 +1080,7 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 			}
 
 			if (!humanize) {
-				p = rspamd_sprintf_num (p, last, ui64, zero, hex, width);
+				p = rspamd_sprintf_num (p, last, ui64, zero, hex, b64 + b32, width);
 			}
 			else {
 				p = rspamd_humanize_number (p, last, ui64, bytes);
