@@ -66,11 +66,11 @@ struct rspamd_regexp_s {
 	gpointer ud;
 	gpointer re_class;
 	guint64 cache_id;
+	gsize match_limit;
 	guint max_hits;
 	gint flags;
 	gint pcre_flags;
 	gint ncaptures;
-	gint nbackref;
 };
 
 struct rspamd_regexp_cache {
@@ -116,28 +116,33 @@ rspamd_regexp_dtor (rspamd_regexp_t *re)
 	if (re) {
 		if (re->raw_re && re->raw_re != re->re) {
 #ifndef WITH_PCRE2
-#ifdef HAVE_PCRE_JIT
+			/* PCRE1 version */
+# ifdef HAVE_PCRE_JIT
 			if (re->raw_extra) {
 				pcre_free_study (re->raw_extra);
 			}
-#endif
+# endif
 #else
-			if (re->mcontext) {
-				pcre2_match_context_free (re->mcontext);
+			/* PCRE 2 version */
+			if (re->raw_mcontext) {
+				pcre2_match_context_free (re->raw_mcontext);
 			}
 #endif
 			PCRE_FREE (re->raw_re);
 		}
+
 		if (re->re) {
 #ifndef WITH_PCRE2
-#ifdef HAVE_PCRE_JIT
+			/* PCRE1 version */
+# ifdef HAVE_PCRE_JIT
 			if (re->extra) {
 				pcre_free_study (re->extra);
 			}
-#endif
+# endif
 #else
-			if (re->raw_mcontext) {
-				pcre2_match_context_free (re->raw_mcontext);
+			/* PCRE 2 version */
+			if (re->mcontext) {
+				pcre2_match_context_free (re->mcontext);
 			}
 #endif
 			PCRE_FREE (re->re);
@@ -169,15 +174,18 @@ rspamd_regexp_post_process (rspamd_regexp_t *r)
 	pcre2_set_recursion_limit (r->mcontext, max_recursion_depth);
 	pcre2_set_match_limit (r->mcontext, max_backtrack);
 
-	if (r->re != r->raw_re) {
+	if (r->raw_re && r->re != r->raw_re) {
 		r->raw_mcontext = pcre2_match_context_create (NULL);
 		g_assert (r->raw_mcontext != NULL);
 		pcre2_set_recursion_limit (r->raw_mcontext, max_recursion_depth);
 		pcre2_set_match_limit (r->raw_mcontext, max_backtrack);
 	}
-	else {
+	else if (r->raw_re) {
 		r->raw_mcontext = r->mcontext;
 	}
+	else {
+		r->raw_mcontext = NULL;
+	}
 
 #ifdef HAVE_PCRE_JIT
 	if (can_jit) {
@@ -306,10 +314,10 @@ rspamd_regexp_post_process (rspamd_regexp_t *r)
 }
 
 rspamd_regexp_t*
-rspamd_regexp_new (const gchar *pattern, const gchar *flags,
+rspamd_regexp_new_len (const gchar *pattern, gsize len, const gchar *flags,
 		GError **err)
 {
-	const gchar *start = pattern, *end, *flags_str = NULL;
+	const gchar *start = pattern, *end = start + len, *flags_str = NULL, *flags_end = NULL;
 	gchar *err_str;
 	rspamd_regexp_t *res;
 	gboolean explicit_utf = FALSE;
@@ -331,12 +339,12 @@ rspamd_regexp_new (const gchar *pattern, const gchar *flags,
 		return NULL;
 	}
 
-	if (flags == NULL) {
+	if (flags == NULL && start + 1 < end) {
 		/* We need to parse pattern and detect flags set */
 		if (*start == '/') {
 			sep = '/';
 		}
-		else if (*start == 'm') {
+		else if (*start == 'm' && start[1] != '\\' && g_ascii_ispunct(start[1])) {
 			start ++;
 			sep = *start;
 
@@ -347,22 +355,23 @@ rspamd_regexp_new (const gchar *pattern, const gchar *flags,
 
 			rspamd_flags |= RSPAMD_REGEXP_FLAG_FULL_MATCH;
 		}
-		if (sep == '\0' || g_ascii_isalnum (sep)) {
+		if (sep == 0) {
 			/* We have no flags, no separators and just use all line as expr */
 			start = pattern;
-			end = start + strlen (pattern);
 			rspamd_flags &= ~RSPAMD_REGEXP_FLAG_FULL_MATCH;
 		}
 		else {
-			end = strrchr (pattern, sep);
+			gchar *last_sep = rspamd_memrchr(pattern, sep, len);
 
-			if (end == NULL || end <= start) {
+			if (last_sep == NULL || last_sep <= start) {
 				g_set_error (err, rspamd_regexp_quark(), EINVAL,
 						"pattern is not enclosed with %c: %s",
 						sep, pattern);
 				return NULL;
 			}
-			flags_str = end + 1;
+			flags_str = last_sep + 1;
+			flags_end = end;
+			end = last_sep;
 			start ++;
 		}
 	}
@@ -370,8 +379,10 @@ rspamd_regexp_new (const gchar *pattern, const gchar *flags,
 		/* Strictly check all flags */
 		strict_flags = TRUE;
 		start = pattern;
-		end = pattern + strlen (pattern);
 		flags_str = flags;
+		if (flags) {
+			flags_end = flags + strlen(flags);
+		}
 	}
 
 	rspamd_flags |= RSPAMD_REGEXP_FLAG_RAW;
@@ -384,7 +395,7 @@ rspamd_regexp_new (const gchar *pattern, const gchar *flags,
 #endif
 
 	if (flags_str != NULL) {
-		while (*flags_str) {
+		while (flags_str < flags_end) {
 			switch (*flags_str) {
 			case 'i':
 				regexp_flags |= PCRE_FLAG(CASELESS);
@@ -512,32 +523,27 @@ fin:
 			&ncaptures) == 0) {
 		res->ncaptures = ncaptures;
 	}
-
-	/* Check number of backrefs */
-	if (pcre_fullinfo (res->raw_re, res->extra, PCRE_INFO_BACKREFMAX,
-			&ncaptures) == 0) {
-		res->nbackref = ncaptures;
-	}
 #else
 	/* Check number of captures */
 	if (pcre2_pattern_info (res->raw_re, PCRE2_INFO_CAPTURECOUNT,
 			&ncaptures) == 0) {
 		res->ncaptures = ncaptures;
 	}
-
-	/* Check number of backrefs */
-	if (pcre2_pattern_info (res->raw_re, PCRE2_INFO_BACKREFMAX,
-			&ncaptures) == 0) {
-		res->nbackref = ncaptures;
-	}
 #endif
 
 	return res;
 }
 
+rspamd_regexp_t *
+rspamd_regexp_new (const gchar *pattern, const gchar *flags,
+									GError **err)
+{
+	return rspamd_regexp_new_len(pattern, strlen(pattern), flags, err);
+}
+
 #ifndef WITH_PCRE2
 gboolean
-rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
+rspamd_regexp_search (const rspamd_regexp_t *re, const gchar *text, gsize len,
 		const gchar **start, const gchar **end, gboolean raw,
 		GArray *captures)
 {
@@ -549,6 +555,7 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 	const gchar *mt;
 	gsize remain = 0;
 	gint rc, match_flags = 0, *ovec, ncaptures, i;
+	const int junk = 0xdeadbabe;
 
 	g_assert (re != NULL);
 	g_assert (text != NULL);
@@ -557,6 +564,10 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 		len = strlen (text);
 	}
 
+	if (re->match_limit > 0 && len > re->match_limit) {
+		len = re->match_limit;
+	}
+
 	if (end != NULL && *end != NULL) {
 		/* Incremental search */
 		mt = (*end);
@@ -605,6 +616,11 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 	ncaptures = (re->ncaptures + 1) * 3;
 	ovec = g_alloca (sizeof (gint) * ncaptures);
 
+
+	for (i = 0; i < ncaptures; i ++) {
+		ovec[i] = junk;
+	}
+
 	if (!(re->flags & RSPAMD_REGEXP_FLAG_NOOPT)) {
 #ifdef HAVE_PCRE_JIT
 # if defined(HAVE_PCRE_JIT_FAST) && !defined(DISABLE_JIT_FAST)
@@ -635,11 +651,21 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 	}
 
 	if (rc >= 0) {
-		if (start) {
-			*start = mt + ovec[0];
+		if (rc > 0) {
+			if (start) {
+				*start = mt + ovec[0];
+			}
+			if (end) {
+				*end = mt + ovec[1];
+			}
 		}
-		if (end) {
-			*end = mt + ovec[1];
+		else {
+			if (start) {
+				*start = mt;
+			}
+			if (end) {
+				*end = mt + remain;
+			}
 		}
 
 		if (captures != NULL && rc >= 1) {
@@ -650,9 +676,16 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 			g_array_set_size (captures, rc);
 
 			for (i = 0; i < rc; i ++) {
-				elt = &g_array_index (captures, struct rspamd_re_capture, i);
-				elt->p = mt + ovec[i * 2];
-				elt->len = (mt + ovec[i * 2 + 1]) - elt->p;
+				if (ovec[i * 2] != junk && ovec[i * 2] >= 0) {
+					elt = &g_array_index (captures, struct rspamd_re_capture, i);
+					elt->p = mt + ovec[i * 2];
+					elt->len = (mt + ovec[i * 2 + 1]) - elt->p;
+				}
+				else {
+					/* Runtime match returned fewer captures than expected */
+					g_array_set_size (captures, i);
+					break;
+				}
 
 			}
 		}
@@ -672,7 +705,7 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 #else
 /* PCRE 2 version */
 gboolean
-rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
+rspamd_regexp_search (const rspamd_regexp_t *re, const gchar *text, gsize len,
 		const gchar **start, const gchar **end, gboolean raw,
 		GArray *captures)
 {
@@ -680,7 +713,8 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 	pcre2_match_context *mcontext;
 	PCRE_T *r;
 	const gchar *mt;
-	gsize remain = 0, *ovec;
+	PCRE2_SIZE remain = 0, *ovec;
+	const PCRE2_SIZE junk = 0xdeadbabeeeeeeeeULL;
 	gint rc, match_flags, novec, i;
 	gboolean ret = FALSE;
 
@@ -691,6 +725,10 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 		len = strlen (text);
 	}
 
+	if (re->match_limit > 0 && len > re->match_limit) {
+		len = re->match_limit;
+	}
+
 	if (end != NULL && *end != NULL) {
 		/* Incremental search */
 		mt = (*end);
@@ -725,6 +763,14 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 	}
 
 	match_data = pcre2_match_data_create (re->ncaptures + 1, NULL);
+	novec = pcre2_get_ovector_count (match_data);
+	ovec = pcre2_get_ovector_pointer (match_data);
+
+	/* Fill ovec with crap, so we can stop if actual matches is less than announced */
+	for (i = 0; i < novec; i ++) {
+		ovec[i * 2] = junk;
+		ovec[i * 2 + 1] = junk;
+	}
 
 #ifdef HAVE_PCRE_JIT
 	if (!(re->flags & RSPAMD_REGEXP_FLAG_DISABLE_JIT) && can_jit) {
@@ -746,14 +792,21 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 #endif
 
 	if (rc >= 0) {
-		novec = pcre2_get_ovector_count (match_data);
-		ovec = pcre2_get_ovector_pointer (match_data);
-
-		if (start) {
-			*start = mt + ovec[0];
+		if (novec > 0) {
+			if (start) {
+				*start = mt + ovec[0];
+			}
+			if (end) {
+				*end = mt + ovec[1];
+			}
 		}
-		if (end) {
-			*end = mt + ovec[1];
+		else {
+			if (start) {
+				*start = mt;
+			}
+			if (end) {
+				*end = mt + remain;
+			}
 		}
 
 		if (captures != NULL && novec >= 1) {
@@ -764,10 +817,15 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 			g_array_set_size (captures, novec);
 
 			for (i = 0; i < novec; i ++) {
-				elt = &g_array_index (captures, struct rspamd_re_capture, i);
-				elt->p = mt + ovec[i * 2];
-				elt->len = (mt + ovec[i * 2 + 1]) - elt->p;
-
+				if (ovec[i * 2] != junk && ovec[i * 2] != PCRE2_UNSET) {
+					elt = &g_array_index (captures, struct rspamd_re_capture, i);
+					elt->p = mt + ovec[i * 2];
+					elt->len = (mt + ovec[i * 2 + 1]) - elt->p;
+				}
+				else {
+					g_array_set_size (captures, i);
+					break;
+				}
 			}
 		}
 
@@ -788,7 +846,7 @@ rspamd_regexp_search (rspamd_regexp_t *re, const gchar *text, gsize len,
 #endif
 
 const char*
-rspamd_regexp_get_pattern (rspamd_regexp_t *re)
+rspamd_regexp_get_pattern (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
@@ -808,7 +866,7 @@ rspamd_regexp_set_flags (rspamd_regexp_t *re, guint new_flags)
 }
 
 guint
-rspamd_regexp_get_flags (rspamd_regexp_t *re)
+rspamd_regexp_get_flags (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
@@ -816,31 +874,15 @@ rspamd_regexp_get_flags (rspamd_regexp_t *re)
 }
 
 guint
-rspamd_regexp_get_pcre_flags (rspamd_regexp_t *re)
+rspamd_regexp_get_pcre_flags (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
 	return re->pcre_flags;
 }
 
-gint
-rspamd_regexp_get_nbackrefs (rspamd_regexp_t *re)
-{
-	g_assert (re != NULL);
-
-	return re->nbackref;
-}
-
-gint
-rspamd_regexp_get_ncaptures (rspamd_regexp_t *re)
-{
-	g_assert (re != NULL);
-
-	return re->ncaptures;
-}
-
 guint
-rspamd_regexp_get_maxhits (rspamd_regexp_t *re)
+rspamd_regexp_get_maxhits (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
@@ -860,7 +902,7 @@ rspamd_regexp_set_maxhits (rspamd_regexp_t *re, guint new_maxhits)
 }
 
 guint64
-rspamd_regexp_get_cache_id (rspamd_regexp_t *re)
+rspamd_regexp_get_cache_id (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
@@ -879,8 +921,28 @@ rspamd_regexp_set_cache_id (rspamd_regexp_t *re, guint64 id)
 	return old;
 }
 
+gsize
+rspamd_regexp_get_match_limit (const rspamd_regexp_t *re)
+{
+	g_assert (re != NULL);
+
+	return re->match_limit;
+}
+
+gsize
+rspamd_regexp_set_match_limit (rspamd_regexp_t *re, gsize lim)
+{
+	gsize old;
+
+	g_assert (re != NULL);
+	old = re->match_limit;
+	re->match_limit = lim;
+
+	return old;
+}
+
 gboolean
-rspamd_regexp_match (rspamd_regexp_t *re, const gchar *text, gsize len,
+rspamd_regexp_match (const rspamd_regexp_t *re, const gchar *text, gsize len,
 		gboolean raw)
 {
 	const gchar *start = NULL, *end = NULL;
@@ -888,7 +950,7 @@ rspamd_regexp_match (rspamd_regexp_t *re, const gchar *text, gsize len,
 	g_assert (re != NULL);
 	g_assert (text != NULL);
 
-	if (len == 0 && text != NULL) {
+	if (len == 0) {
 		len = strlen (text);
 	}
 
@@ -926,7 +988,7 @@ rspamd_regexp_set_ud (rspamd_regexp_t *re, gpointer ud)
 }
 
 gpointer
-rspamd_regexp_get_ud (rspamd_regexp_t *re)
+rspamd_regexp_get_ud (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
@@ -1169,15 +1231,15 @@ rspamd_regexp_library_init (struct rspamd_config *cfg)
 }
 
 gpointer
-rspamd_regexp_get_id (rspamd_regexp_t *re)
+rspamd_regexp_get_id (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
-	return re->id;
+	return (gpointer)re->id;
 }
 
 gpointer
-rspamd_regexp_get_class (rspamd_regexp_t *re)
+rspamd_regexp_get_class (const rspamd_regexp_t *re)
 {
 	g_assert (re != NULL);
 
