@@ -27,6 +27,10 @@
 #endif
 #include <math.h>
 
+#ifdef __x86_64__
+#include <immintrin.h>
+#endif
+
 #include "contrib/fastutf8/fastutf8.h"
 
 const guchar lc_map[256] = {
@@ -98,6 +102,44 @@ rspamd_str_lc (gchar *str, guint size)
 	return size;
 }
 
+gsize
+rspamd_str_copy_lc (const gchar *src, gchar *dst, gsize size)
+{
+	gchar *d = dst;
+
+	/* Find aligned start */
+	while ((0xf & (uintptr_t)src) && size > 0) {
+		*d++ = lc_map[(guchar)*src++];
+		size --;
+	}
+
+	/* Aligned start in src */
+#ifdef __x86_64__
+	while (size >= 16) {
+		__m128i sv = _mm_load_si128((const __m128i*)src);
+		/* From A */
+		__m128i rangeshift = _mm_sub_epi8(sv, _mm_set1_epi8((char)('A'+128)));
+		/* To Z */
+		__m128i nomodify = _mm_cmpgt_epi8(rangeshift, _mm_set1_epi8(-128 + 25));
+		/* ^ ' ' */
+		__m128i flip  = _mm_andnot_si128(nomodify, _mm_set1_epi8(0x20));
+		__m128i uc = _mm_xor_si128(sv, flip);
+		_mm_storeu_si128((__m128i*)d, uc);
+		d += 16;
+		src += 16;
+		size -= 16;
+	}
+#endif
+
+	/* Leftover */
+	while (size > 0) {
+		*d++ = lc_map[(guchar)*src++];
+		size --;
+	}
+
+	return (d - dst);
+}
+
 gint
 rspamd_lc_cmp (const gchar *s, const gchar *d, gsize l)
 {
@@ -484,6 +526,28 @@ rspamd_strtol (const gchar *s, gsize len, glong *value)
 /*
  * Try to convert string of length to long
  */
+#define CONV_STR_LIM_DECIMAL(max_num) do { \
+    while (p < end) { \
+        c = *p; \
+        if (c >= '0' && c <= '9') { \
+            c -= '0'; \
+            if (v > cutoff || (v == cutoff && (guint8)c > cutlim)) { \
+                *value = (max_num); \
+                return FALSE; \
+            } \
+            else { \
+                v *= 10; \
+                v += c; \
+            } \
+        } \
+        else { \
+            *value = v; \
+            return FALSE; \
+        } \
+    p++; \
+    } \
+} while(0)
+
 gboolean
 rspamd_strtoul (const gchar *s, gsize len, gulong *value)
 {
@@ -492,9 +556,39 @@ rspamd_strtoul (const gchar *s, gsize len, gulong *value)
 	gulong v = 0;
 	const gulong cutoff = G_MAXULONG / 10, cutlim = G_MAXULONG % 10;
 
+	/* Some preparations for range errors */
+	CONV_STR_LIM_DECIMAL(G_MAXULONG);
+
+	*value = v;
+	return TRUE;
+}
+
+gboolean
+rspamd_strtou64 (const gchar *s, gsize len, guint64 *value)
+{
+	const gchar *p = s, *end = s + len;
+	gchar c;
+	guint64 v = 0;
+	const guint64 cutoff = G_MAXUINT64 / 10, cutlim = G_MAXUINT64 % 10;
+
+	/* Some preparations for range errors */
+	CONV_STR_LIM_DECIMAL(G_MAXUINT64);
+
+	*value = v;
+	return TRUE;
+}
+
+gboolean
+rspamd_xstrtoul (const gchar *s, gsize len, gulong *value)
+{
+	const gchar *p = s, *end = s + len;
+	gchar c;
+	gulong v = 0;
+	const gulong cutoff = G_MAXULONG / 10, cutlim = G_MAXULONG % 10;
+
 	/* Some preparations for range errors */
 	while (p < end) {
-		c = *p;
+		c = g_ascii_tolower (*p);
 		if (c >= '0' && c <= '9') {
 			c -= '0';
 			if (v > cutoff || (v == cutoff && (guint8)c > cutlim)) {
@@ -503,7 +597,19 @@ rspamd_strtoul (const gchar *s, gsize len, gulong *value)
 				return FALSE;
 			}
 			else {
-				v *= 10;
+				v *= 16;
+				v += c;
+			}
+		}
+		else if (c >= 'a' || c <= 'f') {
+			c = c - 'a' + 10;
+			if (v > cutoff || (v == cutoff && (guint8)c > cutlim)) {
+				/* Range error */
+				*value = G_MAXULONG;
+				return FALSE;
+			}
+			else {
+				v *= 16;
 				v += c;
 			}
 		}
@@ -902,7 +1008,7 @@ rspamd_decode_base32_buf (const gchar *in, gsize inlen, guchar *out, gsize outle
 			}
 		}
 
-		if (processed_bits > 0 && o < end) {
+		if (processed_bits > 0 && o < end && acc != 0) {
 			*o++ = (acc & 0xFF);
 		}
 		else if (o > end) {
@@ -1465,18 +1571,18 @@ rspamd_strings_levenshtein_distance (const gchar *s1, gsize s1len,
 }
 
 GString *
-rspamd_header_value_fold (const gchar *name,
-		const gchar *value,
-		guint fold_max,
-		enum rspamd_newlines_type how,
-		const gchar *fold_on_chars)
+rspamd_header_value_fold (const gchar *name, gsize name_len,
+						  const gchar *value,
+						  gsize value_len,
+						  guint fold_max,
+						  enum rspamd_newlines_type how,
+						  const gchar *fold_on_chars)
 {
 	GString *res;
 	const guint default_fold_max = 76;
 	guint cur_len;
-	const gchar *p, *c;
+	const gchar *p, *c, *end, *fold_sequence;
 	guint nspaces = 0;
-	const gchar *last;
 	gboolean first_token = TRUE;
 	enum {
 		fold_before = 0,
@@ -1497,14 +1603,28 @@ rspamd_header_value_fold (const gchar *name,
 		fold_max = default_fold_max;
 	}
 
-	res = g_string_sized_new (strlen (value));
+	switch (how) {
+	case RSPAMD_TASK_NEWLINES_LF:
+		fold_sequence = "\n\t";
+		break;
+	case RSPAMD_TASK_NEWLINES_CR:
+		fold_sequence = "\r\t";
+		break;
+	case RSPAMD_TASK_NEWLINES_CRLF:
+	default:
+		fold_sequence ="\r\n\t";
+		break;
+	}
+
+	res = g_string_sized_new (value_len);
 
 	c = value;
 	p = c;
+	end = value + value_len;
 	/* name:<WSP> */
-	cur_len = strlen (name) + 2;
+	cur_len = name_len + 2;
 
-	while (*p) {
+	while (p < end) {
 		switch (state) {
 
 		case read_token:
@@ -1546,16 +1666,44 @@ rspamd_header_value_fold (const gchar *name,
 						state = fold_token;
 						next_state = read_token;
 					} else {
-						/* Reset line length */
-						cur_len = 0;
+						/* We need to ensure that it is a folding and not something else */
 
-						while (g_ascii_isspace (*p)) {
-							p++;
+						const char *t = p;
+						bool seen_fold = false;
+
+						while (t < end) {
+							if (*t == ' ' || *t == '\t') {
+								seen_fold = true;
+								break;
+							}
+							else if (!g_ascii_isspace(*t)) {
+								break;
+							}
+
+							t++;
 						}
 
-						g_string_append_len (res, c, p - c);
-						c = p;
-						first_token = TRUE;
+						if (seen_fold) {
+							/* Reset line length */
+							cur_len = 0;
+
+							while (g_ascii_isspace (*p)) {
+								p++;
+							}
+
+							g_string_append_len(res, c, p - c);
+							c = p;
+							first_token = TRUE;
+						}
+						else {
+							/* Not seen folding, inject it */
+							g_string_append_len (res, c, p - c);
+							g_string_append (res, fold_sequence);
+							p = t; /* Adjust p to ensure that we do not append extra stuff */
+							state = read_token;
+							first_token = TRUE;
+							c = p;
+						}
 					}
 				} else if (g_ascii_isspace (*p)) {
 					if (cur_len > fold_max * 0.8 && cur_len < fold_max) {
@@ -1591,7 +1739,7 @@ rspamd_header_value_fold (const gchar *name,
 					 * Check any spaces that are appended to the result
 					 * before folding
 					 */
-					last = &res->str[res->len - 1];
+					const gchar *last = &res->str[res->len - 1];
 
 					while (g_ascii_isspace (*last)) {
 						last --;
@@ -1600,18 +1748,7 @@ rspamd_header_value_fold (const gchar *name,
 					}
 				}
 
-				switch (how) {
-				case RSPAMD_TASK_NEWLINES_LF:
-					g_string_append_len (res, "\n\t", 2);
-					break;
-				case RSPAMD_TASK_NEWLINES_CR:
-					g_string_append_len (res, "\r\t", 2);
-					break;
-				case RSPAMD_TASK_NEWLINES_CRLF:
-				default:
-					g_string_append_len (res, "\r\n\t", 3);
-					break;
-				}
+				g_string_append (res, fold_sequence);
 
 				/* Skip space if needed */
 				if (g_ascii_isspace (*p)) {
@@ -1646,18 +1783,7 @@ rspamd_header_value_fold (const gchar *name,
 						res->len --;
 					}
 
-					switch (how) {
-					case RSPAMD_TASK_NEWLINES_LF:
-						g_string_append_len (res, "\n\t", 2);
-						break;
-					case RSPAMD_TASK_NEWLINES_CR:
-						g_string_append_len (res, "\r\t", 2);
-						break;
-					case RSPAMD_TASK_NEWLINES_CRLF:
-					default:
-						g_string_append_len (res, "\r\n\t", 3);
-						break;
-					}
+					g_string_append (res, fold_sequence);
 				}
 
 				/* Move leftover spaces */
@@ -1709,18 +1835,7 @@ rspamd_header_value_fold (const gchar *name,
 			if (g_ascii_isspace (*c)) {
 				c ++;
 			}
-			switch (how) {
-			case RSPAMD_TASK_NEWLINES_LF:
-				g_string_append_len (res, "\n\t", 2);
-				break;
-			case RSPAMD_TASK_NEWLINES_CR:
-				g_string_append_len (res, "\r\t", 2);
-				break;
-			case RSPAMD_TASK_NEWLINES_CRLF:
-			default:
-				g_string_append_len (res, "\r\n\t", 3);
-				break;
-			}
+			g_string_append (res, fold_sequence);
 			g_string_append_len (res, c, p - c);
 		}
 		else {
@@ -1739,18 +1854,7 @@ rspamd_header_value_fold (const gchar *name,
 		else {
 			if (*c != '\r' && *c != '\n') {
 				/* We need to add folding as well */
-				switch (how) {
-				case RSPAMD_TASK_NEWLINES_LF:
-					g_string_append_len (res, "\n\t", 2);
-					break;
-				case RSPAMD_TASK_NEWLINES_CR:
-					g_string_append_len (res, "\r\t", 2);
-					break;
-				case RSPAMD_TASK_NEWLINES_CRLF:
-				default:
-					g_string_append_len (res, "\r\n\t", 3);
-					break;
-				}
+				g_string_append (res, fold_sequence);
 				g_string_append_len (res, c, p - c);
 			}
 			else {
@@ -2338,9 +2442,15 @@ decode:
 			remain --;
 			ret = 0;
 
-			if      (c >= '0' && c <= '9') { ret = c - '0'; }
-			else if (c >= 'A' && c <= 'F') { ret = c - 'A' + 10; }
-			else if (c >= 'a' && c <= 'f') { ret = c - 'a' + 10; }
+			if (c >= '0' && c <= '9') {
+				ret = c - '0';
+			}
+			else if (c >= 'A' && c <= 'F') {
+				ret = c - 'A' + 10;
+			}
+			else if (c >= 'a' && c <= 'f') {
+				ret = c - 'a' + 10;
+			}
 			else if (c == '\r') {
 				/* Eat one more endline */
 				if (remain > 0 && *p == '\n') {
@@ -2355,10 +2465,14 @@ decode:
 				continue;
 			}
 			else {
-				/* Hack, hack, hack, treat =<garbadge> as =<garbadge> */
-				if (remain > 0) {
+				/* Hack, hack, hack, treat =<garbage> as =<garbage> */
+				if (end - o > 1) {
+					*o++ = '=';
 					*o++ = *(p - 1);
 				}
+				else {
+					return (-1);
+				}
 
 				continue;
 			}
@@ -2366,10 +2480,30 @@ decode:
 			if (remain > 0) {
 				c = *p++;
 				ret *= 16;
+				remain --;
 
-				if      (c >= '0' && c <= '9') { ret += c - '0'; }
-				else if (c >= 'A' && c <= 'F') { ret += c - 'A' + 10; }
-				else if (c >= 'a' && c <= 'f') { ret += c - 'a' + 10; }
+				if (c >= '0' && c <= '9') {
+					ret += c - '0';
+				}
+				else if (c >= 'A' && c <= 'F') {
+					ret += c - 'A' + 10;
+				}
+				else if (c >= 'a' && c <= 'f') {
+					ret += c - 'a' + 10;
+				}
+				else {
+					/* Treat =<good><rubbish> as =<good><rubbish> */
+					if (end - o > 2) {
+						*o++ = '=';
+						*o++ = *(p - 2);
+						*o++ = *(p - 1);
+					}
+					else {
+						return (-1);
+					}
+
+					continue;
+				}
 
 				if (end - o > 0) {
 					*o++ = (gchar)ret;
@@ -2377,8 +2511,6 @@ decode:
 				else {
 					return (-1);
 				}
-
-				remain --;
 			}
 		}
 		else {
@@ -2973,135 +3105,6 @@ rspamd_get_unicode_normalizer (void)
 #endif
 }
 
-
-enum rspamd_normalise_result
-rspamd_normalise_unicode_inplace (rspamd_mempool_t *pool, gchar *start,
-		guint *len)
-{
-#if U_ICU_VERSION_MAJOR_NUM >= 44
-	UErrorCode uc_err = U_ZERO_ERROR;
-	UConverter *utf8_conv = rspamd_get_utf8_converter ();
-	const UNormalizer2 *norm = rspamd_get_unicode_normalizer ();
-	gint32 nsym, end;
-	UChar *src = NULL, *dest = NULL;
-	enum rspamd_normalise_result ret = 0;
-	gboolean has_invisible = FALSE;
-
-	/* We first need to convert data to UChars :( */
-	src = g_malloc ((*len + 1) * sizeof (*src));
-	nsym = ucnv_toUChars (utf8_conv, src, *len + 1,
-			start, *len, &uc_err);
-
-	if (!U_SUCCESS (uc_err)) {
-		msg_warn_pool_check ("cannot normalise URL, cannot convert to unicode: %s",
-				u_errorName (uc_err));
-		ret |= RSPAMD_UNICODE_NORM_ERROR;
-		goto out;
-	}
-
-	/* We can now check if we need to decompose */
-	end = unorm2_spanQuickCheckYes (norm, src, nsym, &uc_err);
-
-	if (!U_SUCCESS (uc_err)) {
-		msg_warn_pool_check ("cannot normalise URL, cannot check normalisation: %s",
-				u_errorName (uc_err));
-		ret |= RSPAMD_UNICODE_NORM_ERROR;
-		goto out;
-	}
-
-	for (gint32 i = 0; i < nsym; i ++) {
-		if (IS_ZERO_WIDTH_SPACE (src[i])) {
-			has_invisible = TRUE;
-			break;
-		}
-	}
-
-	uc_err = U_ZERO_ERROR;
-
-	if (end != nsym) {
-		/* No normalisation needed, but we may still have invisible spaces */
-		/* We copy sub(src, 0, end) to dest and normalise the rest */
-		ret |= RSPAMD_UNICODE_NORM_UNNORMAL;
-		dest = g_malloc (nsym * sizeof (*dest));
-		memcpy (dest, src, end * sizeof (*dest));
-		nsym = unorm2_normalizeSecondAndAppend (norm, dest, end, nsym,
-				src + end, nsym - end, &uc_err);
-
-		if (!U_SUCCESS (uc_err)) {
-			if (uc_err != U_BUFFER_OVERFLOW_ERROR) {
-				msg_warn_pool_check ("cannot normalise URL: %s",
-						u_errorName (uc_err));
-				ret |= RSPAMD_UNICODE_NORM_ERROR;
-			}
-
-			goto out;
-		}
-	}
-	else if (!has_invisible) {
-		goto out;
-	}
-	else {
-		dest = src;
-		src = NULL;
-	}
-
-	if (has_invisible) {
-		/* Also filter zero width spaces */
-		gint32 new_len = 0;
-		UChar *t = dest, *h = dest;
-
-		ret |= RSPAMD_UNICODE_NORM_ZERO_SPACES;
-
-		for (gint32 i = 0; i < nsym; i ++) {
-			if (!IS_ZERO_WIDTH_SPACE (*h)) {
-				*t++ = *h++;
-				new_len ++;
-			}
-			else {
-				h ++;
-			}
-		}
-
-		nsym = new_len;
-	}
-
-	/* We now convert it back to utf */
-	nsym = ucnv_fromUChars (utf8_conv, start, *len, dest, nsym, &uc_err);
-
-	if (!U_SUCCESS (uc_err)) {
-		msg_warn_pool_check ("cannot normalise URL, cannot convert to UTF8: %s"
-					   " input length: %d chars, unicode length: %d utf16 symbols",
-				u_errorName (uc_err), (gint)*len, (gint)nsym);
-
-		if (uc_err == U_BUFFER_OVERFLOW_ERROR) {
-			ret |= RSPAMD_UNICODE_NORM_OVERFLOW;
-		}
-		else {
-			ret |= RSPAMD_UNICODE_NORM_ERROR;
-		}
-
-		goto out;
-	}
-
-	*len = nsym;
-
-out:
-
-	if (src) {
-		g_free (src);
-	}
-
-	if (dest) {
-		g_free (dest);
-	}
-
-	return ret;
-#else
-	/* Kill that with fire please */
-	return FALSE;
-#endif
-}
-
 gchar *
 rspamd_str_regexp_escape (const gchar *pattern, gsize slen,
 		gsize *dst_len, enum rspamd_regexp_escape_flags flags)
@@ -3200,7 +3203,6 @@ rspamd_str_regexp_escape (const gchar *pattern, gsize slen,
 		switch (t) {
 		case '[':
 		case ']':
-		case '-':
 		case '\\':
 		case '{':
 		case '}':
@@ -3216,6 +3218,36 @@ rspamd_str_regexp_escape (const gchar *pattern, gsize slen,
 				*d++ = '\\';
 			}
 			break;
+		case '-':
+			if (flags & RSPAMD_REGEXP_ESCAPE_GLOB) {
+				/*
+				 * For glob patterns, we need to ensure that a previous character is alphanumeric
+				 * and there is `[` symbol somewhere before
+				 */
+				bool seen_brace = false;
+				const char *search = p;
+
+				while (search > pattern) {
+					if (!g_ascii_isalnum(*search) && *search != '-') {
+						break;
+					}
+					if (*search == '[' ) {
+						seen_brace = true;
+						break;
+					}
+
+					search --;
+				}
+
+				if (!seen_brace) {
+					/* Escape `-` symbol */
+					*d++ = '\\';
+				}
+			}
+			else if (!(flags & RSPAMD_REGEXP_ESCAPE_RE)) {
+				*d++ = '\\';
+			}
+			break;
 		case '*':
 		case '?':
 		case '+':
@@ -3223,10 +3255,8 @@ rspamd_str_regexp_escape (const gchar *pattern, gsize slen,
 				/* Treat * as .* and ? as .? */
 				*d++ = '.';
 			}
-			else {
-				if (!(flags & RSPAMD_REGEXP_ESCAPE_RE)) {
-					*d++ = '\\';
-				}
+			else if (!(flags & RSPAMD_REGEXP_ESCAPE_RE)) {
+				*d++ = '\\';
 			}
 			break;
 		default:
@@ -3564,12 +3594,15 @@ rspamd_str_has_8bit_u64 (const guchar *beg, gsize len)
 	guint8 orb = 0;
 
 	if (len >= 16) {
-		const guchar *nextd = beg+8;
+		const guchar *nextd = beg + sizeof(guint64);
 		guint64 n1 = 0, n2 = 0;
 
 		do {
-			n1 |= *(const guint64 *)beg;
-			n2 |= *(const guint64 *)nextd;
+			guint64 t;
+			memcpy(&t, beg, sizeof(t));
+			n1 |= t;
+			memcpy(&t, nextd, sizeof(t));
+			n2 |= t;
 			beg += 16;
 			nextd += 16;
 			len -= 16;
