@@ -59,6 +59,7 @@ struct upstream {
 	gchar *name;
 	ev_timer ev;
 	gdouble last_fail;
+	gdouble last_resolve;
 	gpointer ud;
 	enum rspamd_upstream_flag flags;
 	struct upstream_list *ls;
@@ -127,6 +128,14 @@ struct upstream_ctx {
         rspamd_upstream_log_id, "upstream", upstream->uid, \
         G_STRFUNC, \
         __VA_ARGS__)
+#define msg_info_upstream(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
+        "upstream", upstream->uid, \
+        G_STRFUNC, \
+        __VA_ARGS__)
+#define msg_err_upstream(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
+        "upstream", upstream->uid, \
+        G_STRFUNC, \
+        __VA_ARGS__)
 
 INIT_LOG_MODULE(upstream)
 
@@ -175,7 +184,7 @@ rspamd_upstreams_library_config (struct rspamd_config *cfg,
 		ctx->limits.max_errors = cfg->upstream_max_errors;
 	}
 	if (cfg->upstream_revive_time) {
-		ctx->limits.revive_time = cfg->upstream_max_errors;
+		ctx->limits.revive_time = cfg->upstream_revive_time;
 	}
 	if (cfg->upstream_lazy_resolve_time) {
 		ctx->limits.lazy_resolve_time = cfg->upstream_lazy_resolve_time;
@@ -623,42 +632,87 @@ rspamd_upstream_revive_cb (struct ev_loop *loop, ev_timer *w, int revents)
 
 static void
 rspamd_upstream_resolve_addrs (const struct upstream_list *ls,
-		struct upstream *up)
+		struct upstream *upstream)
 {
-	if (up->ctx->res != NULL &&
-			up->ctx->configured &&
-			up->dns_requests == 0 &&
-			!(up->flags & RSPAMD_UPSTREAM_FLAG_NORESOLVE)) {
+	/* XXX: maybe make it configurable */
+	static const gdouble min_resolve_interval = 60.0;
+
+	if (upstream->ctx->res != NULL &&
+		upstream->ctx->configured &&
+		upstream->dns_requests == 0 &&
+		!(upstream->flags & RSPAMD_UPSTREAM_FLAG_NORESOLVE)) {
+
+		gdouble now = ev_now (upstream->ctx->event_loop);
+
+		if (now - upstream->last_resolve < min_resolve_interval) {
+			msg_info_upstream ("do not resolve upstream %s as it was checked %.0f "
+					  "seconds ago (%.0f is minimum)",
+					upstream->name, now - upstream->last_resolve,
+					min_resolve_interval);
+
+			return;
+		}
+
 		/* Resolve name of the upstream one more time */
-		if (up->name[0] != '/') {
-			if (up->flags & RSPAMD_UPSTREAM_FLAG_SRV_RESOLVE) {
-				if (rdns_make_request_full (up->ctx->res,
-						rspamd_upstream_dns_srv_cb, up,
-						ls->limits->dns_timeout, ls->limits->dns_retransmits,
-						1, up->name, RDNS_REQUEST_SRV) != NULL) {
-					up->dns_requests++;
-					REF_RETAIN (up);
+		if (upstream->name[0] != '/') {
+			upstream->last_resolve = now;
+
+			/*
+			 * If upstream name has a port, then we definitely need to resolve
+			 * merely host part!
+			 */
+			char dns_name[253 + 1]; /* 253 == max dns name + \0 */
+			const char *semicolon_pos = strchr(upstream->name, ':');
+
+			if (semicolon_pos != NULL && semicolon_pos > upstream->name) {
+				if (sizeof (dns_name) > semicolon_pos - upstream->name) {
+					rspamd_strlcpy(dns_name, upstream->name,
+							semicolon_pos - upstream->name + 1);
+				}
+				else {
+					/* XXX: truncated */
+					msg_err_upstream ("internal error: upstream name is larger than"
+									  "max DNS name: %s", upstream->name);
+					rspamd_strlcpy(dns_name, upstream->name, sizeof(dns_name));
 				}
 			}
 			else {
-				if (rdns_make_request_full (up->ctx->res,
-						rspamd_upstream_dns_cb, up,
+				rspamd_strlcpy(dns_name, upstream->name, sizeof(dns_name));
+			}
+
+			if (upstream->flags & RSPAMD_UPSTREAM_FLAG_SRV_RESOLVE) {
+				if (rdns_make_request_full (upstream->ctx->res,
+						rspamd_upstream_dns_srv_cb, upstream,
 						ls->limits->dns_timeout, ls->limits->dns_retransmits,
-						1, up->name, RDNS_REQUEST_A) != NULL) {
-					up->dns_requests++;
-					REF_RETAIN (up);
+						1, dns_name, RDNS_REQUEST_SRV) != NULL) {
+					upstream->dns_requests++;
+					REF_RETAIN (upstream);
+				}
+			}
+			else {
+				if (rdns_make_request_full (upstream->ctx->res,
+						rspamd_upstream_dns_cb, upstream,
+						ls->limits->dns_timeout, ls->limits->dns_retransmits,
+						1, dns_name, RDNS_REQUEST_A) != NULL) {
+					upstream->dns_requests++;
+					REF_RETAIN (upstream);
 				}
 
-				if (rdns_make_request_full (up->ctx->res,
-						rspamd_upstream_dns_cb, up,
+				if (rdns_make_request_full (upstream->ctx->res,
+						rspamd_upstream_dns_cb, upstream,
 						ls->limits->dns_timeout, ls->limits->dns_retransmits,
-						1, up->name, RDNS_REQUEST_AAAA) != NULL) {
-					up->dns_requests++;
-					REF_RETAIN (up);
+						1, dns_name, RDNS_REQUEST_AAAA) != NULL) {
+					upstream->dns_requests++;
+					REF_RETAIN (upstream);
 				}
 			}
 		}
 	}
+	else if (upstream->dns_requests != 0) {
+		msg_info_upstream ("do not resolve upstream %s as another request for "
+					 "resolving has been already issued",
+					 upstream->name);
+	}
 }
 
 static void
@@ -698,7 +752,7 @@ rspamd_upstream_set_inactive (struct upstream_list *ls, struct upstream *upstrea
 	g_ptr_array_remove_index (ls->alive, upstream->active_idx);
 	upstream->active_idx = -1;
 
-	/* We need to update all indicies */
+	/* We need to update all indices */
 	for (i = 0; i < ls->alive->len; i ++) {
 		cur = g_ptr_array_index (ls->alive, i);
 		cur->active_idx = i;
@@ -748,7 +802,7 @@ rspamd_upstream_fail (struct upstream *upstream,
 			upstream->name,
 			reason);
 
-	if (upstream->ctx && upstream->active_idx != -1) {
+	if (upstream->ctx && upstream->active_idx != -1 && upstream->ls) {
 		sec_cur = rspamd_get_ticks (FALSE);
 
 		RSPAMD_UPSTREAM_LOCK (upstream);
@@ -757,6 +811,11 @@ rspamd_upstream_fail (struct upstream *upstream,
 			upstream->last_fail = sec_cur;
 			upstream->errors = 1;
 
+			if (upstream->ls && upstream->dns_requests == 0) {
+				/* Try to re-resolve address immediately */
+				rspamd_upstream_resolve_addrs (upstream->ls, upstream);
+			}
+
 			DL_FOREACH (upstream->ls->watchers, w) {
 				if (w->events_mask & RSPAMD_UPSTREAM_WATCH_FAILURE) {
 					w->func (upstream, RSPAMD_UPSTREAM_WATCH_FAILURE, 1, w->ud);
@@ -853,7 +912,7 @@ rspamd_upstream_ok (struct upstream *upstream)
 	struct upstream_list_watcher *w;
 
 	RSPAMD_UPSTREAM_LOCK (upstream);
-	if (upstream->errors > 0 && upstream->active_idx != -1) {
+	if (upstream->errors > 0 && upstream->active_idx != -1 && upstream->ls) {
 		/* We touch upstream if and only if it is active */
 		msg_debug_upstream ("reset errors on upstream %s (was %ud)", upstream->name, upstream->errors);
 		upstream->errors = 0;
@@ -989,6 +1048,15 @@ rspamd_upstream_name (struct upstream *up)
 	return up->name;
 }
 
+gint
+rspamd_upstream_port (struct upstream *up)
+{
+	struct upstream_addr_elt *elt;
+
+	elt = g_ptr_array_index (up->addrs.addr, up->addrs.cur);
+	return rspamd_inet_address_get_port (elt->addr);
+}
+
 gboolean
 rspamd_upstreams_add_upstream (struct upstream_list *ups, const gchar *str,
 		guint16 def_port, enum rspamd_upstream_parse_type parse_type,
@@ -1099,7 +1167,7 @@ rspamd_upstreams_add_upstream (struct upstream_list *ups, const gchar *str,
 		}
 		for (i = 0; i < addrs->len; i ++) {
 			addr = g_ptr_array_index (addrs, i);
-			rspamd_upstream_add_addr (upstream, rspamd_inet_address_copy (addr));
+			rspamd_upstream_add_addr (upstream, rspamd_inet_address_copy(addr, NULL));
 		}
 	}
 
@@ -1200,10 +1268,6 @@ rspamd_upstreams_parse_line_len (struct upstream_list *ups,
 		ups->rot_alg = RSPAMD_UPSTREAM_HASHED;
 		p += sizeof ("hash:") - 1;
 	}
-	else if (RSPAMD_LEN_CHECK_STARTS_WITH(p, len, "sequential:")) {
-		ups->rot_alg = RSPAMD_UPSTREAM_SEQUENTIAL;
-		p += sizeof ("sequential:") - 1;
-	}
 
 	while (p < end) {
 		span_len = rspamd_memcspn (p, separators, end - p);
@@ -1378,13 +1442,20 @@ rspamd_upstream_get_round_robin (struct upstream_list *ups,
 			}
 		}
 
-		if (up->checked * (up->errors + 1) < min_checked) {
+		/*
+		 * This code is used when all upstreams have zero weight
+		 * The logic is to select least currently used upstream and penalise
+		 * upstream with errors. The error penalty should no be too high
+		 * to avoid sudden traffic drop in this case.
+		 */
+		if (up->checked + up->errors * 2 < min_checked) {
 			min_checked_sel = up;
 			min_checked = up->checked;
 		}
 	}
 
 	if (max_weight == 0) {
+		/* All upstreams have zero weight */
 		if (min_checked > G_MAXUINT / 2) {
 			/* Reset all checked counters to avoid overflow */
 			for (i = 0; i < ups->alive->len; i ++) {
@@ -1675,3 +1746,16 @@ void rspamd_upstreams_add_watch_callback (struct upstream_list *ups,
 
 	DL_APPEND (ups->watchers, nw);
 }
+
+struct upstream*
+rspamd_upstream_ref (struct upstream *up)
+{
+	REF_RETAIN (up);
+	return up;
+}
+
+void
+rspamd_upstream_unref (struct upstream *up)
+{
+	REF_RELEASE (up);
+}
