@@ -248,6 +248,13 @@ gboolean rspamd_upstream_add_addr (struct upstream *up,
  */
 const gchar *rspamd_upstream_name (struct upstream *up);
 
+/**
+ * Returns the port of the current address for the upstream
+ * @param up
+ * @return
+ */
+gint rspamd_upstream_port (struct upstream *up);
+
 /**
  * Sets opaque user data associated with this upstream
  * @param up
@@ -299,6 +306,18 @@ struct upstream *rspamd_upstream_get_except (struct upstream_list *ups,
  */
 void rspamd_upstream_reresolve (struct upstream_ctx *ctx);
 
+/**
+ * Share ownership on upstream
+ * @param up
+ * @return
+ */
+struct upstream* rspamd_upstream_ref (struct upstream *up);
+/**
+ * Unshare ownership on upstream
+ * @param up
+ */
+void rspamd_upstream_unref (struct upstream *up);
+
 #ifdef  __cplusplus
 }
 #endif
