@@ -87,7 +87,7 @@ struct lua_dns_cbdata {
 	gint cbref;
 	gchar *to_resolve;
 	gchar *user_str;
-	struct rspamd_symcache_item *item;
+	struct rspamd_symcache_dynamic_item *item;
 	struct rspamd_async_session *s;
 };
 
@@ -109,28 +109,7 @@ lua_dns_get_type (lua_State *L, int argno)
 	}
 	else {
 		strtype = lua_tostring (L, argno);
-
-		if (g_ascii_strcasecmp (strtype, "a") == 0) {
-			type = RDNS_REQUEST_A;
-		}
-		else if (g_ascii_strcasecmp (strtype, "aaaa") == 0) {
-			type = RDNS_REQUEST_AAAA;
-		}
-		else if (g_ascii_strcasecmp (strtype, "mx") == 0) {
-			type = RDNS_REQUEST_MX;
-		}
-		else if (g_ascii_strcasecmp (strtype, "txt") == 0) {
-			type = RDNS_REQUEST_TXT;
-		}
-		else if (g_ascii_strcasecmp (strtype, "ptr") == 0) {
-			type = RDNS_REQUEST_PTR;
-		}
-		else if (g_ascii_strcasecmp (strtype, "soa") == 0) {
-			type = RDNS_REQUEST_SOA;
-		}
-		else {
-			msg_err ("bad DNS type: %s", strtype);
-		}
+		type = rdns_type_fromstr (strtype);
 	}
 
 	return type;
@@ -237,6 +216,11 @@ lua_push_dns_reply (lua_State *L, const struct rdns_reply *reply)
 
 		LL_FOREACH (reply->entries, elt)
 		{
+			if (!rdns_request_has_type(reply->request, elt->type)) {
+				/* Unrequested type has been returned, ignore it */
+				continue;
+			}
+
 			switch (elt->type) {
 			case RDNS_REQUEST_A:
 				addr = rspamd_inet_address_new (AF_INET, &elt->content.a.addr);
@@ -294,6 +278,10 @@ lua_push_dns_reply (lua_State *L, const struct rdns_reply *reply)
 				lua_pushinteger (L, elt->content.soa.minimum);
 				lua_settable (L, -3);
 
+				lua_rawseti (L, -2, ++i);
+				break;
+			case RDNS_REQUEST_CNAME:
+				lua_pushstring (L, elt->content.cname.name);
 				lua_rawseti (L, -2, ++i);
 				break;
 			default:
@@ -358,7 +346,7 @@ lua_dns_resolver_resolve_common (lua_State *L,
 	struct rspamd_task *task = NULL;
 	GError *err = NULL;
 	gboolean forced = FALSE;
-	struct rspamd_symcache_item *item = NULL;
+	struct rspamd_symcache_dynamic_item *item = NULL;
 
 	/* Check arguments */
 	if (!rspamd_lua_parse_table_arguments (L, first, &err,
@@ -759,6 +747,7 @@ luaopen_dns_resolver (lua_State * L)
 		LUA_ENUM (L, DNS_SPF, RDNS_REQUEST_SPF);
 		LUA_ENUM (L, DNS_AAAA, RDNS_REQUEST_AAAA);
 		LUA_ENUM (L, DNS_SOA, RDNS_REQUEST_SOA);
+		LUA_ENUM (L, DNS_CNAME, RDNS_REQUEST_CNAME);
 	}
 
 	lua_pop (L, 1);
