@@ -16,9 +16,11 @@
 #include "lua_common.h"
 #include "lua_thread_pool.h"
 #include "libserver/http/http_private.h"
+#include "libutil/upstream.h"
 #include "ref.h"
 #include "unix-std.h"
 #include "zlib.h"
+#include "utlist.h"
 
 /***
  * @module rspamd_http
@@ -65,7 +67,7 @@ static const struct luaL_reg httplib_m[] = {
 struct lua_http_cbdata {
 	struct rspamd_http_connection *conn;
 	struct rspamd_async_session *session;
-	struct rspamd_symcache_item *item;
+	struct rspamd_symcache_dynamic_item *item;
 	struct rspamd_http_message *msg;
 	struct ev_loop *event_loop;
 	struct rspamd_config *cfg;
@@ -77,6 +79,7 @@ struct lua_http_cbdata {
 	gchar *mime_type;
 	gchar *host;
 	gchar *auth;
+	struct upstream *up;
 	const gchar *url;
 	gsize max_size;
 	gint flags;
@@ -126,6 +129,10 @@ lua_http_fin (gpointer arg)
 		rspamd_inet_address_free (cbd->addr);
 	}
 
+	if (cbd->up) {
+		rspamd_upstream_unref(cbd->up);
+	}
+
 	if (cbd->mime_type) {
 		g_free (cbd->mime_type);
 	}
@@ -201,6 +208,11 @@ static void
 lua_http_error_handler (struct rspamd_http_connection *conn, GError *err)
 {
 	struct lua_http_cbdata *cbd = (struct lua_http_cbdata *)conn->ud;
+
+	if (cbd->up) {
+		rspamd_upstream_fail(cbd->up, false, err ? err->message : "unknown error");
+	}
+
 	if (cbd->cbref == -1) {
 		if (cbd->flags & RSPAMD_LUA_HTTP_FLAG_YIELDED) {
 			cbd->flags &= ~RSPAMD_LUA_HTTP_FLAG_YIELDED;
@@ -249,6 +261,10 @@ lua_http_finish_handler (struct rspamd_http_connection *conn,
 	}
 	lua_thread_pool_prepare_callback (cbd->cfg->lua_thread_pool, &lcbd);
 
+	if (cbd->up) {
+		rspamd_upstream_ok(cbd->up);
+	}
+
 	L = lcbd.L;
 
 	lua_rawgeti (L, LUA_REGISTRYINDEX, cbd->cbref);
@@ -446,7 +462,8 @@ lua_http_make_connection (struct lua_http_cbdata *cbd)
 			if (cbd->item) {
 				rspamd_session_add_event_full (cbd->session,
 						(event_finalizer_t) lua_http_fin, cbd,
-						M, rspamd_symcache_item_name (cbd->item));
+						M,
+						rspamd_symcache_dyn_item_name (cbd->task, cbd->item));
 			}
 			else {
 				rspamd_session_add_event (cbd->session,
@@ -484,7 +501,7 @@ static void
 lua_http_dns_handler (struct rdns_reply *reply, gpointer ud)
 {
 	struct lua_http_cbdata *cbd = (struct lua_http_cbdata *)ud;
-	struct rspamd_symcache_item *item;
+	struct rspamd_symcache_dynamic_item *item;
 	struct rspamd_task *task;
 
 	task = cbd->task;
@@ -495,28 +512,40 @@ lua_http_dns_handler (struct rdns_reply *reply, gpointer ud)
 		REF_RELEASE (cbd);
 	}
 	else {
-		if (reply->entries->type == RDNS_REQUEST_A) {
-			cbd->addr = rspamd_inet_address_new (AF_INET,
-					&reply->entries->content.a.addr);
-		}
-		else if (reply->entries->type == RDNS_REQUEST_AAAA) {
-			cbd->addr = rspamd_inet_address_new (AF_INET6,
-					&reply->entries->content.aaa.addr);
-		}
+		struct rdns_reply_entry *entry;
 
-		REF_RETAIN (cbd);
-		if (!lua_http_make_connection (cbd)) {
-			lua_http_push_error (cbd, "unable to make connection to the host");
-
-			if (cbd->ref.refcount > 1) {
-				REF_RELEASE (cbd);
+		DL_FOREACH(reply->entries, entry) {
+			if (entry->type == RDNS_REQUEST_A) {
+				cbd->addr = rspamd_inet_address_new(AF_INET,
+						&entry->content.a.addr);
+				break;
+			}
+			else if (entry->type == RDNS_REQUEST_AAAA) {
+				cbd->addr = rspamd_inet_address_new(AF_INET6,
+						&entry->content.aaa.addr);
+				break;
 			}
-
-			REF_RELEASE (cbd);
-
-			return;
 		}
-		REF_RELEASE (cbd);
+
+		if (cbd->addr == NULL) {
+			lua_http_push_error (cbd, "unable to resolve host: no records with such name");
+			REF_RELEASE (cbd);
+		}
+		else {
+			REF_RETAIN (cbd);
+			if (!lua_http_make_connection(cbd)) {
+				lua_http_push_error(cbd, "unable to make connection to the host");
+
+				if (cbd->ref.refcount > 1) {
+					REF_RELEASE (cbd);
+				}
+
+				REF_RELEASE (cbd);
+
+				return;
+			}
+			REF_RELEASE (cbd);
+		}
 	}
 
 	if (item) {
@@ -598,6 +627,7 @@ lua_http_request (lua_State *L)
 	struct rspamd_config *cfg = NULL;
 	struct rspamd_cryptobox_pubkey *peer_key = NULL;
 	struct rspamd_cryptobox_keypair *local_kp = NULL;
+	struct upstream *up = NULL;
 	const gchar *url, *lua_body;
 	rspamd_fstring_t *body = NULL;
 	gint cbref = -1;
@@ -939,6 +969,20 @@ lua_http_request (lua_State *L)
 
 		lua_pop (L, 1);
 
+		lua_pushstring (L, "upstream");
+		lua_gettable (L, 1);
+
+		if (lua_type (L, -1) == LUA_TUSERDATA) {
+			struct rspamd_lua_upstream *lup = lua_check_upstream(L, -1);
+
+			if (lup) {
+				/* Preserve pointer in case if lup is destructed */
+				up = lup->up;
+			}
+		}
+
+		lua_pop (L, 1);
+
 		lua_pushstring (L, "user");
 		lua_gettable (L, 1);
 
@@ -1036,6 +1080,10 @@ lua_http_request (lua_State *L)
 	cbd->auth = auth;
 	cbd->task = task;
 
+	if (up) {
+		cbd->up = rspamd_upstream_ref(up);
+	}
+
 	if (cbd->cbref == -1) {
 		cbd->thread = lua_thread_pool_get_running_entry (cfg->lua_thread_pool);
 	}
@@ -1064,6 +1112,7 @@ lua_http_request (lua_State *L)
 	bool numeric_ip = false;
 
 	/* Check if we can skip resolving */
+
 	gsize hostlen = 0;
 	const gchar *host = rspamd_http_message_get_http_host (msg, &hostlen);
 
@@ -1071,6 +1120,7 @@ lua_http_request (lua_State *L)
 		cbd->host = g_malloc (hostlen + 1);
 		rspamd_strlcpy (cbd->host, host, hostlen + 1);
 
+		/* Keep-alive entry is available */
 		if (cbd->flags & RSPAMD_LUA_HTTP_FLAG_KEEP_ALIVE) {
 			const rspamd_inet_addr_t *ka_addr = rspamd_http_context_has_keepalive(NULL,
 					cbd->host,
@@ -1083,16 +1133,30 @@ lua_http_request (lua_State *L)
 			}
 		}
 
+		/*
+		 * No keep-alive stuff, check if we have upstream or if we can parse host as
+		 * a numeric address
+		 */
 		if (!cbd->addr) {
-			/* We use msg->host here, not cbd->host ! */
-			if (rspamd_parse_inet_address (&cbd->addr,
-					msg->host->str, msg->host->len,
-					RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
+			if (cbd->up) {
 				numeric_ip = true;
+				cbd->addr = rspamd_inet_address_copy(rspamd_upstream_addr_next(cbd->up), NULL);
+			}
+			else {
+				/* We use msg->host here, not cbd->host ! */
+				if (rspamd_parse_inet_address(&cbd->addr,
+						msg->host->str, msg->host->len,
+						RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
+					numeric_ip = true;
+				}
 			}
 		}
 	}
 	else {
+		if (cbd->up) {
+			numeric_ip = true;
+			cbd->addr = rspamd_inet_address_copy(rspamd_upstream_addr_next(cbd->up), NULL);
+		}
 		cbd->host = NULL;
 	}
 
@@ -1104,6 +1168,9 @@ lua_http_request (lua_State *L)
 		ret = lua_http_make_connection (cbd);
 
 		if (!ret) {
+			if (cbd->up) {
+				rspamd_upstream_fail(cbd->up, true, "HTTP connection failed");
+			}
 			if (cbd->ref.refcount > 1) {
 				/* Not released by make_connection */
 				REF_RELEASE (cbd);
