@@ -210,7 +210,7 @@ lua_ip_new (lua_State *L, struct rspamd_lua_ip *old)
 	ip = g_malloc0 (sizeof (*ip));
 
 	if (old != NULL && old->addr != NULL) {
-		ip->addr = rspamd_inet_address_copy (old->addr);
+		ip->addr = rspamd_inet_address_copy(old->addr, NULL);
 	}
 
 	pip = lua_newuserdata (L, sizeof (struct rspamd_lua_ip *));
@@ -588,11 +588,16 @@ rspamd_lua_ip_push (lua_State *L, rspamd_inet_addr_t *addr)
 {
 	struct rspamd_lua_ip *ip, **pip;
 
-	ip = g_malloc0 (sizeof (struct rspamd_lua_ip));
-	ip->addr = rspamd_inet_address_copy (addr);
-	pip = lua_newuserdata (L, sizeof (struct rspamd_lua_ip *));
-	rspamd_lua_setclass (L, "rspamd{ip}", -1);
-	*pip = ip;
+	if (addr) {
+		ip = g_malloc0(sizeof(struct rspamd_lua_ip));
+		ip->addr = rspamd_inet_address_copy(addr, NULL);
+		pip = lua_newuserdata(L, sizeof(struct rspamd_lua_ip *));
+		rspamd_lua_setclass(L, "rspamd{ip}", -1);
+		*pip = ip;
+	}
+	else {
+		lua_pushnil (L);
+	}
 }
 
 void
