@@ -27,6 +27,35 @@
  *
  * All maps could be obtained by function `rspamd_config:get_maps()`
  * Also see [`lua_maps` module description](lua_maps.html).
+ *
+ * **Important notice** maps cannot be queried outside of the worker context.
+ * For example, you cannot add even a file map and query some keys from it during
+ * some module initialisation, you need to add the appropriate event loop context
+ * for a worker (e.g. you cannot use `get_key` outside of the symbols callbacks or
+ * a worker `on_load` scripts).
+ *
+@example
+
+local hash_map = rspamd_config:add_map{
+  type = "hash",
+  urls = ['file:///path/to/file'],
+  description = 'sample map'
+}
+
+local function sample_symbol_cb(task)
+    -- Check whether hash map contains from address of message
+    if hash_map:get_key((task:get_from() or {})[1]) then
+      -- key found
+    end
+end
+
+rspamd_config:register_symbol{
+  name = 'SAMPLE_SYMBOL',
+  type = 'normal',
+  score = 1.0,
+  description = "A sample symbol",
+  callback = sample_symbol_cb,
+}
  */
 
 /***
@@ -430,60 +459,76 @@ lua_map_fin (struct map_cb_data *data, void **target)
 
 	map = data->map;
 
-	if (data->cur_data) {
-		cbdata = (struct lua_map_callback_data *)data->cur_data;
+	if (data->errored) {
+		if (data->cur_data) {
+			cbdata = (struct lua_map_callback_data *)data->cur_data;
+			if (cbdata->ref != -1) {
+				luaL_unref (cbdata->L, LUA_REGISTRYINDEX, cbdata->ref);
+			}
+
+			if (cbdata->data) {
+				rspamd_fstring_free (cbdata->data);
+			}
+
+			data->cur_data = NULL;
+		}
 	}
 	else {
-		msg_err_map ("no data read for map");
-		return;
-	}
-
-	if (cbdata->ref == -1) {
-		msg_err_map ("map has no callback set");
-	}
-	else if (cbdata->data != NULL && cbdata->data->len != 0) {
-
-		lua_pushcfunction (cbdata->L, &rspamd_lua_traceback);
-		int err_idx = lua_gettop (cbdata->L);
-
-		lua_rawgeti (cbdata->L, LUA_REGISTRYINDEX, cbdata->ref);
-
-		if (!cbdata->opaque) {
-			lua_pushlstring (cbdata->L, cbdata->data->str, cbdata->data->len);
+		if (data->cur_data) {
+			cbdata = (struct lua_map_callback_data *) data->cur_data;
 		}
 		else {
-			struct rspamd_lua_text *t;
-
-			t = lua_newuserdata (cbdata->L, sizeof (*t));
-			rspamd_lua_setclass (cbdata->L, "rspamd{text}", -1);
-			t->flags = 0;
-			t->len = cbdata->data->len;
-			t->start = cbdata->data->str;
+			msg_err_map ("no data read for map");
+			return;
 		}
 
-		pmap = lua_newuserdata (cbdata->L, sizeof (void *));
-		*pmap = cbdata->lua_map;
-		rspamd_lua_setclass (cbdata->L, "rspamd{map}", -1);
+		if (cbdata->ref == -1) {
+			msg_err_map ("map has no callback set");
+		}
+		else if (cbdata->data != NULL && cbdata->data->len != 0) {
 
-		gint ret = lua_pcall (cbdata->L, 2, 0, err_idx);
+			lua_pushcfunction (cbdata->L, &rspamd_lua_traceback);
+			int err_idx = lua_gettop(cbdata->L);
 
-		if (ret != 0) {
-			msg_info_map ("call to %s failed (%d): %s", "map fin function",
-				ret,
-				lua_tostring (cbdata->L, -1));
+			lua_rawgeti(cbdata->L, LUA_REGISTRYINDEX, cbdata->ref);
+
+			if (!cbdata->opaque) {
+				lua_pushlstring(cbdata->L, cbdata->data->str, cbdata->data->len);
+			}
+			else {
+				struct rspamd_lua_text *t;
+
+				t = lua_newuserdata(cbdata->L, sizeof(*t));
+				rspamd_lua_setclass(cbdata->L, "rspamd{text}", -1);
+				t->flags = 0;
+				t->len = cbdata->data->len;
+				t->start = cbdata->data->str;
+			}
+
+			pmap = lua_newuserdata(cbdata->L, sizeof(void *));
+			*pmap = cbdata->lua_map;
+			rspamd_lua_setclass(cbdata->L, "rspamd{map}", -1);
+
+			gint ret = lua_pcall(cbdata->L, 2, 0, err_idx);
+
+			if (ret != 0) {
+				msg_info_map ("call to %s failed (%d): %s", "map fin function",
+						ret,
+						lua_tostring(cbdata->L, -1));
+			}
+
+			lua_settop(cbdata->L, err_idx - 1);
 		}
 
-		lua_settop (cbdata->L, err_idx - 1);
-	}
+		cbdata->data = rspamd_fstring_assign(cbdata->data, "", 0);
 
-	cbdata->data = rspamd_fstring_assign (cbdata->data, "", 0);
+		if (target) {
+			*target = data->cur_data;
+		}
 
-	if (target) {
-		*target = data->cur_data;
-	}
-
-	if (data->prev_data) {
-		data->prev_data = NULL;
+		if (data->prev_data) {
+			data->prev_data = NULL;
+		}
 	}
 }
 
