@@ -450,7 +450,6 @@ lua_rsa_signature_load (lua_State *L)
 			lua_pushnil (L);
 		}
 		else {
-			sig = g_malloc (sizeof (rspamd_fstring_t));
 			if (fstat (fd, &st) == -1 ||
 				(data =
 				mmap (NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0))
@@ -621,9 +620,10 @@ lua_rsa_verify_memory (lua_State *L)
 				signature->str, signature->len, rsa);
 
 		if (ret == 0) {
-			msg_info ("cannot check rsa signature for data: %s",
-				ERR_error_string (ERR_get_error (), NULL));
 			lua_pushboolean (L, FALSE);
+			lua_pushstring (L, ERR_error_string (ERR_get_error (), NULL));
+
+			return 2;
 		}
 		else {
 			lua_pushboolean (L, TRUE);
@@ -660,14 +660,19 @@ lua_rsa_sign_memory (lua_State *L)
 
 	if (rsa != NULL && data != NULL) {
 		signature = rspamd_fstring_sized_new (RSA_size (rsa));
+
+		guint siglen = signature->len;
 		ret = RSA_sign (NID_sha256, data, sz,
-				signature->str, (guint *)&signature->len, rsa);
+				signature->str, &siglen, rsa);
 
 		if (ret != 1) {
+			rspamd_fstring_free (signature);
+
 			return luaL_error (L, "cannot sign: %s",
 					ERR_error_string (ERR_get_error (), NULL));
 		}
 		else {
+			signature->len = siglen;
 			psig = lua_newuserdata (L, sizeof (rspamd_fstring_t *));
 			rspamd_lua_setclass (L, "rspamd{rsa_signature}", -1);
 			*psig = signature;
