@@ -339,11 +339,12 @@ struct lua_tcp_cbdata {
 	struct lua_tcp_dtor *dtors;
 	ref_entry_t ref;
 	struct rspamd_task *task;
-	struct rspamd_symcache_item *item;
+	struct rspamd_symcache_dynamic_item *item;
 	struct thread_entry *thread;
 	struct rspamd_config *cfg;
 	struct rspamd_ssl_connection *ssl_conn;
 	gchar *hostname;
+	struct upstream *up;
 	gboolean eof;
 };
 
@@ -461,6 +462,10 @@ lua_tcp_fin (gpointer arg)
 		rspamd_inet_address_free (cbd->addr);
 	}
 
+	if (cbd->up) {
+		rspamd_upstream_unref(cbd->up);
+	}
+
 	while (lua_tcp_shift_handler (cbd)) {}
 	g_queue_free (cbd->handlers);
 
@@ -537,6 +542,10 @@ lua_tcp_push_error (struct lua_tcp_cbdata *cbd, gboolean is_fatal,
 	lua_State *L;
 	gboolean callback_called = FALSE;
 
+	if (is_fatal && cbd->up) {
+		rspamd_upstream_fail(cbd->up, false, err);
+	}
+
 	if (cbd->thread) {
 		va_start (ap, err);
 		lua_tcp_resume_thread_error_argp (cbd, err, ap);
@@ -593,6 +602,12 @@ lua_tcp_push_error (struct lua_tcp_cbdata *cbd, gboolean is_fatal,
 
 			TCP_RELEASE (cbd);
 
+			if ((cbd->flags & (LUA_TCP_FLAG_FINISHED|LUA_TCP_FLAG_CONNECTED)) ==
+				(LUA_TCP_FLAG_FINISHED|LUA_TCP_FLAG_CONNECTED)) {
+				/* A callback has called `close` method, so we need to release a refcount */
+				TCP_RELEASE (cbd);
+			}
+
 			callback_called = TRUE;
 		}
 
@@ -685,6 +700,12 @@ lua_tcp_push_data (struct lua_tcp_cbdata *cbd, const guint8 *str, gsize len)
 
 		lua_settop (L, top);
 		TCP_RELEASE (cbd);
+
+		if ((cbd->flags & (LUA_TCP_FLAG_FINISHED|LUA_TCP_FLAG_CONNECTED)) ==
+			(LUA_TCP_FLAG_FINISHED|LUA_TCP_FLAG_CONNECTED)) {
+			/* A callback has called `close` method, so we need to release a refcount */
+			TCP_RELEASE (cbd);
+		}
 	}
 
 	lua_thread_pool_restore_callback (&cbs);
@@ -907,6 +928,10 @@ call_finish_handler:
 		cbd->flags &= ~LUA_TCP_FLAG_SHUTDOWN;
 	}
 
+	if (cbd->up) {
+		rspamd_upstream_ok(cbd->up);
+	}
+
 	lua_tcp_push_data (cbd, NULL, 0);
 	if (!IS_SYNC (cbd)) {
 		lua_tcp_shift_handler (cbd);
@@ -1019,6 +1044,11 @@ lua_tcp_process_read (struct lua_tcp_cbdata *cbd,
 		}
 		else {
 			lua_tcp_push_error (cbd, TRUE, "IO read error: connection terminated");
+
+			if ((cbd->flags & LUA_TCP_FLAG_FINISHED)) {
+				/* A callback has called `close` method, so we need to release a refcount */
+				TCP_RELEASE (cbd);
+			}
 		}
 
 		lua_tcp_plan_handler_event (cbd, FALSE, FALSE);
@@ -1042,6 +1072,11 @@ lua_tcp_process_read (struct lua_tcp_cbdata *cbd,
 			lua_tcp_push_error (cbd, TRUE,
 					"IO read error while trying to read data: %s",
 					strerror (errno));
+
+			if ((cbd->flags & LUA_TCP_FLAG_FINISHED)) {
+				/* A callback has called `close` method, so we need to release a refcount */
+				TCP_RELEASE (cbd);
+			}
 		}
 
 		lua_tcp_plan_handler_event (cbd, FALSE, FALSE);
@@ -1121,6 +1156,12 @@ lua_tcp_handler (int fd, short what, gpointer ud)
 					lua_settop (L, top);
 					TCP_RELEASE (cbd);
 					lua_thread_pool_restore_callback (&cbs);
+
+					if ((cbd->flags & (LUA_TCP_FLAG_FINISHED|LUA_TCP_FLAG_CONNECTED)) ==
+						(LUA_TCP_FLAG_FINISHED|LUA_TCP_FLAG_CONNECTED)) {
+						/* A callback has called `close` method, so we need to release a refcount */
+						TCP_RELEASE (cbd);
+					}
 				}
 			}
 		}
@@ -1134,6 +1175,12 @@ lua_tcp_handler (int fd, short what, gpointer ud)
 		else {
 			g_assert_not_reached ();
 		}
+
+		if ((cbd->flags & (LUA_TCP_FLAG_FINISHED|LUA_TCP_FLAG_CONNECTED)) ==
+			(LUA_TCP_FLAG_FINISHED|LUA_TCP_FLAG_CONNECTED)) {
+			/* A callback has called `close` method, so we need to release a refcount */
+			TCP_RELEASE (cbd);
+		}
 	}
 	else {
 		lua_tcp_push_error (cbd, TRUE, "IO timeout");
@@ -1156,8 +1203,8 @@ lua_tcp_plan_handler_event (struct lua_tcp_cbdata *cbd, gboolean can_read,
 		if (!(cbd->flags & LUA_TCP_FLAG_FINISHED)) {
 			/* We are finished with a connection */
 			msg_debug_tcp ("no handlers left, finish session");
-			TCP_RELEASE (cbd);
 			cbd->flags |= LUA_TCP_FLAG_FINISHED;
+			TCP_RELEASE (cbd);
 		}
 	}
 	else {
@@ -1233,7 +1280,7 @@ lua_tcp_register_event (struct lua_tcp_cbdata *cbd)
 
 		if (cbd->item) {
 			cbd->async_ev = rspamd_session_add_event_full (cbd->session, fin, cbd, M,
-					rspamd_symcache_item_name (cbd->item));
+					rspamd_symcache_dyn_item_name (cbd->task, cbd->item));
 		}
 		else {
 			cbd->async_ev = rspamd_session_add_event (cbd->session, fin, cbd, M);
@@ -1365,17 +1412,30 @@ lua_tcp_dns_handler (struct rdns_reply *reply, gpointer ud)
 		 * We set this flag as it means that we have already registered the watcher
 		 * when started DNS query
 		 */
+		struct rdns_reply_entry *entry;
+
+		DL_FOREACH(reply->entries, entry) {
+			if (entry->type == RDNS_REQUEST_A) {
+				cbd->addr = rspamd_inet_address_new(AF_INET,
+						&entry->content.a.addr);
+				break;
+			}
+			else if (entry->type == RDNS_REQUEST_AAAA) {
+				cbd->addr = rspamd_inet_address_new(AF_INET6,
+						&entry->content.aaa.addr);
+				break;
+			}
+		}
+
+		if (cbd->addr == NULL){
+			rn = rdns_request_get_name (reply->request, NULL);
+			lua_tcp_push_error (cbd, TRUE, "unable to resolve host: %s; no records with this name",
+					rn->name);
+			TCP_RELEASE (cbd);
+			return;
+		}
+
 		cbd->flags |= LUA_TCP_FLAG_RESOLVED;
-
-		if (reply->entries->type == RDNS_REQUEST_A) {
-			cbd->addr = rspamd_inet_address_new (AF_INET,
-					&reply->entries->content.a.addr);
-		}
-		else if (reply->entries->type == RDNS_REQUEST_AAAA) {
-			cbd->addr = rspamd_inet_address_new (AF_INET6,
-					&reply->entries->content.aaa.addr);
-		}
-
 		rspamd_inet_address_set_port (cbd->addr, cbd->port);
 
 		if (!lua_tcp_make_connection (cbd)) {
@@ -1455,6 +1515,7 @@ lua_tcp_arg_toiovec (lua_State *L, gint pos, struct lua_tcp_cbdata *cbd,
  * - `stop_pattern`: stop reading on finding a certain pattern (e.g. \r\n.\r\n for smtp)
  * - `shutdown`: half-close socket after writing (boolean: default false)
  * - `read`: read response after sending request (boolean: default true)
+ * - `upstream`: optional upstream object that would be used to get an address
  * @return {boolean} true if request has been sent
  */
 static gint
@@ -1473,6 +1534,7 @@ lua_tcp_request (lua_State *L)
 	struct rspamd_task *task = NULL;
 	struct rspamd_config *cfg = NULL;
 	struct iovec *iov = NULL;
+	struct upstream *up = NULL;
 	guint niov = 0, total_out;
 	guint64 h;
 	gdouble timeout = default_tcp_timeout;
@@ -1642,6 +1704,20 @@ lua_tcp_request (lua_State *L)
 			lua_pop (L, 1);
 		}
 
+		lua_pushstring (L, "upstream");
+		lua_gettable (L, 1);
+
+		if (lua_type (L, -1) == LUA_TUSERDATA) {
+			struct rspamd_lua_upstream *lup = lua_check_upstream(L, -1);
+
+			if (lup) {
+				/* Preserve pointer in case if lup is destructed */
+				up = lup->up;
+			}
+		}
+
+		lua_pop (L, 1);
+
 		lua_pushstring (L, "data");
 		lua_gettable (L, -2);
 		total_out = 0;
@@ -1786,6 +1862,10 @@ lua_tcp_request (lua_State *L)
 	cbd->connect_cb = conn_cbref;
 	REF_INIT_RETAIN (cbd, lua_tcp_maybe_free);
 
+	if (up) {
+		cbd->up = rspamd_upstream_ref(up);
+	}
+
 	if (session) {
 		cbd->session = session;
 
@@ -1799,7 +1879,26 @@ lua_tcp_request (lua_State *L)
 		}
 	}
 
-	if (rspamd_parse_inet_address (&cbd->addr,
+	if (cbd->up) {
+		/* Use upstream to get addr */
+		cbd->addr = rspamd_inet_address_copy(rspamd_upstream_addr_next(cbd->up), NULL);
+
+		/* Host is numeric IP, no need to resolve */
+		lua_tcp_register_watcher (cbd);
+
+		if (!lua_tcp_make_connection (cbd)) {
+			lua_tcp_push_error (cbd, TRUE, "cannot connect to the host: %s", host);
+			lua_pushboolean (L, FALSE);
+
+			rspamd_upstream_fail(cbd->up, true, "failed to connect");
+
+			/* No reset of the item as watcher has been registered */
+			TCP_RELEASE (cbd);
+
+			return 1;
+		}
+	}
+	else if (rspamd_parse_inet_address (&cbd->addr,
 			host, strlen (host), RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
 		rspamd_inet_address_set_port (cbd->addr, port);
 		/* Host is numeric IP, no need to resolve */
@@ -2017,6 +2116,7 @@ lua_tcp_connect_sync (lua_State *L)
 
 			if (!rspamd_dns_resolver_request_task (task, lua_tcp_dns_handler, cbd,
 					RDNS_REQUEST_A, host)) {
+				cbd->item = NULL; /* We have not registered watcher */
 				lua_pushboolean (L, FALSE);
 				lua_pushliteral (L, "Failed to initiate dns request");
 				TCP_RELEASE (cbd);
@@ -2043,7 +2143,30 @@ lua_tcp_close (lua_State *L)
 	}
 
 	cbd->flags |= LUA_TCP_FLAG_FINISHED;
-	TCP_RELEASE (cbd);
+
+	if (cbd->ssl_conn) {
+		/* TODO: postpone close in case ssl is used ! */
+		rspamd_ssl_connection_free (cbd->ssl_conn);
+		cbd->ssl_conn = NULL;
+	}
+
+	if (cbd->fd != -1) {
+		rspamd_ev_watcher_stop (cbd->event_loop, &cbd->ev);
+		close (cbd->fd);
+		cbd->fd = -1;
+	}
+
+	if (cbd->addr) {
+		rspamd_inet_address_free (cbd->addr);
+		cbd->addr = NULL;
+	}
+
+	if (cbd->up) {
+		rspamd_upstream_unref(cbd->up);
+		cbd->up = NULL;
+	}
+	/* Do not release refcount as it will be handled elsewhere */
+	/* TCP_RELEASE (cbd); */
 
 	return 0;
 }
