@@ -101,7 +101,7 @@ LUA_FUNCTION_DEF (text, save_in_file);
 LUA_FUNCTION_DEF (text, span);
 /***
  * @method rspamd_text:sub(start[, len])
- * Returns a substrin for lua_text similar to string.sub from Lua
+ * Returns a substring for lua_text similar to string.sub from Lua
  * @return {rspamd_text} new rspamd_text with span (must be careful when using with owned texts...)
  */
 LUA_FUNCTION_DEF (text, sub);
@@ -194,7 +194,7 @@ LUA_FUNCTION_DEF (text, base32);
  * @method rspamd_text:base64([line_length, [nline, [fold]]])
  * Returns a text encoded in base64 (new rspamd_text is allocated)
  *
- * @param {number} line_length return text splited with newlines up to this attribute
+ * @param {number} line_length return text split with newlines up to this attribute
  * @param {string} nline newline type: `cr`, `lf`, `crlf`
  * @param {boolean} fold use folding when splitting into lines (false by default)
  * @return {rspamd_text} new text encoded in base64
@@ -225,6 +225,7 @@ LUA_FUNCTION_DEF (text, gc);
 LUA_FUNCTION_DEF (text, eq);
 LUA_FUNCTION_DEF (text, lt);
 LUA_FUNCTION_DEF (text, concat);
+LUA_FUNCTION_DEF (text, strtoul);
 
 static const struct luaL_reg textlib_f[] = {
 		LUA_INTERFACE_DEF (text, fromstring),
@@ -257,6 +258,7 @@ static const struct luaL_reg textlib_m[] = {
 		LUA_INTERFACE_DEF (text, base64),
 		LUA_INTERFACE_DEF (text, hex),
 		LUA_INTERFACE_DEF (text, find),
+		LUA_INTERFACE_DEF (text, strtoul),
 		{"write", lua_text_save_in_file},
 		{"__len", lua_text_len},
 		{"__tostring", lua_text_str},
@@ -286,19 +288,26 @@ lua_check_text_or_string (lua_State * L, gint pos)
 		return ud ? (struct rspamd_lua_text *) ud : NULL;
 	}
 	else if (pos_type == LUA_TSTRING) {
-		/* Fake static lua_text */
-		static struct rspamd_lua_text fake_text;
+		/*
+		 * Fake static lua_text, we allow to use this function multiple times
+		 * by having a small array of static structures.
+		 */
+		static unsigned cur_txt_idx = 0;
+		static struct rspamd_lua_text fake_text[4];
 		gsize len;
+		int sel_idx;
+
+		sel_idx = cur_txt_idx++ % G_N_ELEMENTS (fake_text);
+		fake_text[sel_idx].start = lua_tolstring (L, pos, &len);
 
-		fake_text.start = lua_tolstring (L, pos, &len);
 		if (len >= G_MAXUINT) {
 			return NULL;
 		}
 
-		fake_text.len = len;
-		fake_text.flags = RSPAMD_TEXT_FLAG_FAKE;
+		fake_text[sel_idx].len = len;
+		fake_text[sel_idx].flags = RSPAMD_TEXT_FLAG_FAKE;
 
-		return &fake_text;
+		return &fake_text[sel_idx];
 	}
 
 	return NULL;
@@ -339,6 +348,23 @@ lua_new_text (lua_State *L, const gchar *start, gsize len, gboolean own)
 	return t;
 }
 
+bool
+lua_is_text_binary(struct rspamd_lua_text *t)
+{
+	if (t == NULL || t->len == 0) {
+		return false;
+	}
+
+	if (rspamd_str_has_8bit(t->start, t->len)) {
+		if (rspamd_fast_utf8_validate(t->start, t->len) == 0) {
+			return false;
+		}
+		return true;
+	}
+
+	return false;
+}
+
 
 static gint
 lua_text_fromstring (lua_State *L)
@@ -911,6 +937,10 @@ lua_text_split (lua_State *L)
 	struct rspamd_lua_regexp *re;
 	gboolean stringify = FALSE, own_re = FALSE;
 
+	if (t == NULL) {
+		return luaL_error (L, "invalid arguments");
+	}
+
 	if (lua_type (L, 2) == LUA_TUSERDATA) {
 		re = lua_check_regexp (L, 2);
 	}
@@ -938,7 +968,7 @@ lua_text_split (lua_State *L)
 		own_re = TRUE;
 	}
 
-	if (t && re) {
+	if (re) {
 		if (lua_isboolean (L, 3)) {
 			stringify = lua_toboolean (L, 3);
 		}
@@ -1385,7 +1415,7 @@ lua_text_find (lua_State *L)
 }
 
 #define BITOP(a,b,op) \
-		((a)[(gsize)(b)/(8*sizeof *(a))] op (gsize)1<<((gsize)(b)%(8*sizeof *(a))))
+		((a)[(guint64)(b)/(8u*sizeof *(a))] op (guint64)1<<((guint64)(b)%(8u*sizeof *(a))))
 
 static gint
 lua_text_exclude_chars (lua_State *L)
@@ -1395,7 +1425,7 @@ lua_text_exclude_chars (lua_State *L)
 	gssize patlen;
 	const gchar *pat = lua_tolstring (L, 2, &patlen), *p, *end;
 	gchar *dest, *d;
-	gsize byteset[32 / sizeof(gsize)]; /* Bitset for ascii */
+	guint64 byteset[32 / sizeof(guint64)]; /* Bitset for ascii */
 	gboolean copy = TRUE;
 	guint *plen;
 
@@ -1436,7 +1466,7 @@ lua_text_exclude_chars (lua_State *L)
 
 				if (patlen > 0) {
 					/*
-					 * This stuff assumes little endian, but GSIZE_FROM_LE should
+					 * This stuff assumes little endian, but GUINT64_FROM_LE should
 					 * deal with proper conversion
 					 */
 					switch (*pat) {
@@ -1445,22 +1475,22 @@ lua_text_exclude_chars (lua_State *L)
 						break;
 					case 's':
 						/* "\r\n\t\f " */
-						byteset[0] |= GSIZE_FROM_LE (0x100003600);
+						byteset[0] |= GUINT64_FROM_LE(0x100003600LLU);
 						break;
 					case 'n':
 						/* newlines: "\r\n" */
-						byteset[0] |= GSIZE_FROM_LE (0x2400);
+						byteset[0] |= GUINT64_FROM_LE (0x2400LLU);
 						break;
 					case '8':
 						/* 8 bit characters */
-						byteset[2] |= GSIZE_FROM_LE (0xffffffffffffffffLLU);
-						byteset[3] |= GSIZE_FROM_LE (0xffffffffffffffffLLU);
+						byteset[2] |= GUINT64_FROM_LE (0xffffffffffffffffLLU);
+						byteset[3] |= GUINT64_FROM_LE (0xffffffffffffffffLLU);
 						break;
 					case 'c':
 						/* Non printable (control) characters */
-						byteset[0] |= GSIZE_FROM_LE (0xffffffff);
+						byteset[0] |= GUINT64_FROM_LE (0xffffffffLLU);
 						/* Del character */
-						byteset[1] |= GSIZE_FROM_LE (0x8000000000000000);
+						byteset[1] |= GUINT64_FROM_LE (0x8000000000000000LLU);
 						break;
 					}
 				}
@@ -1506,7 +1536,7 @@ lua_text_oneline (lua_State *L)
 	struct rspamd_lua_text *t = lua_check_text (L, 1);
 	const gchar *p, *end;
 	gchar *dest, *d;
-	gsize byteset[32 / sizeof(gsize)]; /* Bitset for ascii */
+	guint64 byteset[32 / sizeof(guint64)]; /* Bitset for ascii */
 	gboolean copy = TRUE, seen_8bit = FALSE;
 	guint *plen;
 
@@ -1540,14 +1570,14 @@ lua_text_oneline (lua_State *L)
 		/* Fill pattern bitset */
 		memset (byteset, 0, sizeof byteset);
 		/* All spaces */
-		byteset[0] |= GSIZE_FROM_LE (0x100003600);
+		byteset[0] |= GUINT64_FROM_LE (0x100003600LLU);
 		/* Control characters */
-		byteset[0] |= GSIZE_FROM_LE (0xffffffff);
+		byteset[0] |= GUINT64_FROM_LE (0xffffffffLLU);
 		/* Del character */
-		byteset[1] |= GSIZE_FROM_LE (0x8000000000000000);
+		byteset[1] |= GUINT64_FROM_LE (0x8000000000000000LLU);
 		/* 8 bit characters */
-		byteset[2] |= GSIZE_FROM_LE (0xffffffffffffffffLLU);
-		byteset[3] |= GSIZE_FROM_LE (0xffffffffffffffffLLU);
+		byteset[2] |= GUINT64_FROM_LE (0xffffffffffffffffLLU);
+		byteset[3] |= GUINT64_FROM_LE (0xffffffffffffffffLLU);
 
 		p = t->start;
 		end = t->start + t->len;
@@ -1676,6 +1706,29 @@ lua_text_lower (lua_State *L)
 	return 1;
 }
 
+static gint
+lua_text_strtoul (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t = lua_check_text (L, 1);
+
+	if (t) {
+		unsigned long ll;
+
+		if (rspamd_strtoul (t->start, t->len, &ll)) {
+			lua_pushinteger (L, ll);
+		}
+		else {
+			lua_pushnil (L);
+		}
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
 /* Used to distinguish lua text metatable */
 static const guint rspamd_lua_text_cookie = 0x2b21ef6fU;
 
