@@ -53,7 +53,7 @@ lua_thread_pool_free (struct lua_thread_pool *pool);
 
 /**
  * Extracts a thread from the list of available ones.
- * It immediately becames running one and should be used to run a Lua script/function straight away.
+ * It immediately becomes the running one and should be used to run a Lua script/function straight away.
  * as soon as the code is finished, it should be either returned into list of available threads by
  * calling lua_thread_pool_return() or terminated by calling lua_thread_pool_terminate_entry()
  * if the code finished with error.
@@ -183,6 +183,19 @@ lua_thread_resume_full (struct thread_entry *thread_entry,
 #define lua_thread_resume(thread_entry, narg) \
     lua_thread_resume_full (thread_entry, narg, G_STRLOC)
 
+/**
+ * Terminates thread pool entry and fill the pool with another thread entry if needed
+ * @param pool
+ * @param thread_entry
+ * @param loc
+ */
+void
+lua_thread_pool_terminate_entry_full (struct lua_thread_pool *pool,
+								 struct thread_entry *thread_entry,
+								 const gchar *loc, bool enforce);
+#define lua_thread_pool_terminate_entry(pool, thread_entry) \
+    lua_thread_pool_terminate_entry_full (pool, thread_entry, G_STRLOC, false)
+
 #ifdef  __cplusplus
 }
 #endif
