@@ -19,7 +19,7 @@
  * Allowed options:
  * - symbol (string): symbol to insert (default: 'R_BAD_CHARSET')
  * - threshold (double): value that would be used as threshold in expression characters_changed / total_characters
- *   (e.g. if threshold is 0.1 than charset change should occure more often than in 10 symbols), default: 0.1
+ *   (e.g. if threshold is 0.1 than charset change should occur more often than in 10 symbols), default: 0.1
  */
 
 
@@ -87,10 +87,10 @@ chartable_get_context (struct rspamd_config *cfg)
 }
 
 static void chartable_symbol_callback (struct rspamd_task *task,
-									   struct rspamd_symcache_item *item,
+									   struct rspamd_symcache_dynamic_item *item,
 									   void *unused);
 static void chartable_url_symbol_callback (struct rspamd_task *task,
-										   struct rspamd_symcache_item *item,
+										   struct rspamd_symcache_dynamic_item *item,
 										   void *unused);
 
 gint
@@ -596,7 +596,7 @@ rspamd_chartable_process_part (struct rspamd_task *task,
 	gdouble cur_score = 0.0;
 
 	if (part == NULL || part->utf_words == NULL ||
-			part->utf_words->len == 0) {
+			part->utf_words->len == 0 || part->nwords == 0) {
 		return FALSE;
 	}
 
@@ -640,7 +640,7 @@ rspamd_chartable_process_part (struct rspamd_task *task,
 
 static void
 chartable_symbol_callback (struct rspamd_task *task,
-		struct rspamd_symcache_item *item,
+		struct rspamd_symcache_dynamic_item *item,
 		void *unused)
 {
 	guint i;
@@ -674,7 +674,7 @@ chartable_symbol_callback (struct rspamd_task *task,
 		ignore_diacritics = TRUE;
 	}
 
-	if (task->meta_words != NULL) {
+	if (task->meta_words != NULL && task->meta_words->len > 0) {
 		rspamd_stat_token_t *w;
 		gdouble cur_score = 0;
 		gsize arlen = task->meta_words->len;
@@ -710,7 +710,7 @@ chartable_symbol_callback (struct rspamd_task *task,
 
 static void
 chartable_url_symbol_callback (struct rspamd_task *task,
-		struct rspamd_symcache_item *item,
+		struct rspamd_symcache_dynamic_item *item,
 		void *unused)
 {
 	/* XXX: TODO: unbreak module once URLs unicode project is over */
