@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2011-2015, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 Copyright (c) 2013-2015, Andrew Lewis <nerf@judo.za.org>
 
 Licensed under the Apache License, Version 2.0 (the "License");
@@ -22,12 +22,14 @@ end
 local hash = require 'rspamd_cryptobox_hash'
 local rspamd_logger = require 'rspamd_logger'
 local rspamd_util = require 'rspamd_util'
+local rspamd_ip = require "rspamd_ip"
 local fun = require 'fun'
 local lua_util = require 'lua_util'
-local ts = require("tableshape").types
 local selectors = require "lua_selectors"
 local bit = require 'bit'
 local lua_maps = require "lua_maps"
+local rbl_common = require "plugins/rbl"
+local rspamd_url = require "rspamd_url"
 
 -- This plugin implements various types of RBL checks
 -- Documentation can be found here:
@@ -36,28 +38,41 @@ local lua_maps = require "lua_maps"
 local E = {}
 local N = 'rbl'
 
+-- Checks that could be performed by rbl module
 local local_exclusions
 local white_symbols = {}
 local black_symbols = {}
 local monitored_addresses = {}
 local known_selectors = {} -- map from selector string to selector id
+local url_flag_bits = rspamd_url.flags
 
 local function get_monitored(rbl)
+  local function is_random_monitored()
+    -- Explicit definition
+    if type(rbl.random_monitored) == 'boolean' then
+      return rbl.random_monitored
+    end
+
+    -- We check 127.0.0.1 for merely RBLs with `from` or `received` and only if
+    -- they don't have `no_ip` attribute at the same time
+    --
+    -- Convert to a boolean variable using the common idiom
+    return (not (rbl.from or rbl.received)
+        or rbl.no_ip)
+        and true or false
+  end
+
   local default_monitored = '1.0.0.127'
   local ret = {
     rcode = 'nxdomain',
     prefix = default_monitored,
-    random = false,
+    random = is_random_monitored(),
   }
 
   if rbl.monitored_address then
     ret.prefix = rbl.monitored_address
   end
 
-  if rbl.dkim or rbl.urls or rbl.emails then
-    ret.random = true
-  end
-
   lua_util.debugm(N, rspamd_config,
       'added monitored address: %s (%s random)',
       ret.prefix, ret.random)
@@ -121,28 +136,27 @@ local function ip_to_rbl(ip)
 end
 
 local function gen_check_rcvd_conditions(rbl, received_total)
-  local min_pos = tonumber(rbl['received_min_pos'])
-  local max_pos = tonumber(rbl['received_max_pos'])
-  local match_flags = rbl['received_flags']
-  local nmatch_flags = rbl['received_nflags']
+  local min_pos = tonumber(rbl.received_min_pos)
+  local max_pos = tonumber(rbl.received_max_pos)
+  local match_flags = rbl.received_flags
+  local nmatch_flags = rbl.received_nflags
+
   local function basic_received_check(rh)
-    if not (rh['real_ip'] and rh['real_ip']:is_valid()) then return false end
-    if ((rh['real_ip']:get_version() == 6 and rbl['ipv6']) or
-        (rh['real_ip']:get_version() == 4 and rbl['ipv4'])) and
-        ((rbl['exclude_private_ips'] and not rh['real_ip']:is_local()) or
-            not rbl['exclude_private_ips']) and ((rbl['exclude_local_ips'] and
-        not is_excluded_ip(rh['real_ip'])) or not rbl['exclude_local_ips']) then
+    if not (rh.real_ip and rh.real_ip:is_valid()) then return false end
+    if ((rh.real_ip:get_version() == 6 and rbl.ipv6) or
+        (rh.real_ip:get_version() == 4 and rbl.ipv4)) and
+        ((rbl.exclude_private_ips and not rh.real_ip:is_local()) or
+            not rbl.exclude_private_ips) and ((rbl.exclude_local_ips and
+        not is_excluded_ip(rh.real_ip)) or not rbl.exclude_local_ips) then
       return true
     else
       return false
     end
   end
-  if not (max_pos or min_pos or match_flags or nmatch_flags) then
-    return basic_received_check
-  end
-  return function(rh, pos)
-    if not basic_received_check() then return false end
-    local got_flags = rh['flags'] or E
+
+  local function positioned_received_check(rh, pos)
+    if not rh or not basic_received_check(rh) then return false end
+    local got_flags = rh.flags or E
     if min_pos then
       if min_pos < 0 then
         if min_pos == -1 then
@@ -150,7 +164,7 @@ local function gen_check_rcvd_conditions(rbl, received_total)
             return false
           end
         else
-          if pos <= (received_total - (min_pos*-1)) then
+          if pos <= (received_total - math.abs(min_pos)) then
             return false
           end
         end
@@ -160,7 +174,7 @@ local function gen_check_rcvd_conditions(rbl, received_total)
     end
     if max_pos then
       if max_pos < -1 then
-        if (received_total - (max_pos*-1)) >= pos then
+        if (received_total - math.abs(max_pos)) >= pos then
           return false
         end
       elseif max_pos > 0 then
@@ -185,34 +199,47 @@ local function gen_check_rcvd_conditions(rbl, received_total)
     end
     return true
   end
+
+
+  if not (max_pos or min_pos or match_flags or nmatch_flags) then
+    return basic_received_check
+  else
+    return positioned_received_check
+  end
 end
 
 local function rbl_dns_process(task, rbl, to_resolve, results, err, resolve_table_elt)
-  local function make_option(ip)
+  local function make_option(ip, label)
     if ip then
       return string.format('%s:%s:%s',
           resolve_table_elt.orig,
-          resolve_table_elt.what,
+          label,
           ip)
     else
       return string.format('%s:%s',
           resolve_table_elt.orig,
-          resolve_table_elt.what)
+          label)
     end
   end
 
-  local function insert_result(s, ip)
+  local function insert_result(s, ip, label)
     if rbl.symbols_prefixes then
-      local prefix = rbl.symbols_prefixes[resolve_table_elt.what]
+      local prefix = rbl.symbols_prefixes[label]
 
       if not prefix then
-        rspamd_logger.warnx(task, 'unlisted symbol prefix for %s', resolve_table_elt.what)
-        task:insert_result(s, 1.0, make_option(ip))
+        rspamd_logger.warnx(task, 'unlisted symbol prefix for %s', label)
+        task:insert_result(s, 1.0, make_option(ip, label))
       else
-        task:insert_result(prefix .. '_' .. s, 1.0, make_option(ip))
+        task:insert_result(prefix .. '_' .. s, 1.0, make_option(ip, label))
       end
     else
-      task:insert_result(s, 1.0, make_option(ip))
+      task:insert_result(s, 1.0, make_option(ip, label))
+    end
+  end
+
+  local function insert_results(s, ip)
+    for label in pairs(resolve_table_elt.what) do
+      insert_result(s, ip, label)
     end
   end
 
@@ -236,7 +263,7 @@ local function rbl_dns_process(task, rbl, to_resolve, results, err, resolve_tabl
   end
 
   if rbl.returncodes == nil and rbl.returnbits == nil and rbl.symbol ~= nil then
-    insert_result(rbl.symbol)
+    insert_results(rbl.symbol)
     return
   end
 
@@ -251,7 +278,7 @@ local function rbl_dns_process(task, rbl, to_resolve, results, err, resolve_tabl
         for _,check_bit in ipairs(bits) do
           if bit.band(ipnum, check_bit) == check_bit then
             foundrc = true
-            insert_result(s)
+            insert_results(s)
             -- Here, we continue with other bits
           end
         end
@@ -261,7 +288,7 @@ local function rbl_dns_process(task, rbl, to_resolve, results, err, resolve_tabl
         for _,v in ipairs(codes) do
           if string.find(ipstr, '^' .. v .. '$') then
             foundrc = true
-            insert_result(s)
+            insert_results(s)
             break
           end
         end
@@ -270,10 +297,10 @@ local function rbl_dns_process(task, rbl, to_resolve, results, err, resolve_tabl
 
     if not foundrc then
       if rbl.unknown and rbl.symbol then
-        insert_result(rbl.symbol, ipstr)
+        insert_results(rbl.symbol, ipstr)
       else
-        rspamd_logger.errx(task, 'RBL %1 returned unknown result: %2',
-            rbl.rbl, ipstr)
+        lua_util.debugm(N, task, '%1 returned unknown result: %2',
+            to_resolve, ipstr)
       end
     end
   end
@@ -282,7 +309,12 @@ end
 
 local function gen_rbl_callback(rule)
   local function is_whitelisted(task, req, req_str, whitelist, what)
-    if rule.ignore_whitelist then return false end
+    if rule.ignore_whitelist then
+      lua_util.debugm(N, task,
+          'ignore whitelisting checks to %s by %s: ignore whitelist is being set',
+          req_str, rule.symbol)
+      return false
+    end
 
     if rule.whitelist then
       if rule.whitelist:get_key(req) then
@@ -296,25 +328,31 @@ local function gen_rbl_callback(rule)
 
     -- Maybe whitelisted by some other rbl rule
     if whitelist then
-      local wl_what = whitelist[req_str]
-      if wl_what then
+      local wl = whitelist[req_str]
+      if wl then
         lua_util.debugm(N, task,
-            'whitelisted %s on %s by %s rbl rule (%s checked)',
-            req_str, wl_what, what)
-        return wl_what == what
+            'whitelisted request to %s by %s (%s) rbl rule (%s checked type, %s whitelist type)',
+            req_str, wl.type, wl.symbol, what, wl.type)
+        if wl.type == what then
+          -- This was decided to be a bad idea as in case of whitelisting a request to blacklist
+          -- is not even sent
+          --task:adjust_result(wl.symbol, 0.0 / 0.0, rule.symbol)
+
+          return true
+        end
       end
     end
 
     return false
   end
 
-  local function add_dns_request(task, req, forced, is_ip, requests_table, what, whitelist)
+  local function add_dns_request(task, req, forced, is_ip, requests_table, label, whitelist)
     local req_str = req
     if is_ip then
       req_str = tostring(req)
     end
 
-    if whitelist and is_whitelisted(task, req, req_str, whitelist, what) then
+    if whitelist and is_whitelisted(task, req, req_str, whitelist, label) then
       return
     end
 
@@ -328,6 +366,9 @@ local function gen_rbl_callback(rule)
       if forced and not nreq.forced then
         nreq.forced = true
       end
+      if not nreq.what[label] then
+        nreq.what[label] = true
+      end
 
       return true,nreq -- Duplicate
     else
@@ -343,22 +384,22 @@ local function gen_rbl_callback(rule)
             n = processed,
             orig = req_str,
             resolve_ip = resolve_ip,
-            what = what,
+            what = {[label] = true},
           }
           requests_table[req] = nreq
         end
       else
         local to_resolve
-        local orign = req
+        local origin = req
 
         if not resolve_ip then
-          orign = maybe_make_hash(req, rule)
+          origin = maybe_make_hash(req, rule)
           to_resolve = string.format('%s.%s',
-              orign,
+              origin,
               rule.rbl)
         else
           -- First, resolve origin stuff without hashing or anything
-          to_resolve = orign
+          to_resolve = origin
         end
 
         nreq = {
@@ -366,7 +407,7 @@ local function gen_rbl_callback(rule)
           n = to_resolve,
           orig = req_str,
           resolve_ip = resolve_ip,
-          what = what,
+          what = {[label] = true},
         }
         requests_table[req] = nreq
       end
@@ -388,6 +429,14 @@ local function gen_rbl_callback(rule)
     return true
   end
 
+  local function check_required_symbols(task, _)
+    if rule.require_symbols then
+      return fun.all(function(sym) task:has_symbol(sym) end, rule.require_symbols)
+    end
+
+    return true
+  end
+
   local function check_user(task, _)
     if task:get_user() then
       return false
@@ -419,6 +468,8 @@ local function gen_rbl_callback(rule)
 
     add_dns_request(task, helo, true, false, requests_table,
         'helo', whitelist)
+
+    return true
   end
 
   local function check_dkim(task, requests_table, whitelist)
@@ -431,7 +482,8 @@ local function gen_rbl_callback(rule)
         -- We check merely mime from
         mime_from_domain = ((task:get_from('mime') or E)[1] or E).domain
         if mime_from_domain then
-          local mime_from_domain_tld = rspamd_util.get_tld(mime_from_domain)
+          local mime_from_domain_tld = rule.url_full_hostname and
+               mime_from_domain or rspamd_util.get_tld(mime_from_domain)
 
           if rule.url_compose_map then
             mime_from_domain = rule.url_compose_map:process_url(task, mime_from_domain_tld, mime_from_domain)
@@ -456,6 +508,8 @@ local function gen_rbl_callback(rule)
 
               if rule.url_compose_map then
                 domain_tld = rule.url_compose_map:process_url(task, domain_tld, domain)
+              elseif rule.url_full_hostname then
+                domain_tld = domain
               end
             end
 
@@ -468,6 +522,8 @@ local function gen_rbl_callback(rule)
               local domain_tld = rspamd_util.get_tld(domain)
               if rule.url_compose_map then
                 domain_tld = rule.url_compose_map:process_url(task, domain_tld, domain)
+              elseif rule.url_full_hostname then
+                domain_tld = domain
               end
               add_dns_request(task, domain_tld,
                   false, false, requests_table, 'dkim', whitelist)
@@ -496,19 +552,47 @@ local function gen_rbl_callback(rule)
       ignore_ip = rule.no_ip,
       need_images = rule.images,
       need_emails = false,
+      need_content = rule.content_urls or false,
       esld_limit = esld_lim,
       no_cache = true,
     }
 
+    if not rule.urls then
+      ex_params.flags_mode = 'explicit'
+      ex_params.flags = {}
+      if rule.content_urls then
+        table.insert(ex_params.flags, 'content')
+      end
+      if rule.images then
+        table.insert(ex_params.flags, 'image')
+      end
+    end
+
     local urls = lua_util.extract_specific_urls(ex_params)
 
     for _,u in ipairs(urls) do
-      local url_tld = u:get_tld()
-      if rule.url_compose_map then
-        url_tld = rule.url_compose_map:process_url(task, url_tld, u:get_host())
+      local flags = u:get_flags_num()
+
+      if bit.band(flags, url_flag_bits.numeric) ~= 0 then
+        -- For numeric urls we convert data to the ip address and
+        -- reverse octets. See #3948 for details
+        local to_resolve = u:get_host()
+        local addr = rspamd_ip.from_string(to_resolve)
+
+        if addr then
+          to_resolve = table.concat(addr:inversed_str_octets(), ".")
+        end
+        add_dns_request(task, to_resolve, false,
+            false, requests_table, 'url', whitelist)
+      else
+        local url_hostname = u:get_host()
+        local url_tld = rule.url_full_hostname and url_hostname or u:get_tld()
+        if rule.url_compose_map then
+          url_tld = rule.url_compose_map:process_url(task, url_tld, url_hostname)
+        end
+        add_dns_request(task, url_tld, false,
+            false, requests_table, 'url', whitelist)
       end
-      add_dns_request(task, url_tld, false,
-          false, requests_table, 'url', whitelist)
     end
 
     return true
@@ -562,12 +646,17 @@ local function gen_rbl_callback(rule)
   end
 
   local function check_selector(task, requests_table, whitelist)
-    local res = rule.selector(task)
+    for selector_label, selector in pairs(rule.selectors) do
+      local res = selector(task)
 
-    if res then
-      for _,r in ipairs(res) do
-        add_dns_request(task, r, false, false, requests_table,
-            'sel' .. rule.selector_id, whitelist)
+      if res and type(res) == 'table' then
+          for _,r in ipairs(res) do
+            add_dns_request(task, r, false, false, requests_table,
+                    selector_label, whitelist)
+          end
+      elseif res then
+        add_dns_request(task, res, false, false,
+                requests_table, selector_label, whitelist)
       end
     end
 
@@ -576,7 +665,14 @@ local function gen_rbl_callback(rule)
 
   local function check_email_table(task, email_tbl, requests_table, whitelist, what)
     lua_util.remove_email_aliases(email_tbl)
-    email_tbl.addr = email_tbl.addr:lower()
+    email_tbl.domain = email_tbl.domain:lower()
+    email_tbl.user = email_tbl.user:lower()
+
+    if email_tbl.domain == '' or email_tbl.user == '' then
+      rspamd_logger.infox(task, "got an email with some empty parts: '%s@%s'; skip it in the checks",
+          email_tbl.user, email_tbl.domain)
+      return
+    end
 
     if rule.emails_domainonly then
       add_dns_request(task, email_tbl.domain, false, false, requests_table,
@@ -624,7 +720,7 @@ local function gen_rbl_callback(rule)
 
     for _,email in ipairs(emails) do
       local domain
-      if rule.emails_domainonly then
+      if rule.emails_domainonly and not rule.url_full_hostname then
         if rule.url_compose_map then
           domain = rule.url_compose_map:process_url(task, email:get_tld(), email:get_host())
         else
@@ -635,8 +731,8 @@ local function gen_rbl_callback(rule)
       end
 
       local email_tbl = {
-        domain = domain,
-        user = email:get_user(),
+        domain = domain or '',
+        user = email:get_user() or '',
         addr = tostring(email),
       }
       check_email_table(task, email_tbl, requests_table, whitelist, 'email')
@@ -665,7 +761,8 @@ local function gen_rbl_callback(rule)
 
   -- Create function pipeline depending on rbl settings
   local pipeline = {
-    is_alive, -- generic for all
+    is_alive, -- check monitored status
+    check_required_symbols -- if we have require_symbols then check those symbols
   }
   local description = {
     'alive',
@@ -700,7 +797,7 @@ local function gen_rbl_callback(rule)
     description[#description + 1] = 'replyto'
   end
 
-  if rule.urls then
+  if rule.urls or rule.content_urls or rule.images then
     pipeline[#pipeline + 1] = check_urls
     description[#description + 1] = 'urls'
   end
@@ -760,8 +857,9 @@ local function gen_rbl_callback(rule)
             -- Check if we have rspamd{ip} userdata
             if type(dns_res) == 'userdata' then
               -- Add result as an actual RBL request
+              local label = next(orig_resolve_table_elt.what)
               local dup,nreq = add_dns_request(task, dns_res, false, true,
-                  resolved_req, orig_resolve_table_elt.what)
+                  resolved_req, label)
               -- Add original name
               if not dup then
                 nreq.orig = nreq.orig .. ':' .. orig_resolve_table_elt.n
@@ -853,26 +951,38 @@ local function add_rbl(key, rbl, global_opts)
   end
 
   if rbl.selector then
-    if known_selectors[rbl.selector] then
-      lua_util.debugm(N, rspamd_config, 'reuse selector id %s',
-          known_selectors[rbl.selector].id)
-      rbl.selector = known_selectors[rbl.selector].selector
-      rbl.selector_id = known_selectors[rbl.selector].id
-    else
-      -- Create a new flattened closure
-      local sel = selectors.create_selector_closure(rspamd_config, rbl.selector, '', true)
 
-      if not sel then
-        rspamd_logger.errx('invalid selector for rbl rule %s: %s', key, rbl.selector)
-        return false
+    rbl.selectors = {}
+    if type(rbl.selector) ~= 'table' then
+      rbl.selector = {['selector'] = rbl.selector}
+    end
+
+    for selector_label, selector in pairs(rbl.selector) do
+      if known_selectors[selector] then
+        lua_util.debugm(N, rspamd_config, 'reuse selector id %s',
+            known_selectors[selector].id)
+        rbl.selectors[selector_label] = known_selectors[selector].selector
+      else
+
+        if type(rbl.selector_flatten) ~= 'boolean' then
+          -- Fail-safety
+          rbl.selector_flatten = true
+        end
+        local sel = selectors.create_selector_closure(rspamd_config, selector, '',
+                rbl.selector_flatten)
+
+        if not sel then
+          rspamd_logger.errx('invalid selector for rbl rule %s: %s', key, selector)
+          return false
+        end
+
+        rbl.selector = sel
+        known_selectors[selector] = {
+          selector = sel,
+          id = #lua_util.keys(known_selectors) + 1,
+        }
+        rbl.selectors[selector_label] = known_selectors[selector].selector
       end
-
-      rbl.selector = sel
-      known_selectors[rbl.selector] = {
-        selector = sel,
-        id = #lua_util.keys(known_selectors) + 1,
-      }
-      rbl.selector_id = known_selectors[rbl.selector].id
     end
 
   end
@@ -939,21 +1049,33 @@ local function add_rbl(key, rbl, global_opts)
         rspamd_config:register_symbol{
           type = 'virtual',
           parent = id,
+          group = 'rbl',
+          score = 0,
           name = prefix .. '_' .. rbl.symbol,
         }
       end
-      if not rbl.is_whitelist and rbl.ignore_whitelist == false then
+      if not (rbl.is_whitelist or rbl.ignore_whitelist) then
         table.insert(black_symbols, rbl.symbol .. '_CHECK')
+      else
+        lua_util.debugm(N, rspamd_config, 'rule %s ignores whitelists: rbl.is_whitelist = %s, ' ..
+            'rbl.ignore_whitelist = %s',
+            rbl.symbol, rbl.is_whitelist, rbl.ignore_whitelist)
       end
     else
       id = rspamd_config:register_symbol{
         type = 'callback',
         callback = callback,
         name = rbl.symbol,
+        group = 'rbl',
+        score = 0,
         flags = table.concat(flags_tbl, ',')
       }
-      if not rbl.is_whitelist and rbl.ignore_whitelist == false then
+      if not (rbl.is_whitelist or rbl.ignore_whitelist) then
         table.insert(black_symbols, rbl.symbol)
+      else
+        lua_util.debugm(N, rspamd_config, 'rule %s ignores whitelists: rbl.is_whitelist = %s, ' ..
+            'rbl.ignore_whitelist = %s',
+            rbl.symbol, rbl.is_whitelist, rbl.ignore_whitelist)
       end
     end
 
@@ -967,9 +1089,16 @@ local function add_rbl(key, rbl, global_opts)
       rspamd_config:register_dependency(rbl.symbol, 'DKIM_CHECK')
     end
 
+    if rbl.require_symbols then
+      for _,dep in ipairs(rbl.require_symbols) do
+        rspamd_config:register_dependency(rbl.symbol, dep)
+      end
+    end
+
     -- Failure symbol
     rspamd_config:register_symbol{
-      type = 'virtual,nostat',
+      type = 'virtual',
+      flags = 'nostat',
       name = rbl.symbol .. '_FAIL',
       parent = id,
       score = 0.0,
@@ -984,6 +1113,8 @@ local function add_rbl(key, rbl, global_opts)
             type = 'virtual',
             parent = id,
             name = s,
+            group = 'rbl',
+            score = 0,
           }
         end
         if rbl.is_whitelist then
@@ -1002,7 +1133,7 @@ local function add_rbl(key, rbl, global_opts)
             table.insert(white_symbols, s)
           end
         else
-          if rbl.ignore_whitelist == false then
+          if not rbl.ignore_whitelist then
             table.insert(black_symbols, s)
           end
         end
@@ -1054,102 +1185,35 @@ end
 
 -- Plugin defaults should not be changed - override these in config
 -- New defaults should not alter behaviour
-local default_options = {
-  ['default_enabled'] = true,
-  ['default_ipv4'] = true,
-  ['default_ipv6'] = true,
-  ['default_received'] = false,
-  ['default_from'] = true,
-  ['default_unknown'] = false,
-  ['default_rdns'] = false,
-  ['default_helo'] = false,
-  ['default_dkim'] = false,
-  ['default_dkim_domainonly'] = true,
-  ['default_emails'] = false,
-  ['default_urls'] = false,
-  ['default_emails_domainonly'] = false,
-  ['default_exclude_private_ips'] = true,
-  ['default_exclude_users'] = false,
-  ['default_exclude_local'] = true,
-  ['default_is_whitelist'] = false,
-  ['default_ignore_whitelist'] = false,
-  ['default_resolve_ip'] = false,
-  ['default_no_ip'] = false,
-  ['default_images'] = false,
-  ['default_replyto'] = false,
-  ['default_dkim_match_from'] = false,
-}
 
-opts = lua_util.override_defaults(default_options, opts)
+
+opts = lua_util.override_defaults(rbl_common.default_options, opts)
+
+if opts.rules and opts.rbls then
+  -- Common issue :(
+  rspamd_logger.infox(rspamd_config, 'merging `rules` and `rbls` keys for compatibility')
+  opts.rbls = lua_util.override_defaults(opts.rbls, opts.rules)
+end
 
 if(opts['local_exclude_ip_map'] ~= nil) then
   local_exclusions = lua_maps.map_add(N, 'local_exclude_ip_map', 'radix',
     'RBL exclusions map')
 end
 
-local return_codes_schema = ts.map_of(
-    ts.string / string.upper, -- Symbol name
-    (
-        ts.array_of(ts.string) +
-            (ts.string / function(s)
-              return { s }
-            end) -- List of IP patterns
-    )
-)
-local return_bits_schema = ts.map_of(
-    ts.string / string.upper, -- Symbol name
-    (
-        ts.array_of(ts.number + ts.string / tonumber) +
-            (ts.string / function(s)
-              return { tonumber(s) }
-            end) +
-            (ts.number / function(s)
-              return { s }
-            end)
-    )
-)
-
-local rule_schema_tbl = {
-  enabled = ts.boolean:is_optional(),
-  disabled = ts.boolean:is_optional(),
-  rbl = ts.string,
-  selector = ts.string:is_optional(),
-  symbol = ts.string:is_optional(),
-  returncodes = return_codes_schema:is_optional(),
-  return_codes = return_codes_schema:is_optional(),
-  returnbits = return_bits_schema:is_optional(),
-  return_bits = return_bits_schema:is_optional(),
-  whitelist_exception = (
-      ts.array_of(ts.string) + (ts.string / function(s) return {s} end)
-  ):is_optional(),
-  whitelist = lua_maps.map_schema:is_optional(),
-  url_compose_map = lua_maps.map_schema:is_optional(),
-  local_exclude_ip_map = ts.string:is_optional(),
-  hash = ts.one_of{"sha1", "sha256", "sha384", "sha512", "md5", "blake2"}:is_optional(),
-  hash_format = ts.one_of{"hex", "base32", "base64"}:is_optional(),
-  hash_len = (ts.integer + ts.string / tonumber):is_optional(),
-  monitored_address = ts.string:is_optional(),
-  requests_limit = (ts.integer + ts.string / tonumber):is_optional(),
-  process_script = ts.string:is_optional(),
-  emails_delimiter = ts.string:is_optional(),
-  ignore_defaults = ts.boolean:is_optional(),
-  disable_monitoring = ts.boolean:is_optional(),
-  symbols_prefixes = ts.map_of(ts.string, ts.string):is_optional(),
-}
--- Add default boolean flags to the schema
-for def_k,_ in pairs(default_options) do
-  rule_schema_tbl[def_k:sub(#('default_') + 1)] = ts.boolean:is_optional()
-end
-
-local rule_schema = ts.shape(rule_schema_tbl)
-
-for key,rbl in pairs(opts.rbls or opts.rules) do
+for key,rbl in pairs(opts.rbls ) do
   if type(rbl) ~= 'table' or rbl.disabled == true or rbl.enabled == false then
     rspamd_logger.infox(rspamd_config, 'disable rbl "%s"', key)
   else
+    -- Aliases
+    if type(rbl.ignore_default) == 'boolean' then
+      rbl.ignore_defaults = rbl.ignore_default
+    end
+    if type(rbl.ignore_whitelists) == 'boolean' then
+      rbl.ignore_whitelist = rbl.ignore_whitelists
+    end
     -- Propagate default options from opts to rule
     if not rbl.ignore_defaults then
-      for default_opt_key,_ in pairs(default_options) do
+      for default_opt_key,_ in pairs(rbl_common.default_options) do
         local rbl_opt = default_opt_key:sub(#('default_') + 1)
         if rbl[rbl_opt] == nil then
           rbl[rbl_opt] = opts[default_opt_key]
@@ -1161,15 +1225,22 @@ for key,rbl in pairs(opts.rbls or opts.rules) do
       rbl.requests_limit = rspamd_config:get_dns_max_requests()
     end
 
-    local res,err = rule_schema:transform(rbl)
+    local res,err = rbl_common.rule_schema:transform(rbl)
     if not res then
       rspamd_logger.errx(rspamd_config, 'invalid config for %s: %s, RBL is DISABLED',
           key, err)
     else
+      res = rbl_common.convert_checks(res)
       -- Aliases
       if res.return_codes then res.returncodes = res.return_codes end
       if res.return_bits then res.returnbits = res.return_bits end
-      add_rbl(key, res, opts)
+
+      if not res then
+        rspamd_logger.errx(rspamd_config, 'invalid config for %s: %s, RBL is DISABLED',
+            key, err)
+      else
+        add_rbl(key, res, opts)
+      end
     end
   end -- rbl.enabled
 end
@@ -1189,7 +1260,10 @@ local function rbl_callback_white(task)
         lua_util.debugm(N, task,'found whitelist from %s: %s(%s)', w,
             elt, what)
         if elt and what then
-          whitelisted_elements[elt] = what
+          whitelisted_elements[elt] = {
+            type = what,
+            symbol = w,
+          }
         end
       end
     end
@@ -1209,14 +1283,16 @@ rspamd_config:register_symbol{
   type = 'callback',
   callback = rbl_callback_white,
   name = 'RBL_CALLBACK_WHITE',
-  flags = 'nice,empty,no_squeeze'
+  flags = 'nice,empty,no_squeeze',
+  augmentations = {string.format("timeout=%f", rspamd_config:get_dns_timeout() or 0.0)},
 }
 
 rspamd_config:register_symbol{
   type = 'callback',
   callback = rbl_callback_fin,
   name = 'RBL_CALLBACK',
-  flags = 'empty,no_squeeze'
+  flags = 'empty,no_squeeze',
+  augmentations = {string.format("timeout=%f", rspamd_config:get_dns_timeout() or 0.0)},
 }
 
 for _, w in ipairs(white_symbols) do
