@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2017-2018, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -70,14 +70,29 @@ local function generic_reputation_calc(token, rule, mult, task)
 end
 
 local function add_symbol_score(task, rule, mult, params)
-  if not params then params = {tostring(mult)};
-
+  if not params then
+    params = {tostring(mult)}
   end
+
   if rule.selector.config.split_symbols then
+    local sym_spam = rule.symbol .. '_SPAM'
+    local sym_ham = rule.symbol .. '_HAM'
+    if not rule.static_symbols then
+      rule.static_symbols = {}
+      rule.static_symbols.ham = rspamd_config:get_symbol(sym_ham)
+      rule.static_symbols.spam = rspamd_config:get_symbol(sym_spam)
+    end
     if mult >= 0 then
-      task:insert_result(rule.symbol .. '_SPAM', mult, params)
+      task:insert_result(sym_spam, mult, params)
     else
-      task:insert_result(rule.symbol .. '_HAM', mult, params)
+      -- Avoid multiplication of negative the `mult` by negative static score of the
+      -- ham symbol
+      if rule.static_symbols.ham and rule.static_symbols.ham.score then
+        if rule.static_symbols.ham.score < 0 then
+          mult = math.abs(mult)
+        end
+      end
+      task:insert_result(sym_ham, mult, params)
     end
   else
     task:insert_result(rule.symbol, mult, params)
@@ -157,39 +172,53 @@ end
 local function dkim_reputation_filter(task, rule)
   local requests = gen_dkim_queries(task, rule)
   local results = {}
-  local nchecked = 0
+  local dkim_tlds = lua_util.keys(requests)
+  local requests_left = #dkim_tlds
   local rep_accepted = 0.0
-  local rep_rejected = 0.0
 
   lua_util.debugm(N, task, 'dkim reputation tokens: %s', requests)
 
   local function tokens_cb(err, token, values)
-    nchecked = nchecked + 1
+    requests_left = requests_left - 1
 
     if values then
       results[token] = values
     end
 
-    if nchecked == #requests then
+    if requests_left == 0 then
       for k,v in pairs(results) do
-        if requests[k] == 'a' then
-          rep_accepted = rep_accepted + generic_reputation_calc(v, rule, 1.0, task)
-        elseif requests[k] == 'r' then
-          rep_rejected = rep_rejected + generic_reputation_calc(v, rule, 1.0, task)
+        -- `k` in results is a prefixed and suffixed tld, so we need to look through
+        -- all requests to find any request with the matching tld
+        local sel_tld
+        for _,tld in ipairs(dkim_tlds) do
+          if k:find(tld, 1, true) then
+            sel_tld = tld
+            break
+          end
+        end
+
+        if sel_tld and requests[sel_tld] then
+          if requests[sel_tld] == 'a' then
+            rep_accepted = rep_accepted + generic_reputation_calc(v, rule, 1.0, task)
+          end
+        else
+          rspamd_logger.warnx(task, "cannot find the requested tld for a request: %s (%s tlds noticed)",
+              k, dkim_tlds)
         end
       end
 
       -- Set local reputation symbol
-      if rep_accepted > 0 or rep_rejected > 0 then
-        if rep_accepted > rep_rejected then
-          add_symbol_score(task, rule, -(rep_accepted - rep_rejected))
-        else
-          add_symbol_score(task, rule, (rep_rejected - rep_accepted))
-        end
+      local rep_accepted_abs = math.abs(rep_accepted or 0)
+      lua_util.debugm(N, task, "dkim reputation accepted: %s",
+          rep_accepted_abs)
+      if rep_accepted_abs then
+        local final_rep = rep_accepted
+        if rep_accepted > 1.0 then final_rep = 1.0 end
+        if rep_accepted < -1.0 then final_rep = -1.0 end
+        add_symbol_score(task, rule, final_rep)
 
         -- Store results for future DKIM results adjustments
         task:get_mempool():set_variable("dkim_reputation_accept", tostring(rep_accepted))
-        task:get_mempool():set_variable("dkim_reputation_reject", tostring(rep_rejected))
       end
     end
   end
@@ -197,7 +226,7 @@ local function dkim_reputation_filter(task, rule)
   for dom,res in pairs(requests) do
     -- tld + "." + check_result, e.g. example.com.+ - reputation for valid sigs
     local query = string.format('%s.%s', dom, res)
-    rule.backend.get_token(task, rule, query, tokens_cb)
+    rule.backend.get_token(task, rule, nil, query, tokens_cb, 'string')
   end
 end
 
@@ -209,28 +238,26 @@ local function dkim_reputation_idempotent(task, rule)
     for dom,res in pairs(requests) do
       -- tld + "." + check_result, e.g. example.com.+ - reputation for valid sigs
       local query = string.format('%s.%s', dom, res)
-      rule.backend.set_token(task, rule, query, sc)
+      rule.backend.set_token(task, rule, nil, query, sc)
     end
   end
 end
 
 local function dkim_reputation_postfilter(task, rule)
-  local sym_accepted = task:get_symbol('R_DKIM_ALLOW')
+  local sym_accepted = (task:get_symbol('R_DKIM_ALLOW') or E)[1]
   local accept_adjustment = task:get_mempool():get_variable("dkim_reputation_accept")
+  local cfg = rule.selector.config or E
 
-  if sym_accepted and accept_adjustment then
-    local final_adjustment = rule.config.max_accept_adjustment *
-        rspamd_util.tanh(tonumber(accept_adjustment))
-    task:adjust_result('R_DKIM_ALLOW', sym_accepted.score * final_adjustment)
-  end
+  if sym_accepted and sym_accepted.score and
+      accept_adjustment and type(cfg.max_accept_adjustment) == 'number' then
+    local final_adjustment = cfg.max_accept_adjustment *
+        rspamd_util.tanh(tonumber(accept_adjustment) or 0)
+    lua_util.debugm(N, task, "adjust DKIM_ALLOW: " ..
+        "cfg.max_accept_adjustment=%s accept_adjustment=%s final_adjustment=%s sym_accepted.score=%s",
+        cfg.max_accept_adjustment, accept_adjustment, final_adjustment,
+        sym_accepted.score)
 
-  local sym_rejected = task:get_symbol('R_DKIM_REJECT')
-  local reject_adjustment = task:get_mempool():get_variable("dkim_reputation_reject")
-
-  if sym_rejected and reject_adjustment then
-    local final_adjustment = rule.config.max_reject_adjustment *
-        rspamd_util.tanh(tonumber(reject_adjustment))
-    task:adjust_result('R_DKIM_REJECT', sym_rejected.score * final_adjustment)
+    task:adjust_result('R_DKIM_ALLOW', sym_accepted.score + final_adjustment)
   end
 end
 
@@ -243,7 +270,6 @@ local dkim_selector = {
     outbound = true,
     inbound = true,
     max_accept_adjustment = 2.0, -- How to adjust accepted DKIM score
-    max_reject_adjustment = 3.0 -- How to adjust rejected DKIM score
   },
   dependencies = {"DKIM_TRACE"},
   filter = dkim_reputation_filter, -- used to get scores
@@ -286,30 +312,54 @@ end
 
 local function url_reputation_filter(task, rule)
   local requests = gen_url_queries(task, rule)
+  local url_keys = lua_util.keys(requests)
+  local requests_left = #url_keys
   local results = {}
-  local nchecked = 0
 
   local function indexed_tokens_cb(err, index, values)
-    nchecked = nchecked + 1
+    requests_left = requests_left - 1
 
     if values then
       results[index] = values
     end
 
-    if nchecked == #requests then
+    if requests_left == 0 then
       -- Check the url with maximum hits
       local mhits = 0
+
+      local result_request_match_tbl = {}
+      -- XXX: ugly O(N^2) loop to match requests and responses
+      for _,res_pair in ipairs(results) do
+        local result_k = res_pair[1]
+        for _, request_k in ipairs(url_keys) do
+          if result_k:find(request_k) then
+            result_request_match_tbl[result_k] = request_k
+            break
+          end
+        end
+      end
+
       for k,_ in pairs(results) do
-        if requests[k][2] > mhits then
-          mhits = requests[k][2]
+        local req = result_request_match_tbl[k]
+        if req then
+          if requests[req] > mhits then
+            mhits = requests[req][2]
+          else
+            rspamd_logger.warnx(task, "cannot find the requested response for a request: %s (%s requests noticed)",
+                k, url_keys)
+          end
         end
       end
 
       if mhits > 0 then
         local score = 0
         for k,v in pairs(results) do
-          score = score + generic_reputation_calc(v, rule,
-              requests[k][2] / mhits, task)
+          local req = result_request_match_tbl[k]
+          if req then
+            score = score + generic_reputation_calc(v, rule,
+                requests[req][2] / mhits, task)
+
+          end
         end
 
         if math.abs(score) > 1e-3 then
@@ -325,7 +375,7 @@ local function url_reputation_filter(task, rule)
       indexed_tokens_cb(err, i, values)
     end
 
-    rule.backend.get_token(task, rule, req[1], tokens_cb)
+    rule.backend.get_token(task, rule, nil, req[1], tokens_cb, 'string')
   end
 end
 
@@ -335,7 +385,7 @@ local function url_reputation_idempotent(task, rule)
 
   if sc then
     for _,tld in ipairs(requests) do
-      rule.backend.set_token(task, rule, tld[1], sc)
+      rule.backend.set_token(task, rule, nil, tld[1], sc)
     end
   end
 end
@@ -411,19 +461,22 @@ local function ip_reputation_filter(task, rule)
     if asn_stats then
       local asn_score = generic_reputation_calc(asn_stats, rule, cfg.scores.asn, task)
       score = score + asn_score
-      table.insert(description_t, string.format('asn: %s(%.2f)', asn, asn_score))
+      table.insert(description_t, string.format('asn: %s(%.2f)',
+              asn, asn_score))
     end
     if country_stats then
       local country_score = generic_reputation_calc(country_stats, rule,
           cfg.scores.country, task)
       score = score + country_score
-      table.insert(description_t, string.format('country: %s(%.2f)', country, country_score))
+      table.insert(description_t, string.format('country: %s(%.2f)',
+              country, country_score))
     end
     if ip_stats then
       local ip_score = generic_reputation_calc(ip_stats, rule, cfg.scores.ip,
         task)
       score = score + ip_score
-      table.insert(description_t, string.format('ip: %s(%.2f)', ip, ip_score))
+      table.insert(description_t, string.format('ip: %s(%.2f)',
+              tostring(ip), ip_score))
     end
 
     if math.abs(score) > 0.001 then
@@ -462,13 +515,16 @@ local function ip_reputation_filter(task, rule)
   end
 
   if asn then
-    rule.backend.get_token(task, rule, cfg.asn_prefix .. asn, gen_token_callback('asn'))
+    rule.backend.get_token(task, rule, cfg.asn_prefix, asn,
+            gen_token_callback('asn'), 'string')
   end
   if country then
-    rule.backend.get_token(task, rule, cfg.country_prefix .. country, gen_token_callback('country'))
+    rule.backend.get_token(task, rule, cfg.country_prefix, country,
+            gen_token_callback('country'), 'string')
   end
 
-  rule.backend.get_token(task, rule, cfg.ip_prefix .. tostring(ip), gen_token_callback('ip'))
+  rule.backend.get_token(task, rule, cfg.ip_prefix, ip,
+          gen_token_callback('ip'), 'ip')
 end
 
 -- Used to set scores
@@ -502,13 +558,13 @@ local function ip_reputation_idempotent(task, rule)
   local sc = extract_task_score(task, rule)
   if sc then
     if asn then
-      rule.backend.set_token(task, rule, cfg.asn_prefix .. asn, sc)
+      rule.backend.set_token(task, rule, cfg.asn_prefix, asn, sc, nil, 'string')
     end
     if country then
-      rule.backend.set_token(task, rule, cfg.country_prefix .. country, sc)
+      rule.backend.set_token(task, rule, cfg.country_prefix, country, sc, nil, 'string')
     end
 
-    rule.backend.set_token(task, rule, cfg.ip_prefix .. tostring(ip), sc)
+    rule.backend.set_token(task, rule, cfg.ip_prefix, ip, sc, nil, 'ip')
   end
 end
 
@@ -565,7 +621,7 @@ local function spf_reputation_filter(task, rule)
     end
   end
 
-  rule.backend.get_token(task, rule, hkey, tokens_cb)
+  rule.backend.get_token(task, rule, nil, hkey, tokens_cb, 'string')
 end
 
 local function spf_reputation_idempotent(task, rule)
@@ -580,7 +636,7 @@ local function spf_reputation_idempotent(task, rule)
 
   lua_util.debugm(N, task, 'set spf record %s -> %s = %s',
       spf_record, hkey, sc)
-  rule.backend.set_token(task, rule, hkey, sc)
+  rule.backend.set_token(task, rule, nil, hkey, sc)
 end
 
 
@@ -593,8 +649,6 @@ local spf_selector = {
     max_score = nil,
     outbound = true,
     inbound = true,
-    max_accept_adjustment = 2.0, -- How to adjust accepted DKIM score
-    max_reject_adjustment = 3.0 -- How to adjust rejected DKIM score
   },
   dependencies = {"R_SPF_ALLOW"},
   filter = spf_reputation_filter, -- used to get scores
@@ -652,12 +706,12 @@ local function generic_reputation_filter(task, rule)
       fun.each(function(e)
         lua_util.debugm(N, task, 'check generic reputation (%s) %s',
           rule['symbol'], e)
-        rule.backend.get_token(task, rule, e, tokens_cb)
+        rule.backend.get_token(task, rule, nil, e, tokens_cb, 'string')
       end, selector_res)
     else
       lua_util.debugm(N, task, 'check generic reputation (%s) %s',
         rule['symbol'], selector_res)
-      rule.backend.get_token(task, rule, selector_res, tokens_cb)
+      rule.backend.get_token(task, rule, nil, selector_res, tokens_cb, 'string')
     end
   end
 end
@@ -674,12 +728,12 @@ local function generic_reputation_idempotent(task, rule)
       fun.each(function(e)
         lua_util.debugm(N, task, 'set generic selector (%s) %s = %s',
             rule['symbol'], e, sc)
-        rule.backend.set_token(task, rule, e, sc)
+        rule.backend.set_token(task, rule, nil, e, sc)
       end, selector_res)
     else
       lua_util.debugm(N, task, 'set generic selector (%s) %s = %s',
           rule['symbol'], selector_res, sc)
-      rule.backend.set_token(task, rule, selector_res, sc)
+      rule.backend.set_token(task, rule, nil, selector_res, sc)
     end
   end
 end
@@ -733,7 +787,10 @@ local function reputation_dns_init(rule, _, _, _)
 end
 
 
-local function gen_token_key(token, rule)
+local function gen_token_key(prefix, token, rule)
+  if prefix then
+    token = prefix .. token
+  end
   local res = token
   if rule.backend.config.hashed then
     local hash_alg = rule.backend.config.hash_alg or "blake2"
@@ -766,7 +823,7 @@ end
 
 --[[
 -- Generic interface for get and set tokens functions:
--- get_token(task, rule, token, continuation), where `continuation` is the following function:
+-- get_token(task, rule, prefix, token, continuation, token_type), where `continuation` is the following function:
 --
 -- function(err, token, values) ... end
 -- `err`: string value for error (similar to redis or DNS callbacks)
@@ -785,37 +842,57 @@ end
 -- example of tokens: {'s': 0, 'h': 0, 'p': 1}
 --]]
 
-local function reputation_dns_get_token(task, rule, token, continuation_cb)
+local function reputation_dns_get_token(task, rule, prefix, token, continuation_cb, token_type)
   -- local r = task:get_resolver()
-  local key = gen_token_key(token, rule)
+  -- In DNS we never ever use prefix as prefix, we use if as a suffix!
+  if token_type == 'ip' then
+    token = table.concat(token:inversed_str_octets(), '.')
+  end
+
+  local key = gen_token_key(nil, token, rule)
   local dns_name = key .. '.' .. rule.backend.config.list
 
+  if prefix then
+    dns_name = string.format('%s.%s.%s', key, prefix,
+            rule.backend.config.list)
+  else
+    dns_name = string.format('%s.%s', key, rule.backend.config.list)
+  end
+
   local function dns_cb(_, _, results, err)
     if err and (err ~= 'requested record is not found' and
         err ~= 'no records with this name') then
-      rspamd_logger.errx(task, 'error looking up %s: %s', dns_name, err)
+      rspamd_logger.warnx(task, 'error looking up %s: %s', dns_name, err)
     end
 
     lua_util.debugm(N, task, 'DNS RESPONSE: label=%1 results=%2 err=%3 list=%4',
         dns_name, results, err, rule.backend.config.list)
 
     -- Now split tokens to list of values
-    if results  then
-      local values = {}
-      -- Format: key1=num1;key2=num2...keyn=numn
-      fun.each(function(e)
-        local vals = lua_util.rspamd_str_split(e, "=")
-        if vals and #vals == 2 then
-          local nv = tonumber(vals[2])
-          if nv then
-            values[vals[1]] = nv
-          end
-        end
+    if results and results[1]  then
+      -- Format: num_messages;sc1;sc2...scn
+      local dns_tokens = lua_util.rspamd_str_split(results[1], ";")
+      -- Convert all to numbers excluding any possible non-numbers
+      dns_tokens = fun.totable(fun.filter(function(e)
+        return type(e) == 'number'
       end,
-          lua_util.rspamd_str_split(results[1], ";"))
+      fun.map(function(e)
+        local n = tonumber(e)
+        if n then return n end
+        return "BAD"
+      end, dns_tokens)))
 
-      continuation_cb(nil, dns_name, values)
+      if #dns_tokens < 2 then
+        rspamd_logger.warnx(task, 'cannot parse response for reputation token %s: %s',
+                dns_name, results[1])
+        continuation_cb(results, dns_name, nil)
+      else
+        local cnt = table.remove(dns_tokens, 1)
+        continuation_cb(nil, dns_name, { cnt, dns_tokens })
+      end
     else
+      rspamd_logger.messagex(task, 'invalid response for reputation token %s: %s',
+              dns_name, results[1])
       continuation_cb(results, dns_name, nil)
     end
   end
@@ -918,8 +995,11 @@ local function reputation_redis_init(rule, cfg, ev_base, worker)
   return true
 end
 
-local function reputation_redis_get_token(task, rule, token, continuation_cb)
-  local key = gen_token_key(token, rule)
+local function reputation_redis_get_token(task, rule, prefix, token, continuation_cb, token_type)
+  if token_type and token_type == 'ip' then
+    token = tostring(token)
+  end
+  local key = gen_token_key(prefix, token, rule)
 
   local function redis_get_cb(err, data)
     if data then
@@ -953,8 +1033,11 @@ local function reputation_redis_get_token(task, rule, token, continuation_cb)
   end
 end
 
-local function reputation_redis_set_token(task, rule, token, sc, continuation_cb)
-  local key = gen_token_key(token, rule)
+local function reputation_redis_set_token(task, rule, prefix, token, sc, continuation_cb, token_type)
+  if token_type and token_type == 'ip' then
+    token = tostring(token)
+  end
+  local key = gen_token_key(prefix, token, rule)
 
   local function redis_set_cb(err, data)
     if err then
@@ -1219,6 +1302,7 @@ local function parse_rule(name, tbl)
     name = rule.symbol,
     type = rule_type,
     callback = callback_gen(reputation_filter_cb, rule),
+    augmentations = {string.format("timeout=%f", redis_params.timeout or 0.0)},
   }
 
   if rule.selector.config.split_symbols then
@@ -1245,8 +1329,9 @@ local function parse_rule(name, tbl)
     rspamd_config:register_symbol{
       name = rule.symbol .. '_POST',
       type = 'postfilter',
-      flags = 'nostat',
+      flags = 'nostat,explicit_disable,ignore_passthrough',
       callback = callback_gen(reputation_postfilter_cb, rule),
+      augmentations = {string.format("timeout=%f", redis_params.timeout or 0.0)},
     }
   end
 
@@ -1255,7 +1340,9 @@ local function parse_rule(name, tbl)
     rspamd_config:register_symbol{
       name = rule.symbol .. '_IDEMPOTENT',
       type = 'idempotent',
+      flags = 'explicit_disable,ignore_passthrough',
       callback = callback_gen(reputation_idempotent_cb, rule),
+      augmentations = {string.format("timeout=%f", redis_params.timeout or 0.0)},
     }
   end
 
