@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2017, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -59,7 +59,7 @@ local settings = {
 
 local function adjust_url(task, orig_url, redir_url)
   if type(redir_url) == 'string' then
-    redir_url = rspamd_url.create(task:get_mempool(), redir_url)
+    redir_url = rspamd_url.create(task:get_mempool(), redir_url, {'redirect_target'})
   end
 
   if redir_url then
@@ -107,7 +107,7 @@ local function cache_url(task, orig_url, url, key, param)
             redis_trim_cb, --callback
             'ZREMRANGEBYRANK', -- command
             {settings.top_urls_key, '0',
-              tostring(settings.top_urls_count + 1)} -- arguments
+              tostring(-(settings.top_urls_count + 1))} -- arguments
           )
           if not ret then
             rspamd_logger.errx(task, 'cannot trim top urls set')
@@ -356,7 +356,10 @@ if opts then
       local id = rspamd_config:register_symbol{
         name = 'URL_REDIRECTOR_CHECK',
         type = 'callback,prefilter',
+        priority = lua_util.symbols_priorities.medium,
         callback = url_redirector_handler,
+        -- In fact, the real timeout is nested_limit * timeout...
+        augmentations = {string.format("timeout=%f", settings.timeout)}
       }
 
       if settings.redirector_symbol then
