@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2016, Andrew Lewis <nerf@judo.za.org>
- * Copyright (c) 2016, Vsevolod Stakhov <vsevolod@highsecure.ru>
+ * Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -94,22 +94,22 @@ rspamadm_fuzzyconvert (gint argc, gchar **argv, const struct rspamadm_command *c
 		rspamd_fprintf (stderr, "option parsing failed: %s\n", error->message);
 		g_error_free (error);
 		g_option_context_free (context);
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 
 	g_option_context_free (context);
 
 	if (!source_db) {
 		rspamd_fprintf (stderr, "source db is missing\n");
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 	if (!redis_host) {
 		rspamd_fprintf (stderr, "redis host is missing\n");
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 	if (!fuzzy_expiry) {
 		rspamd_fprintf (stderr, "expiry is missing\n");
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 
 	obj = ucl_object_typed_new (UCL_OBJECT);
