@@ -129,7 +129,7 @@ rspamadm_statconvert (gint argc, gchar **argv, const struct rspamadm_command *cm
 		rspamd_fprintf (stderr, "option parsing failed: %s\n", error->message);
 		g_error_free (error);
 		g_option_context_free (context);
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 
 	g_option_context_free (context);
