@@ -36,7 +36,6 @@
 #include "libserver/milter.h"
 #include "libserver/milter_internal.h"
 #include "libmime/lang_detection.h"
-#include "contrib/zstd/zstd.h"
 
 #include <math.h>
 
@@ -44,26 +43,32 @@
 #include <netinet/tcp.h> /* for TCP_NODELAY */
 #endif
 
+#ifdef SYS_ZSTD
+#  include "zstd.h"
+#else
+#  include "contrib/zstd/zstd.h"
+#endif
+
 /* Rotate keys each minute by default */
 #define DEFAULT_ROTATION_TIME 60.0
 #define DEFAULT_RETRIES 5
 
 #define msg_err_session(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         session->pool->tag.tagname, session->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_session(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         session->pool->tag.tagname, session->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_session(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         session->pool->tag.tagname, session->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 #define msg_debug_session(...)  rspamd_conditional_debug_fast (NULL, session->client_addr, \
         rspamd_proxy_log_id, "proxy", session->pool->tag.uid, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 INIT_LOG_MODULE(proxy)
@@ -152,6 +157,7 @@ struct rspamd_proxy_ctx {
 	struct rspamd_milter_context milter_ctx;
 	/* Language detector */
 	struct rspamd_lang_detector *lang_det;
+	gdouble task_timeout;
 };
 
 enum rspamd_backend_flags {
@@ -323,12 +329,27 @@ rspamd_proxy_parse_upstream (rspamd_mempool_t *pool,
 	L = ctx->lua_state;
 
 	if (ucl_object_type (obj) != UCL_OBJECT) {
+		if (ucl_object_type (obj) != UCL_NULL) {
+			msg_info_pool_check("upstream %s is disabled by setting it to NULL",
+					ucl_object_key (obj));
+			return TRUE;
+		}
+
 		g_set_error (err, rspamd_proxy_quark (), 100,
 				"upstream option must be an object");
 
 		return FALSE;
 	}
 
+	if (!rspamd_config_is_enabled_from_ucl (pool, obj)) {
+		/* Upstream is valid but disabled */
+		msg_info_pool_check("upstream %s is disabled",
+				ucl_object_lookup (obj, "name") ?
+					ucl_object_tostring(ucl_object_lookup (obj, "name")) :
+					ucl_object_key (obj));
+		return TRUE;
+	}
+
 	up = rspamd_mempool_alloc0 (pool, sizeof (*up));
 
 	elt = ucl_object_lookup (obj, "name");
@@ -486,12 +507,27 @@ rspamd_proxy_parse_mirror (rspamd_mempool_t *pool,
 	L = ctx->lua_state;
 
 	if (ucl_object_type (obj) != UCL_OBJECT) {
+		if (ucl_object_type (obj) != UCL_NULL) {
+			msg_info_pool_check("mirror %s is disabled by setting it to NULL",
+					ucl_object_key (obj));
+			return TRUE;
+		}
+
 		g_set_error (err, rspamd_proxy_quark (), 100,
 				"mirror option must be an object");
 
 		return FALSE;
 	}
 
+	if (!rspamd_config_is_enabled_from_ucl (pool, obj)) {
+		/* Upstream is valid but disabled */
+		msg_info_pool_check("mirror %s is disabled",
+				ucl_object_lookup (obj, "name") ?
+				ucl_object_tostring(ucl_object_lookup (obj, "name")) :
+				ucl_object_key (obj));
+		return TRUE;
+	}
+
 	up = rspamd_mempool_alloc0 (pool, sizeof (*up));
 
 	elt = ucl_object_lookup (obj, "name");
@@ -976,13 +1012,17 @@ proxy_call_cmp_script (struct rspamd_proxy_session *session, gint cbref)
 		}
 	}
 
-	if (lua_pcall (L, 1, 0, err_idx) != 0) {
+	gchar log_tag[RSPAMD_LOG_ID_LEN + 1];
+	rspamd_strlcpy(log_tag, session->pool->tag.uid, sizeof(log_tag));
+	lua_pushstring(L, log_tag);
+
+	if (lua_pcall (L, 2, 0, err_idx) != 0) {
 		msg_err_session (
 				"cannot run lua compare script: %s",
 				lua_tostring (L, -1));
 	}
 
-	lua_settop (L, 0);
+	lua_settop (L, err_idx - 1);
 }
 
 static void
@@ -1431,7 +1471,7 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 
 			msg->method = HTTP_GET;
 			rspamd_http_connection_write_message_shared (bk_conn->backend_conn,
-					msg, NULL, NULL, bk_conn,
+					msg, rspamd_upstream_name(bk_conn->up), NULL, bk_conn,
 					bk_conn->timeout);
 		}
 		else {
@@ -1458,7 +1498,7 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 			}
 
 			rspamd_http_connection_write_message (bk_conn->backend_conn,
-					msg, NULL, NULL, bk_conn,
+					msg, rspamd_upstream_name(bk_conn->up), NULL, bk_conn,
 					bk_conn->timeout);
 		}
 
@@ -1539,11 +1579,21 @@ proxy_backend_master_error_handler (struct rspamd_http_connection *conn, GError
 		msg_err_session ("cannot connect to upstream, maximum retries "
 				"has been reached: %d", session->retries);
 		/* Terminate session immediately */
-		proxy_client_write_error (session, err->code, err->message);
+		if (err) {
+			proxy_client_write_error(session, err->code, err->message);
+		}
+		else {
+			proxy_client_write_error(session, 503, "Unknown error after no retries left");
+		}
 	}
 	else {
 		if (!proxy_send_master_message (session)) {
-			proxy_client_write_error (session, err->code, err->message);
+			if (err) {
+				proxy_client_write_error(session, err->code, err->message);
+			}
+			else {
+				proxy_client_write_error(session, 503, "Unknown error on write");
+			}
 		}
 		else {
 			msg_info_session ("retry connection to: %s"
@@ -1737,7 +1787,7 @@ rspamd_proxy_scan_self_reply (struct rspamd_task *task)
 				NULL,
 				ctype,
 				session,
-				0);
+				session->ctx->timeout / 10.0);
 	}
 }
 
@@ -1789,11 +1839,11 @@ rspamd_proxy_self_scan (struct rspamd_proxy_session *session)
 	task->sock = -1;
 
 	if (session->client_milter_conn) {
-		task->client_addr = rspamd_inet_address_copy (
-				session->client_milter_conn->addr);
+		task->client_addr = rspamd_inet_address_copy(
+				session->client_milter_conn->addr, NULL);
 	}
 	else {
-		task->client_addr = rspamd_inet_address_copy (session->client_addr);
+		task->client_addr = rspamd_inet_address_copy(session->client_addr, NULL);
 	}
 
 	task->fin_arg = session;
@@ -1837,7 +1887,7 @@ rspamd_proxy_self_scan (struct rspamd_proxy_session *session)
 
 	}
 	else if (session->ctx->has_self_scan) {
-		if (session->ctx->cfg->task_timeout > 0) {
+		if (!isnan(session->ctx->task_timeout) && session->ctx->task_timeout > 0) {
 			task->timeout_ev.data = task;
 			ev_timer_init (&task->timeout_ev, rspamd_task_timeout,
 					session->ctx->cfg->task_timeout,
@@ -1985,7 +2035,8 @@ retry:
 
 			rspamd_http_connection_write_message_shared (
 					session->master_conn->backend_conn,
-					msg, NULL, NULL, session->master_conn,
+					msg, rspamd_upstream_name(session->master_conn->up),
+					NULL, session->master_conn,
 					session->master_conn->timeout);
 		}
 		else {
@@ -2013,7 +2064,8 @@ retry:
 
 			rspamd_http_connection_write_message (
 					session->master_conn->backend_conn,
-					msg, NULL, NULL, session->master_conn,
+					msg, rspamd_upstream_name(session->master_conn->up),
+					NULL, session->master_conn,
 					session->master_conn->timeout);
 		}
 	}
@@ -2180,7 +2232,7 @@ proxy_accept_socket (EV_P_ ev_io *w, int revents)
 {
 	struct rspamd_worker *worker = (struct rspamd_worker *)w->data;
 	struct rspamd_proxy_ctx *ctx;
-	rspamd_inet_addr_t *addr;
+	rspamd_inet_addr_t *addr = NULL;
 	struct rspamd_proxy_session *session;
 	gint nfd;
 
@@ -2194,6 +2246,7 @@ proxy_accept_socket (EV_P_ ev_io *w, int revents)
 	}
 	/* Check for EAGAIN */
 	if (nfd == 0) {
+		rspamd_inet_address_free (addr);
 		return;
 	}
 
@@ -2322,6 +2375,8 @@ start_rspamd_proxy (struct rspamd_worker *worker)
 		/* Additional initialisation needed */
 		rspamd_worker_init_scanner (worker, ctx->event_loop, ctx->resolver,
 				&ctx->lang_det);
+		/* Always yse cfg->task_timeout */
+		ctx->task_timeout = rspamd_worker_check_and_adjust_timeout(ctx->cfg, NAN);
 
 		if (worker->index == 0) {
 			/*
@@ -2406,6 +2461,7 @@ start_rspamd_proxy (struct rspamd_worker *worker)
 
 	REF_RELEASE (ctx->cfg);
 	rspamd_log_close (worker->srv->logger);
+	rspamd_unset_crash_handler (worker->srv);
 
 	exit (EXIT_SUCCESS);
 }
