@@ -53,7 +53,13 @@ context("Selectors test", function()
 
     ["header Subject lower"] = {
                 selector = "header(Subject).lower",
-                expect = {"second, lower-cased header subject"}},
+                expect = {"second, lower-cased header subject"}
+    },
+
+    ["header Subject lower_utf8"] = {
+                selector = "header(Subject).lower_utf8",
+                expect = {"second, lower-cased header subject"}
+    },
 
     ["header full Subject lower"] = {
                 selector = "header(Subject, 'full').lower",
@@ -201,9 +207,10 @@ context("Selectors test", function()
       selector = "specific_urls({need_emails = true, limit = 2})",
       expect = {{"test@example.net", "http://subdomain.example.net"}}},
 
-    ["specific_urls + emails limit"] = {
-      selector = "specific_urls({need_emails = true, limit = 1})",
-      expect = {{"test@example.net"}}},
+    -- Broken test as order depends on the hash function internally
+    --["specific_urls + emails limit"] = {
+    --  selector = "specific_urls({need_emails = true, limit = 1})",
+    --  expect = {{"test@example.net"}}},
 
     ["pool_var str, default type"] = {
                 selector = [[pool_var("str_var")]],
