@@ -10,7 +10,7 @@ context("URL check functions", function()
   local ffi = require("ffi")
 
   ffi.cdef[[
-  void rspamd_http_normalize_path_inplace(char *path, size_t len, size_t *nlen);
+  void rspamd_normalize_path_inplace(char *path, size_t len, size_t *nlen);
   ]]
 
   test_helper.init_url_parser()
@@ -133,6 +133,16 @@ context("URL check functions", function()
     {"http://hehe｡example。com#test", true, {
       host = 'hehe.example.com', fragment = 'test'
     }},
+    {"http:////$%^&****((@example.org//#f@f", true, {
+      user = '$%^&****((', host = 'example.org', fragment = 'f@f'
+    }},
+    {"http://@@example.com", true, {
+      user = "@", host = "example.com"
+    }},
+    {"https://example.com\\_Resources\\ClientImages\\UserData?ol\\o#ololo\\", true, {
+      host = "example.com", path = "_Resources\\ClientImages\\UserData",
+      query = "ol\\o", fragment = "ololo\\",
+    }},
   }
 
   -- Some cases from https://code.google.com/p/google-url/source/browse/trunk/src/url_canon_unittest.cc
@@ -192,7 +202,7 @@ context("URL check functions", function()
       local buf = ffi.new("uint8_t[?]", #v[1])
       local sizbuf = ffi.new("size_t[1]")
       ffi.copy(buf, v[1], #v[1])
-      ffi.C.rspamd_http_normalize_path_inplace(buf, #v[1], sizbuf)
+      ffi.C.rspamd_normalize_path_inplace(buf, #v[1], sizbuf)
       local res = ffi.string(buf, tonumber(sizbuf[0]))
       assert_equal(v[2], res, 'expected ' .. v[2] .. ' but got ' .. res .. ' in path ' .. v[1])
     end)
