@@ -35,6 +35,8 @@ main (int argc, char **argv)
 	rspamd_main->server_pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), NULL, 0);
 	cfg = rspamd_config_new (RSPAMD_CONFIG_INIT_DEFAULT);
 	cfg->libs_ctx = rspamd_init_libs ();
+	/* More aggressive GC, workaround for 'not enough memory' test failures */
+	cfg->lua_gc_step *= 2;
 	rspamd_main->cfg = cfg;
 	cfg->cfg_pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), NULL, 0);
 
@@ -48,11 +50,9 @@ main (int argc, char **argv)
 	if (!g_option_context_parse (context, &argc, &argv, &error)) {
 		fprintf (stderr, "option parsing failed: %s\n", error->message);
 		g_option_context_free (context);
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 
-	/* Setup logger */
-
 	/* Setup logger */
 	if (verbose || g_test_verbose ()) {
 		rspamd_main->logger = rspamd_log_open_emergency (rspamd_main->server_pool,
