@@ -335,7 +335,7 @@ main (int argc, char **argv)
 	if (!g_option_context_parse (context, &argc, &argv, &error)) {
 		rspamd_fprintf (stderr, "option parsing failed: %s\n", error->message);
 		g_error_free (error);
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 
 	rspamd_parse_inet_address (&addr, host, 0);
